/* @(#) Attachments_FT.c 96/07/26 1.2 */

/* 
 *	File:				Attachments_FT.c
 *	Author:				Eric L. Hayes
 *	Creation Date:		
 *
 *	Last Modified by:	Eric L. Hayes
 *	Modification Date:	3/15/96
 *
 *	Version:			M2 2.0
 *		 
 *	This file contains the functional tests for attachments.  
 *
 *	Test 1:	Attach multiple samples to an instrument, link the attachments together, and 
 *			play all four attachments in order.
 *
 *	Test 2: Make an attachment independent to the instrument, then start instrument without
 *			starting the attachment.
 *
 *	Test 3:	Start instrument and start independent attachment.
 *
 *	Test 4:	Start instrument; start independent attachment; and stop attachment in the middle
 *			of play.  Call WhereAttachment to find out where we are within the attachment.
 *
 *
 *	Uses the following calls:
 *
 *	OpenAudioFolio	 		CreateCue			LoadInstrument		LoadSample
 *	GetCueSignal			CreateAttachment	LinkAttachments		StartAttachment
 *	ConnectInstrumentParts	StartInstrument		StopInstrument		WaitSignal
 *	MonitorAttachment		DeleteAttachment	UnloadInstrument	UnloadSample
 *	StopAttachment			WhereAttachment		SleepUntilTime
 *
 *	Dsp Instruments used:
 *
 *	sampler_16_f1.dsp
 *
 *	Audio files Used (These files must be present in the current directory)
 *
 *	Cowbell_44K_16B_M.aiff		Marimba_44K_16B_M.aiff		PanFlute_44K_16B_M.aiff		Trumpet_44K_16B_M.aiff
 *	LongFile_44K_16B_M.AIFF					
 *
 *  Copyright (C) 1996, an unpublished work by The 3DO Company.
 *  All rights reserved. This material contains confidential
 *  information that is the property of The 3DO Company. Any
 *  unauthorized duplication, disclosure or use is prohibited.
 *  
 */


#include <audio/audio.h>
#include <audio/parse_aiff.h>
#include <stdio.h>
#include <string.h>

#include "TestMacros.h"
#include <tests/3do_scaf.h>

/**************************************Test Environment Globals****************/
/******************************************************************************/

int main (int argc, char ** argv)
{
	Err		theError;
	int32	theSignal, bytePosition1, bytePosition2, signals[4], i;
	Item 	theItem, theLineOut, theSample, samples[4], theAttachment, attachments[4],
			theCue, cues[4];



	InitScaffold(AUDIO, argc, argv);					/* initializes this program */
	/******************************************************************************/

	theError = OpenAudioFolio();
	CHECKRESULT_SETUP(FT,theError,"OpenAudioFolio",cleanup);
	
	theCue	= CreateCue( NULL );
	CHECKRESULT_SETUP(FT,theCue,"CreateCue",cleanup);

	for ( i = 0; i < 4; i++ )
	{
		cues[i]	= CreateCue( NULL );
		CHECKRESULT_SETUP(FT,cues[i],"CreateCue",cleanup);
	}

	#define expected 0

	/****************************************************************************************************/
	/*	Load the sample instrument and the line_out instrument (which provides the sound).				*/
	/*	Load the sample and attach it to the sample instrument.  Call MonitorAttachment which will		*/
	/*	let you know through theCue when the end of the sample has been reached when the instrument is	*/
	/*	played.  theSignal is theSignal for that particular attachment.									*/
		
	theItem			= LoadInstrument("sampler_16_f1.dsp", 1, 100);
	CHECKRESULT_SETUP(FT, theItem,"LoadInstrument(sampler_16_f1.dsp)", cleanup);

	theLineOut		= LoadInstrument("line_out.dsp", 1, 100);
	CHECKRESULT_SETUP(FT, theLineOut,"LoadInstrument(line_out.dsp)", cleanup);

	samples[0]		= LoadSample("Cowbell_44K_16B_M.aiff");
	CHECKRESULT_SETUP(FT, samples[0],"LoadSample(Cowbell_44K_16B_M.aiff)", cleanup);
	
	samples[1]		= LoadSample("Marimba_44K_16B_M.aiff");
	CHECKRESULT_SETUP(FT, samples[1],"LoadSample(Marimba_44K_16B_M.aiff)", cleanup);
	
	samples[2]		= LoadSample("PanFlute_44K_16B_M.aiff");
	CHECKRESULT_SETUP(FT, samples[2],"LoadSample(PanFlute_44K_16B_M.aiff)", cleanup);
	
	samples[3]		= LoadSample("Trumpet_44K_16B_M.aiff");
	CHECKRESULT_SETUP(FT, samples[3],"LoadSample(Trumpet_44K_16B_M.aiff)", cleanup);

	theSample		= LoadSample("LongFile_44K_16B_M.AIFF");
	CHECKRESULT_SETUP(FT, theSample,"LoadSample(LongFile_44K_16B_M.AIFF)", cleanup);

	attachments[0]	= CreateAttachment( theItem, samples[0], NULL );
	CHECKRESULT_SETUP(FT, attachments[0],"CreateAttachment(theItem, Cowbell_44K_16B_M.aiff)", cleanup);
	
	attachments[1]	= CreateAttachment( theItem, samples[1], NULL );
	CHECKRESULT_SETUP(FT, attachments[1],"CreateAttachment(theItem, Marimba_44K_16B_M.aiff)", cleanup);
	
	attachments[2]	= CreateAttachment( theItem, samples[2], NULL );
	CHECKRESULT_SETUP(FT, attachments[2],"CreateAttachment(theItem, PanFlute_44K_16B_M.aiff)", cleanup);
	
	attachments[3]	= CreateAttachment( theItem, samples[3], NULL );
	CHECKRESULT_SETUP(FT, attachments[3],"CreateAttachment(theItem, Trumpet_44K_16B_M.aiff)", cleanup);

	theAttachment	= CreateAttachment( theItem, theSample, NULL );
	CHECKRESULT_SETUP(FT, theAttachment,"CreateAttachment(theItem, LongFile_44K_16B_M.AIFF)", cleanup);

	
	for ( i = 0; i < 4; i++ )
	{
		MonitorAttachment( attachments[i], cues[i], CUE_AT_END );
		signals[i]		= GetCueSignal( cues[i] );
	}
		
	MonitorAttachment( theAttachment, theCue, CUE_AT_END );
	theSignal		= GetCueSignal( theCue );

	LinkAttachments( attachments[0], attachments[1] );
	LinkAttachments( attachments[1], attachments[2] );
	LinkAttachments( attachments[2], attachments[3] );

	/****************************************************************************************************/
	/*	Monophonic instrument are listed before the 10th element in the structure						*/
	/*	Stereo instruments are found after the 10th element												*/
		
	ConnectInstrumentParts(	theItem, "Output", 0, 
							theLineOut, "Input", 0 );		/*	Connect the left channel	*/

	ConnectInstrumentParts(	theItem, "Output", 0, 
							theLineOut, "Input", 1 );		/*	Connect the right channel	*/

	/****************************************************************************************************/
	/****************************************************************************************************/
	/*	Test #1:	Play four linked samples															*/
	/****************************************************************************************************/
	BeginTest(FT, "Play four linked samples");

	theError	= StartInstrument( theItem, NULL );					/*	Play sound	*/
	LogComment("StartInstrument(sampler_16_f1.dsp) returned %d (0x%x)", theError, theError);

	theError	= StartInstrument( theLineOut, NULL );
	LogComment("StartInstrument(line_out.dsp) returned %d (0x%x)", theError, theError);
		
	/****************************************************************************************************/
		
	WaitSignal( signals[3] );							/*	Enter a Wait state until sample is finished	*/

	/****************************************************************************************************/
	theError	= StopInstrument( theItem, NULL );						/*	Stop Playing the sound	*/
	LogComment("StopInstrument(sampler_16_f1.dsp) returned %d (0x%x)", theError, theError);

	theError	= StopInstrument( theLineOut, NULL );
	LogComment("StopInstrument(line_out.dsp) returned %d (0x%x)", theError, theError);

	LogResult(InteractiveVerify("Did you hear four different samples being played consecutively?"), "");


	/****************************************************************************************************/
	/*	Test #2:	Start instrument, without starting independent sample															*/
	/****************************************************************************************************/
	BeginTest(FT, "Start instrument without starting independent attachment");

	for ( i = 0; i < 4; i++ )					/*	Delete the other attachments so they will not be mistakenly played */
		DeleteAttachment( attachments[i] );

	/*	Set this option so that the sample will not play when the instrument is started
	*/
	SetAudioItemInfoVA( theAttachment, AF_TAG_SET_FLAGS, AF_ATTF_NOAUTOSTART, TAG_END ); 


	theError	= StartInstrument( theItem, NULL );						/*	Play instrument	*/
	LogComment("StartInstrument(sampler_16_f1.dsp) returned %d (0x%x)", theError, theError);

	theError	= StartInstrument( theLineOut, NULL );
	LogComment("StartInstrument(line_out.dsp) returned %d (0x%x)", theError, theError);
		
	/****************************************************************************************************/
		
	SleepUntilTime( theCue, GetAudioTime() + 5*240 );					/*	Wait about five seconds	*/

	/****************************************************************************************************/
	theError	= StopInstrument( theItem, NULL );						/*	Stop Playing the instrument	*/
	LogComment("StopInstrument(sampler_16_f1.dsp) returned %d (0x%x)", theError, theError);

	theError	= StopInstrument( theLineOut, NULL );
	LogComment("StopInstrument(line_out.dsp) returned %d (0x%x)", theError, theError);

	LogResult(InteractiveVerify("You should not have heard anything. Did Test Pass?"), "");


	/****************************************************************************************************/
	/*	Test #3:	Start instrument, start independent sample															*/
	/****************************************************************************************************/

	BeginTest(FT, "Start instrument and independent attachment together");

	theError	= StartInstrument( theItem, NULL );						/*	Play instrument	*/
	LogComment("StartInstrument(sampler_16_f1.dsp) returned %d (0x%x)", theError, theError);

	theError	= StartInstrument( theLineOut, NULL );
	LogComment("StartInstrument(line_out.dsp) returned %d (0x%x)", theError, theError);
		
	theError	= StartAttachment( theAttachment, NULL );
	LogComment("StartAttachment() returned %d (0x%x)", theError, theError);
		
	/****************************************************************************************************/
		
	WaitSignal(theSignal);									/*	Wait until sample has finished playing	*/

	/****************************************************************************************************/
	theError	= StopInstrument( theItem, NULL );						/*	Stop Playing the instrument	*/
	LogComment("StopInstrument(sampler_16_f1.dsp) returned %d (0x%x)", theError, theError);

	theError	= StopInstrument( theLineOut, NULL );
	LogComment("StopInstrument(line_out.dsp) returned %d (0x%x)", theError, theError);

	theError	= StopAttachment( theAttachment, NULL );
	LogComment("StopAttachment() returned %d (0x%x)", theError, theError);

	LogResult(InteractiveVerify("You should have heard the entire sample. Did Test Pass?"), "");


	/****************************************************************************************************/
	/*	Test #4:	Start instrument, start independent sample, then stop attachment															*/
	/****************************************************************************************************/

	BeginTest(FT, "Start instrument and independent attachment together");

	theError	= StartInstrument( theItem, NULL );						/*	Play instrument	*/
	LogComment("StartInstrument(sampler_16_f1.dsp) returned %d (0x%x)", theError, theError);

	theError	= StartInstrument( theLineOut, NULL );
	LogComment("StartInstrument(line_out.dsp) returned %d (0x%x)", theError, theError);
		
	theError	= StartAttachment( theAttachment, NULL );
	LogComment("StartAttachment() returned %d (0x%x)", theError, theError);
		
	bytePosition1	= WhereAttachment( theAttachment );
	LogComment("WhereAttachment has returned %d", bytePosition1 );
	/****************************************************************************************************/
		
	SleepUntilTime( theCue, GetAudioTime() + 5*240 );					/*	Wait about five seconds	*/

	/****************************************************************************************************/
	bytePosition2	= WhereAttachment( theAttachment );
	LogComment("WhereAttachment has returned %d", bytePosition2 );

	if ( bytePosition2 > bytePosition1 )
		LogComment("PASSED: The value returned by the second call to WhereAttachment is greater than the value returned by the first call" );
	else
		LogComment("FAILED: The value returned by the second call to WhereAttachment is less than the value returned by the first call" );
		
	theError	= StopAttachment( theAttachment, NULL );
	LogComment("StopAttachment() returned %d (0x%x)", theError, theError);

	LogResult(InteractiveVerify("You should have heard only part of the sample. Did Test Pass?"), "");

	
	theError	= StopInstrument( theItem, NULL );						/*	Stop Playing the instrument	*/
	LogComment("StopInstrument(sampler_16_f1.dsp) returned %d (0x%x)", theError, theError);

	theError	= StopInstrument( theLineOut, NULL );
	LogComment("StopInstrument(line_out.dsp) returned %d (0x%x)", theError, theError);



	for ( i = 0; i < 4; i++ )
	{
		DeleteCue( cues[i] );
		DeleteAttachment( attachments[i] );
		UnloadSample( samples[i] );
	}
	
	DeleteCue( theCue );
	DeleteAttachment( theAttachment );
	UnloadSample( theSample );
	UnloadInstrument(theItem);
	UnloadInstrument(theLineOut);
	

	cleanup:
	theError	= CloseAudioFolio();
	CHECKRESULT(FT,theError,"CloseAudioFolio()");


	/******************************************************************************/
	TermScaffold();
	
	return 1;
}

