/* @(#) Trigger_FT.c 96/08/01 1.2 */
/*	
 *	File:				Trigger_FT.c
 *	Author:				Eric L. Hayes
 *	Creation Date:		
 *
 *	Last Modified by:	Eric L. Hayes
 *	Modification Date:	3/15/96
 *
 *	Version:			M2 2.0
 *
 *	This file contains the functional test for schmidt_trigger.dsp. schmidt_trigger.dsp is supposed to yield
 *	a TRUE output when the input rises above the signal level at the SetLevel knob, and a FALSE output when
 *	the input signal falls below the value of the signal at the resetLevel knob.  In any other case the
 *	output should remain the same.  If the resetLevel is greater than the setLevel knob, the schmidt_trigger
 *	will oscillate.  A structure contains the values for input, setLevel, resetLevel, as well as the expected
 *	output has been created.  A schmidt_trigger is created, given these test values and the expected output
 *	is compared with the actual output.  The test program also checks if the trigger has been set.
 *	If resetLevel is greater than setLevel, then the test program checks
 *	to see if the instrument is oscillating.  The structure has one additional element, startStop, which allows
 *	the user to stop and start an instrument before each test.
 *
 *	The folowing patches need to be in the current working directory in order for this test to run.
 *
 *		ampout.patch
 *
 *	Audio Functions called:
 *
 *		CloseAudioFolio		CreateCue			CreateKnob			LoadInstrument			CreateProbe			
 *		OpenAudioFolio		StartInstrument		StopInstrument		SleepUntilTime			SetKnob
 *		DeleteCue			DeleteProbe			ReadProbe			UnloadInstrument		UnloadPatchTemplate
 *		GetCueSignal		ConnectInstrumentParts
 *
 *  Copyright (C) 1996, an unpublished work by The 3DO Company.
 *  All rights reserved. This material contains confidential
 *  information that is the property of The 3DO Company. Any
 *  unauthorized duplication, disclosure or use is prohibited.
 *  
 */



#include <audio/audio.h>
#include <audio/patch.h>
#include <audio/patchfile.h>
#include <stdio.h>
#include <kernel/task.h>

#include "TestMacros.h"
#include <tests/3do_scaf.h>

/**************************************Test Environment Globals****************/
/******************************************************************************/

typedef	struct
		{
			float32	input;
			float32	setLevel;
			float32	resetLevel;
			Boolean	expectedOutput;
			Boolean	startStop;
		}  TestValues;


#define	NUM_OF_TESTS		6


int main (int argc, char ** argv)
{
	Boolean		oscillate	= FALSE;
	Err			theError;
	float32		theOutputValue, theOutputValue2, previousOutputValue = FALSE;
	int			i, j;
	int32		sigMask;
	Item 		theItem, theTestProbe, theInputVoltage, constantVoltagePatch, theVoltageKnob,
				sLKnob, rLKnob, theCue, sleepCue;
	TestValues	testList[NUM_OF_TESTS]	= 	{
												0,		0,		0,		FALSE,	FALSE,
												.6,		.4,		-.9,	TRUE,	FALSE,
												.5,		0,		0,		TRUE,	FALSE,
												.8,		.5,		-.1,	TRUE,	FALSE,
												-.6,	.5,		-.1,	FALSE,	FALSE,
												.1,		.4,		-.9,	FALSE,	FALSE
											
											};



	InitScaffold(AUDIO, argc, argv);					/* initializes this program */
	/******************************************************************************/

	theError = OpenAudioFolio();
	CHECKRESULT_SETUP(FT,theError,"OpenAudioFolio",cleanup);
	
	theCue	= CreateCue( NULL );
	CHECKRESULT_SETUP(FT,theCue,"CreateCue",cleanup);

	sleepCue	= CreateCue( NULL );
	CHECKRESULT_SETUP(FT,sleepCue,"CreateCue",cleanup);

	sigMask	= GetCueSignal( theCue );
	LogComment ( "SigMask is 0x%x.\t\tGetCurrentSignals is  0x%x.", sigMask, GetCurrentSignals() );

	/****************************************************************************************************
	 *	Load the schmidt trigger instrument and the constant voltage patch which will be used as the
	 *	schmidt trigger's input.  Create the knobs for the schmidt trigger and create the knob for the
	 *	constant voltage patch.
	 ****************************************************************************************************/
		
	theItem			= LoadInstrument("schmidt_trigger.dsp", 1, 100);
		
	sLKnob	= CreateKnob( theItem, "SetLevel", NULL );
	rLKnob	= CreateKnob( theItem, "ResetLevel", NULL );


	theTestProbe	= CreateProbeVA( theItem, "Output", AF_TAG_TYPE, AF_SIGNAL_TYPE_WHOLE_NUMBER, TAG_END );
	
	/*	Create an instrument patch which supplies a constant voltage. Set the voltage to .1	*/
	constantVoltagePatch	= LoadPatchTemplate( "ampout.patch" );
	theInputVoltage			= CreateInstrument (constantVoltagePatch, NULL);
		
	if ( theInputVoltage < 0 )
	{
		LogComment("Could not create an instrument for  ampout.patch" );
		PrintfSysErr(theInputVoltage);
	}
		
	theVoltageKnob			= CreateKnob( theInputVoltage, "Amp", NULL );

	/*	Connect the constant voltage instrument to the schmidt trigger		*/		
	ConnectInstrumentParts(	theInputVoltage, "Output", 0, theItem, "Input", 0 	);



	StartInstrument( theInputVoltage, NULL );
	StartInstrument( theItem, NULL );

	theError = ArmTrigger( theItem, NULL, theCue, AF_F_TRIGGER_CONTINUOUS | AF_F_TRIGGER_RESET );
	CHECKRESULT_SETUP(FT,theError,"ArmTrigger",cleanup);

	LogComment ( "SigMask is 0x%x.\t\tGetCurrentSignals is  0x%x.", sigMask, GetCurrentSignals() );
	
	for ( i = 0; i < NUM_OF_TESTS; i++ )
	{

		/*	Set the values for the input signal and the reset and set knobs	*/
		SetKnob( theVoltageKnob, testList[i].input );
		SetKnob( sLKnob, testList[i].setLevel );
		SetKnob( rLKnob, testList[i].resetLevel );

		SleepUntilTime( sleepCue, GetAudioTime() + 2 );
		ReadProbe( theTestProbe, &theOutputValue );
	
		if ( testList[i].resetLevel > testList[i].setLevel )
		{
			for ( j = 0; j < 2; j++ )
			{
				ReadProbe( theTestProbe, &theOutputValue );
				SleepUntilTime( sleepCue, GetAudioTime() + 1 );
				ReadProbe( theTestProbe, &theOutputValue2 );
				
				if ( theOutputValue != theOutputValue2 )
					oscillate	= TRUE;

				ReadProbe( theTestProbe, &theOutputValue );
				SleepUntilTime( sleepCue, GetAudioTime() + 2 );
				ReadProbe( theTestProbe, &theOutputValue2 );
				
				if ( theOutputValue != theOutputValue2 )
					oscillate	= TRUE;

				ReadProbe( theTestProbe, &theOutputValue );
				SleepUntilTime( sleepCue, GetAudioTime() + 3 );
				ReadProbe( theTestProbe, &theOutputValue2 );
				
				if ( theOutputValue != theOutputValue2 )
					oscillate	= TRUE;
				
			}
			if ( oscillate  == TRUE )
				LogComment("%2d. Output is oscillating when Input = %f;  SetLevel = %f; ResetLevel = %f.",
								i, testList[i].input, testList[i].setLevel, testList[i].resetLevel );
			else
				LogComment("%2d. Output is not oscillating when Input = %f;  SetLevel = %f; ResetLevel = %f.",
								i, testList[i].input, testList[i].setLevel, testList[i].resetLevel );
			
			oscillate			= FALSE;
			previousOutputValue = FALSE;
		}
		else if ( testList[i].expectedOutput  == TRUE )
		{
			if ( (Boolean)theOutputValue  > 0 )
			{
				LogComment("%2d. Output is %d when Input = %f;  SetLevel = %f; ResetLevel = %f.",
								i, testList[i].expectedOutput, testList[i].input, testList[i].setLevel, testList[i].resetLevel );

				if ( previousOutputValue == FALSE )
				{
					LogComment ( "SigMask is 0x%x.\t\tGetCurrentSignals is  0x%x.", sigMask, GetCurrentSignals() );
					BeginTest(FT, "Trigger");
					if ( sigMask  == (GetCurrentSignals() & sigMask) )
					{
						LogResult(PASSED, "");
						theError = ClearCurrentSignals( sigMask );
						LogComment("ClearCurrentSignals returned %d", theError);
					}
					else
						LogResult(FAILED, "%2d. Trigger has not been sent when output has gone from FALSE to TRUE", i );
				}
				else
				{
					LogComment ( "SigMask is 0x%x.\t\tGetCurrentSignals is  0x%x.", sigMask, GetCurrentSignals() );
					BeginTest(FT, "Trigger");

					if ( sigMask  != (GetCurrentSignals() & sigMask) )
						LogResult(PASSED, "");
					else
					{
						LogResult(FAILED, "%2d. Trigger has been sent when output has gone from TRUE to TRUE", i );
						theError = ClearCurrentSignals( sigMask );
						LogComment("ClearCurrentSignals returned %d", theError);
					}
				}
				previousOutputValue = TRUE;
			}
			else
			{
				LogComment("%2d. Expected %d, but got %d when Input = %f;  SetLevel = %f; ResetLevel = %f.",
								i, testList[i].expectedOutput, (Boolean)theOutputValue, testList[i].input, testList[i].setLevel, testList[i].resetLevel );
			
				previousOutputValue = FALSE;
			}
		}
		else
		{
			if ( (Boolean)theOutputValue  == 0 )
			{
				LogComment("%2d. Output is %d when Input = %f;  SetLevel = %f; ResetLevel = %f.",
								i, testList[i].expectedOutput, testList[i].input, testList[i].setLevel, testList[i].resetLevel );
				
				previousOutputValue = FALSE;

				LogComment ( "SigMask is 0x%x.\t\tGetCurrentSignals is  0x%x.", sigMask, GetCurrentSignals() );
				
				BeginTest(FT, "Trigger");
				if ( (GetCurrentSignals() & sigMask)  != sigMask )
					LogResult(PASSED, "");
				else
				{
					LogResult(FAILED, "%2d. Trigger has been sent when output has gone FALSE to FALSE", i );
					theError = ClearCurrentSignals( sigMask );
					LogComment("ClearCurrentSignals returned %d", theError);
				}

			}
			else
			{
				LogComment("%2d. Expected %d, but got %d when Input = %f;  SetLevel = %f; ResetLevel = %f.",
								i, testList[i].expectedOutput, (Boolean)theOutputValue, testList[i].input, testList[i].setLevel, testList[i].resetLevel );
				
				previousOutputValue = TRUE;
			}	
		}
	
	}
	StopInstrument( theInputVoltage, NULL );
	StopInstrument( theItem, NULL );



		
	
	DeleteCue( theCue );
	DeleteCue( sleepCue );
	UnloadInstrument( theItem );
	UnloadPatchTemplate( constantVoltagePatch );
	
	cleanup:
	theError	= CloseAudioFolio();
	CHECKRESULT(FT, theError,"CloseAudioFolio");


	/******************************************************************************/
	TermScaffold();
	
	return 1;
}
