/* @(#) Delay_F1_FT.c 96/07/29 1.2 */
/*	
 *	File:				Delay_F1_FT.c
 *	Author:				Eric L. Hayes
 *	Creation Date:		
 *
 *	Last Modified by:	Eric L. Hayes
 *	Modification Date:	3/15/96
 *
 *	Version:			M2 2.0
 *
 *	This file contains the functional tests for delay_f1.dsp.  This program creates a simple reverb loop using
 *	a delay line with the characteristics the user specifies.  The user can set three parameters on the command
 *	line.  The user can set the size of the buffer in bytes (-s), the size of the delay in frames (-d), and the
 *	time for which the sample plays (-t).  Each parameter has a default value.  The user should be able to hear
 *	the delay in the played sample provided  that the playing time is long enough in respect to the length of the
 *	delay.
 *
 *	Audio Functions called:
 *
 *		CloseAudioFolio		CreateCue			CreateAttachment		LoadInstrument			LoadSample
 *		OpenAudioFolio		CreateDelayLine		CreateMixerTemplate		CreateInstrument		CreateKnob
 *		SetAudioItemInfoVA	CalcMixerGainPart	SetKnobPart				StartInstrument			StopInstrument
 *		SleepUntilTime		DeleteAttachment	DeleteCue				DeleteDelayLine			UnloadInstrument
 *		UnloadSample
 *
 *  Copyright (C) 1996, an unpublished work by The 3DO Company.
 *  All rights reserved. This material contains confidential
 *  information that is the property of The 3DO Company. Any
 *  unauthorized duplication, disclosure or use is prohibited.
 *  
 */



#include <audio/audio.h>
#include <audio/parse_aiff.h>
#include <stdio.h>
#include <stdlib.h>

#include <TestMacros.h>
#include <TestUtilities.h>
#include <tests/3do_scaf.h>


#define	NUM_CHANNELS				2
#define	SAMPLE_DURATION_IN_SECS		2
#define	SAMPLES_PER_SECOND			44100
#define	BYTES_PER_SAMPLE			2

#define	PLAY_DURATION				2400

#define	IF_LOOP			TRUE


/**************************************Test Environment Globals****************/
/******************************************************************************/

int main (int argc, char ** argv)
{
	
	char		c;
	Err			Result;
	int			newArgc;
	int32		theMixerKnobPartNum,
				delaySize				= NUM_CHANNELS*SAMPLE_DURATION_IN_SECS*SAMPLES_PER_SECOND*BYTES_PER_SAMPLE,
				delayFrames				= 11025,
				playTime				= PLAY_DURATION;
	Item 		delayLine, delayInstrument, sampleReader, sampleWriter, sampleSource,
				theCue, delayAttachment, sampleReaderAttachment, sampleWriterAttachment,
				theMixerTemplate, mixerInstrument, theMixerKnob, lineOut;
	MixerSpec	theMixerSpec;



	newArgc	= InitScaffold(AUDIO, argc, argv);					/* initializes this program */
	/******************************************************************************/

	while( --newArgc > 0  && (*++argv)[0] == '-' )
	{
		c	= *++argv[0];
		switch( c )
		{
			case	's'	:									/*	Get the size of the delay line in bytes	*/
							if ( --argc > 0 )
							{
								delaySize	= atol( *++argv );
								if ( (delaySize < 0) || (delaySize > 4000000) ) 
								{
									printf( "\nDelay_F1_FT: %d is out of range.\tThe delay Line size must be between 0 and 4 Mb\n", delaySize );
									exit(1);
								}
							}
							break;
			case	'd'	:									/*	Get the delay	*/
							if ( --argc > 0 )
							{
								delayFrames	= atol( *++argv );
								if ( delayFrames < 32 ) 
								{
									printf( "\nDelay_F1_FT: %d is out of range.\tDelay must be greater than 32 frames\n", delayFrames );
									exit(1);
								}
							}
							break;
			case	't'	:									/*	Get duration (the time to wait before reading probes)  from the command line	*/
							if ( --argc > 0 )
							{
								playTime	= atol( *++argv )*240;
								if ( playTime < 0 )
								{
									printf( "\nDelay_F1_FT: %d is out of range.\tPlaying time must be greater than 0 secs\n", playTime );
									exit(1);
								}
							}
							break;
			default		:	
							printf( "\nDelay_F1_FT: Illegal option -%c\n", c );
							printf( "\nUsage:\tDelay_F1_FT:\t-s (delay line size) -d (delay) -t (playing time)\n" );
							exit(1);	
							break;
		}
	}
	


	Result = OpenAudioFolio();
	CHECKRESULT_SETUP(FT, Result,"OpenAudioFolio", cleanup);
	
	theCue	= CreateCue( NULL );
	CHECKRESULT_SETUP(FT, theCue,"CreateCue", cleanup);

	#define expected 0


	/*	Create the delay line, the delay instrument which writes to the delay line, and attach them.
	*/
	delayLine 		= CreateDelayLine( delaySize, NUM_CHANNELS, IF_LOOP );
	CHECKRESULT_SETUP(FT, delayLine,"CreateDelayLine(size, 2, TRUE)", cleanup);

	delayInstrument = LoadInstrument("delay_f1.dsp", 0, 100);
	CHECKRESULT_SETUP(FT, delayInstrument,"LoadInstrument(delay_f1.dsp)", cleanup);

	delayAttachment = CreateAttachment( delayInstrument, delayLine, NULL );
	CHECKRESULT_SETUP(FT, delayAttachment,"CreateAttachment(delayInstrument, delayLine)", cleanup);


	/*	Load the instrument which will play the sample to be written to the delay line,  load the
		sample, and attach them
	*/
	sampleWriter = LoadInstrument("sampler_16_f1.dsp", 0, 100);
	CHECKRESULT_SETUP(FT, sampleWriter,"LoadInstrument(sampler_16_f1.dsp)", cleanup);

	sampleSource = LoadSample("44K_16B_M.AIFF");
	CHECKRESULT_SETUP(FT, sampleSource,"LoadSample(44K_16B_M.AIFF)", cleanup);

	sampleWriterAttachment = CreateAttachment( sampleWriter, sampleSource, NULL );
	CHECKRESULT_SETUP(FT, sampleWriterAttachment,"CreateAttachment(sampleWriter, sampleSource", cleanup);


	/*	Load the sample instrument which will read the info from the delay line and attach it to
		the delay line
	*/
	sampleReader = LoadInstrument("sampler_16_f1.dsp", 0, 100);
	CHECKRESULT_SETUP(FT, sampleReader,"LoadInstrument(sampler_16_f1.dsp)", cleanup);

	sampleReaderAttachment = CreateAttachment( sampleReader, delayLine, NULL );
	CHECKRESULT_SETUP(FT, sampleReaderAttachment,"CreateAttachment(sampleReader, delayLine", cleanup);


	/*	Start writing to the delay line at some specified point
	*/
	Result = SetAudioItemInfoVA( sampleReaderAttachment,
                                 AF_TAG_START_AT, delayFrames,
                                 TAG_END );
	CHECKRESULT_SETUP(FT, Result,"SetAudioItemInfo: START_AT", cleanup);


	/*	Load the line_out Instrument
	*/
	lineOut = LoadInstrument("line_out.dsp", 0, 100);
	CHECKRESULT_SETUP(FT, lineOut,"LoadInstrument(line_out)", cleanup);



	/*	Create Mixer	*/
	theMixerSpec		= MakeMixerSpec( 2, 2, 0);
		
	theMixerTemplate	= CreateMixerTemplate( theMixerSpec, NULL );
	mixerInstrument		= CreateInstrument( theMixerTemplate, NULL );
	CHECKRESULT_SETUP(FT, mixerInstrument,"CreateInstrument(theMixerTemplate)", cleanup);

	theMixerKnob		= CreateKnob( mixerInstrument, "Gain", NULL );

	/*	Set the gain between the input and the output			*/
	theMixerKnobPartNum	= CalcMixerGainPart( theMixerSpec, 0, 0 );
	SetKnobPart( theMixerKnob, theMixerKnobPartNum, .5 );
		
	theMixerKnobPartNum	= CalcMixerGainPart( theMixerSpec, 0, 1 );
	SetKnobPart( theMixerKnob, theMixerKnobPartNum, .5 );
			
	theMixerKnobPartNum	= CalcMixerGainPart( theMixerSpec, 1, 0 );
	SetKnobPart( theMixerKnob, theMixerKnobPartNum, 0.5 );
		
	/*	Set the gain between the input and the output	(feedback loop)		*/
	theMixerKnobPartNum	= CalcMixerGainPart( theMixerSpec, 1, 1 );
	SetKnobPart( theMixerKnob, theMixerKnobPartNum, -.8 );
			
	
	/*	Connect all of the instruments to the mixer
	*/
	Result	= ConnectInstrumentParts(	sampleWriter, "Output", 0, mixerInstrument, "Input", 0 		);
	CHECKRESULT_SETUP(FT, Result,"ConnectInstrumentParts(sampleWriter Output 0, Mixer Input 0)", cleanup);

	Result	= ConnectInstrumentParts(	sampleReader, "Output", 0, mixerInstrument, "Input", 1 		);
	CHECKRESULT_SETUP(FT, Result,"ConnectInstrumentParts(sampleReader Output 0, Mixer Input 1)", cleanup);


	/*	Connect mixer to output for sound
	*/
	Result	= ConnectInstrumentParts(	mixerInstrument, "Output", 0, lineOut, "Input", 0 		);
	CHECKRESULT_SETUP(FT, Result,"ConnectInstrumentParts(Mixer Output 0, Line_Out Input 0)", cleanup);
	Result	= ConnectInstrumentParts(	mixerInstrument, "Output", 0, lineOut, "Input", 1 		);
	CHECKRESULT_SETUP(FT, Result,"ConnectInstrumentParts(Mixer Output 0, Line_Out Input 1)", cleanup);
							

	/*	Connnect feedback loop
	*/
	Result	= ConnectInstrumentParts(	mixerInstrument, "Output", 1, delayInstrument, "Input", 0 		);
	CHECKRESULT_SETUP(FT, Result,"ConnectInstrumentParts(Mixer Output 1, Delay Input 0)", cleanup);

	LogComment("Start playing sample with %d frame delay for %d seconds.", delayFrames, playTime/240 );

	Result	= StartInstrument( delayInstrument, NULL );
	CHECKRESULT_SETUP(FT, Result,"StartInstrument(delayInstrument, NULL)", cleanup);

	Result	= StartInstrument( sampleReader, NULL );
	CHECKRESULT_SETUP(FT, Result,"StartInstrument(sampleReader, NULL)", cleanup);

	Result	= StartInstrument( sampleWriter, NULL ); 
	CHECKRESULT_SETUP(FT, Result,"StartInstrument(sampleWriter, NULL)", cleanup);

	Result	= StartInstrument( lineOut, NULL );
	CHECKRESULT_SETUP(FT, Result,"StartInstrument(lineOut, NULL)", cleanup);

	Result	= StartInstrument( mixerInstrument, NULL );
	CHECKRESULT_SETUP(FT, Result,"StartInstrument(mixerInstrument, NULL)", cleanup);
	
	Result	= SleepUntilTime( theCue, GetAudioTime() + playTime );
	CHECKRESULT(FT, Result,"SleepUntilTime()");
	
	LogComment("Stop playing sample.", delayFrames );
	

	StopInstrument( lineOut, NULL );
	StopInstrument( mixerInstrument, NULL );
	StopInstrument( delayInstrument, NULL );
	StopInstrument( sampleWriter, NULL );
	StopInstrument( sampleReader, NULL );
	
	BeginTest(FT, "delay_f1.dsp");
	LogResult(InteractiveVerify("You should have heard a sample, followed by the specified delay, and then softer echoes of the sample. Did Test Pass?"), "");

	DeleteCue( theCue );
	
	DeleteAttachment( delayAttachment );
	DeleteAttachment( sampleReaderAttachment );
	DeleteAttachment( sampleWriterAttachment );

	DeleteDelayLine( delayLine );

	UnloadSample( sampleSource );
	UnloadInstrument( sampleReader );
	UnloadInstrument( sampleWriter );
	UnloadInstrument( delayInstrument );
	UnloadInstrument( lineOut );
	UnloadInsTemplate( theMixerTemplate );

cleanup:
	Result	= CloseAudioFolio();
	CHECKRESULT(FT, Result,"CloseAudioFolio");

	/******************************************************************************/
	TermScaffold();

	return 0;
}
