/* @(#) PlayArithmeticDSPs_FT.c 96/07/29 1.4 */
/*	
 *	File:				PlayArithmeticDSPs_FT.c
 *	Author:				Eric L. Hayes
 *	Creation Date:		
 *
 *	Last Modified by:	Eric L. Hayes
 *	Modification Date:	3/15/96
 *
 *	Version:			M2 2.0
 *
 *	This file contains the functional tests for the arithmetic dsps.  The test program gets the
 *	name of the arithmetic instrument from PlayArithmeticDSPs_FT_data.h.  Knobs are created for
 *	the instrument and the knobs are set to particular values which should yield some expected
 *	output.  The output from the instrument is read by a probe and compared to the expected output.
 *
 *	The last set of tests are for the cubic_amplifier.dsp.  This instrument is given a range of
 *	input and amplitude values.  The output is probed to see if it is correct.
 *
 *	The folowing patches need to be in the current working directory in order for this test to run.
 *
 *		ampout.patch
 *
 *	Audio Functions called:
 *
 *		CloseAudioFolio				CreateProbe			LoadInstrument			
 *		OpenAudioFolio				CreateKnob			SetKnobPart			StartInstrument
 *		StopInstrument				UnloadInstrument
 *
 *	Dsp Instruments Tested:
 *
 *		add.dsp				maximum.dsp			minimum.dsp				multiply.dsp
 *		subtract.dsp		timesplus.dsp		timesplus_noclip.dsp
 *
 *  Copyright (C) 1996, an unpublished work by The 3DO Company.
 *  All rights reserved. This material contains confidential
 *  information that is the property of The 3DO Company. Any
 *  unauthorized duplication, disclosure or use is prohibited.
 *  
 *
 */



#include <audio/audio.h>
#include <audio/patch.h>
#include <audio/patchfile.h>
#include <math.h>
#include <stdio.h>

#include "TestMacros.h"
#include <tests/3do_scaf.h>


#define		SOUND_DURATION		240


/**************************************Test Environment Globals****************/
/******************************************************************************/

int main (int argc, char ** argv)
{
	#include "PlayArithmeticDSPs_FT_data.h"
	
	Err		theError;
	int32	p[10], i;
	float32	amplitudeGotten, amplitudeExpected, amplitudeA, amplitudeB, amplitudeC;
	Item 	InputSoundA, InputSoundB, InputSoundC = 0, theTestProbe,
			theTestInstrument, theAmplitude,
			theInputVoltage, constantVoltagePatch, theVoltageKnob;



	InitScaffold(AUDIO, argc, argv);					/* initializes this program */
	/******************************************************************************/

	theError = OpenAudioFolio();
	CHECKRESULT_SETUP(FT, theError,"OpenAudioFolio",cleanup);
	
	/****************************************************************************************************/
	/****************************************************************************************************/
	/*	Do tests																					*/
	/****************************************************************************************************/
	#include "DepthTest_TopPortion.h"

	theTestInstrument	= LoadInstrument(ARGUMENT_01, 1, 100);

		
	InputSoundA			= CreateKnob(theTestInstrument, "InputA", NULL);
	InputSoundB			= CreateKnob(theTestInstrument, "InputB", NULL);
	

	theTestProbe		= CreateProbe( theTestInstrument, "Output", NULL );
	
	/*	Set the characteristics for the input sounds 				*/
	switch( p[1] )
	{
		case 0 :	/*	Add.dsp		+ inputs	*/
					SetKnobPart( InputSoundA, 0, .5 );		
					SetKnobPart( InputSoundB, 0, 0.25 );		
					amplitudeExpected	= .75;		
					break;
					
		case 1 :	/*	Add.dsp		- inputs	*/
					SetKnobPart( InputSoundA, 0, -.5 );		
					SetKnobPart( InputSoundB, 0, -0.25 );		
					amplitudeExpected	= -.75;		
					break;
					
		case 2 :	/*	Add.dsp		+ clip	*/
					SetKnobPart( InputSoundA, 0, .5 );		
					SetKnobPart( InputSoundB, 0, 0.9 );		
					amplitudeExpected	= 1.0;		
					break;
					
		case 3 :	/*	Add.dsp		- clip	*/
					SetKnobPart( InputSoundA, 0, -.5 );		
					SetKnobPart( InputSoundB, 0, -0.9 );		
					amplitudeExpected	= -1.0;		
					break;
					
		case 4 :	/*	Multiply.dsp	+ inputs	*/
					SetKnobPart( InputSoundA, 0, .4 );		
					SetKnobPart( InputSoundB, 0, .5 );		
					amplitudeExpected	= .2;		
					break;
					
		case 5 :	/*	Multiply.dsp	- inputs	*/
					SetKnobPart( InputSoundA, 0, -.4 );		
					SetKnobPart( InputSoundB, 0, -.5 );		
					amplitudeExpected	= .2;		
					break;
					
		case 6 :	/*	Multiply.dsp	+/- inputs	*/
					SetKnobPart( InputSoundA, 0, .4 );		
					SetKnobPart( InputSoundB, 0, -.5 );		
					amplitudeExpected	= -.2;		
					break;
					
		case 7 :	/*	Multiply.dsp	+/- inputs	*/
					SetKnobPart( InputSoundA, 0, -.4 );		
					SetKnobPart( InputSoundB, 0, .5 );		
					amplitudeExpected	= -.2;		
					break;
					
		case 8 :	/*	Multiply_unsigned.dsp	+ inputs	*/
					SetKnobPart( InputSoundA, 0, .4 );		
					SetKnobPart( InputSoundB, 0, .5 );		
					amplitudeExpected	= .2;		
					break;
					
		case 9 :	/*	Multiply_unsigned.dsp.dsp	+ inputs greater than 1.0	*/
					SetKnobPart( InputSoundA, 0, 1.3 );		
					SetKnobPart( InputSoundB, 0, 1.2 );		
					amplitudeExpected	= 1.56;		
					break;
					
		case 10 :	/*	Multiply_unsigned.dsp.dsp	Mixed inputs	*/
					SetKnobPart( InputSoundA, 0, 1.5 );		
					SetKnobPart( InputSoundB, 0, .6 );		
					amplitudeExpected	= .9;		
					break;
					
		case 11 :	/*	Multiply_unsigned.dsp.dsp	Clipped output	*/
					SetKnobPart( InputSoundA, 0, 1.8 );		
					SetKnobPart( InputSoundB, 0, 1.6 );		
					amplitudeExpected	= 2.0;		
					break;
					
		case 12 :	/*	Subtract.dsp	+ inputs	*/
					SetKnobPart( InputSoundA, 0, 1.0 );		
					SetKnobPart( InputSoundB, 0,  0.75 );		
					amplitudeExpected	= 0.25;		
					break;
					
		case 13 :	/*	Subtract.dsp	- inputs	*/
					SetKnobPart( InputSoundA, 0, -0.4 );		
					SetKnobPart( InputSoundB, 0,  -0.3 );		
					amplitudeExpected	= -0.1;		
					break;
					
		case 14 :	/*	Subtract.dsp	+/- inputs	*/
					SetKnobPart( InputSoundA, 0, -0.4 );		
					SetKnobPart( InputSoundB, 0,  0.3 );		
					amplitudeExpected	= -0.7;		
					break;
					
		case 15 :	/*	Subtract.dsp	- clip	*/
					SetKnobPart( InputSoundA, 0, -0.4 );		
					SetKnobPart( InputSoundB, 0, 0.8 );		
					amplitudeExpected	= -1.0;		
					break;
					
		case 16 :	/*	Minimum.dsp		+ inputs	*/
					SetKnobPart( InputSoundA, 0, .25 );		
					SetKnobPart( InputSoundB, 0, 1.0 );		
					amplitudeExpected	= .25;		
					break;
					
		case 17 :	/*	Minimum.dsp		- inputs	*/
					SetKnobPart( InputSoundA, 0, -.25 );		
					SetKnobPart( InputSoundB, 0, -1.0 );		
					amplitudeExpected	= -1.0;		
					break;
					
		case 18 :	/*	Minimum.dsp		+/- inputs	*/
					SetKnobPart( InputSoundA, 0, .25 );		
					SetKnobPart( InputSoundB, 0, -1.0 );		
					amplitudeExpected	= -1.0;		
					break;
					
		case 19 :	/*	Minimum.dsp		same inputs	*/
					SetKnobPart( InputSoundA, 0, .25 );		
					SetKnobPart( InputSoundB, 0, .25 );		
					amplitudeExpected	= .25;		
					break;
					
		case 20 :	/*	Maximum.dsp		+ inputs	*/
					SetKnobPart( InputSoundA, 0, 1.0 );		
					SetKnobPart( InputSoundB, 0,  0.25 );		
					amplitudeExpected	= 1.0;		
					break;
					
		case 21 :	/*	Maximum.dsp		- inputs	*/
					SetKnobPart( InputSoundA, 0, -1.0 );		
					SetKnobPart( InputSoundB, 0,  -0.25 );		
					amplitudeExpected	= -0.25;		
					break;
					
		case 22 :	/*	Maximum.dsp		+/- inputs	*/
					SetKnobPart( InputSoundA, 0, -1.0 );		
					SetKnobPart( InputSoundB, 0,  0.25 );		
					amplitudeExpected	= .25;		
					break;
					
		case 23 :	/*	Maximum.dsp		same inputs	*/
					SetKnobPart( InputSoundA, 0, 0.25 );		
					SetKnobPart( InputSoundB, 0,  0.25 );		
					amplitudeExpected	= 0.25;		
					break;
					
		case 24 :	/*	TimesPlus.dsp	+ inputs	*/
					SetKnobPart( InputSoundA, 0, .5 );		
					SetKnobPart( InputSoundB, 0, .5 );
					
					InputSoundC		= CreateKnob(theTestInstrument, "InputC", NULL);
					SetKnobPart( InputSoundC, 0, .25 );
					
					amplitudeExpected	= .5;		
					break;
					
		case 25 :	/*	TimesPlus.dsp	- inputs	*/
					SetKnobPart( InputSoundA, 0, -.5 );		
					SetKnobPart( InputSoundB, 0, -.5 );
					
					InputSoundC		= CreateKnob(theTestInstrument, "InputC", NULL);
					SetKnobPart( InputSoundC, 0, -.25 );
					
					amplitudeExpected	= 0.0;		
					break;
					
		case 26 :	/*	TimesPlus.dsp	+ clip	*/
					SetKnobPart( InputSoundA, 0, .7 );		
					SetKnobPart( InputSoundB, 0, .7 );
					
					InputSoundC		= CreateKnob(theTestInstrument, "InputC", NULL);
					SetKnobPart( InputSoundC, 0, .8 );
					
					amplitudeExpected	= 1.0;		
					break;
					
		case 27 :	/*	TimesPlus.dsp	- clip	*/
					SetKnobPart( InputSoundA, 0, -.7 );		
					SetKnobPart( InputSoundB, 0, .7 );
					
					InputSoundC		= CreateKnob(theTestInstrument, "InputC", NULL);
					SetKnobPart( InputSoundC, 0, -.8 );
					
					amplitudeExpected	= -1.0;		
					break;
					
		case 28 :	/*	TimesPlus_noclip.dsp	+ inputs	*/
					SetKnobPart( InputSoundA, 0, .4 );		
					SetKnobPart( InputSoundB, 0, .8 );
					
					InputSoundC		= CreateKnob(theTestInstrument, "InputC", NULL);
					SetKnobPart( InputSoundC, 0, .3 );
					
					amplitudeExpected	= .62;		
					break;
					
		case 29 :	/*	TimesPlus_noclip.dsp	- inputs	*/
					SetKnobPart( InputSoundA, 0, -.4 );		
					SetKnobPart( InputSoundB, 0, .8 );
					
					InputSoundC		= CreateKnob(theTestInstrument, "InputC", NULL);
					SetKnobPart( InputSoundC, 0, -.3 );
					
					amplitudeExpected	= -.62;		
					break;
					
		case 30 :	/*	TimesPlus_noclip.dsp	+ clip	*/
					SetKnobPart( InputSoundA, 0, .8 );		
					SetKnobPart( InputSoundB, 0, .9 );
					
					InputSoundC		= CreateKnob(theTestInstrument, "InputC", NULL);
					SetKnobPart( InputSoundC, 0, .75 );
					
					amplitudeExpected	= -.53;	/*	This value was chosen because it is the two's complement 
													of .47.  The output you see depends on whether the probe
													interprets the value as signed or unsigned.  Confusing,
													but that is how it is
												*/
					break;
					
		case 31 :	/*	TimesPlus_noclip.dsp	- clip	*/
					SetKnobPart( InputSoundA, 0, -.8 );		
					SetKnobPart( InputSoundB, 0, .9 );
					
					InputSoundC		= CreateKnob(theTestInstrument, "InputC", NULL);
					SetKnobPart( InputSoundC, 0, -.75 );
					
					amplitudeExpected	= .53;		
					break;
					
		default :	/*	Any other	*/
					SetKnobPart( InputSoundA, 0, 1.0 );		
					SetKnobPart( InputSoundB, 0, 1.0 );		
					
					InputSoundC		= CreateKnob(theTestInstrument, "InputC", NULL);
					SetKnobPart( InputSoundC, 0, 1.0 );	
					
					amplitudeExpected	= 1.0;
					break;
	}
		

	StartInstrument( theTestInstrument, NULL );
	ReadProbe( theTestProbe, &amplitudeGotten );	
	StopInstrument( theTestInstrument, NULL );

	ReadKnobPart( InputSoundA, 0, &amplitudeA );		
	ReadKnobPart( InputSoundB, 0, &amplitudeB );	
	

	if ( p[1] > 23 )
	{
		ReadKnobPart( InputSoundC, 0, &amplitudeC );		

		BeginTest(FT, "%s", ARGUMENT_01);
		
		if ( (amplitudeGotten - amplitudeExpected < .01) && (amplitudeGotten - amplitudeExpected > -.01) )
			LogResult(PASSED, "");
		else
			LogResult(FAILED, "Expected %f, but got %f from %s with (%f, %f, %f) values", 
							amplitudeExpected, amplitudeGotten,ARGUMENT_01, amplitudeA, amplitudeB, amplitudeC );
	}
	else
	{
		BeginTest(FT, "%s", ARGUMENT_01);
		
		if ( (amplitudeGotten - amplitudeExpected < .01) && (amplitudeGotten - amplitudeExpected > -.01) )
			LogResult(PASSED, "");
		else
			LogResult(FAILED, "Expected %f, but got %f from %s with (%f, %f) values", 
							amplitudeExpected, amplitudeGotten,ARGUMENT_01, amplitudeA, amplitudeB );
	}

	DeleteKnob( InputSoundA );
	DeleteKnob( InputSoundB );
	if ( p[1] > 23 )
		DeleteKnob( InputSoundC );
		
	DeleteProbe(theTestProbe);
	UnloadInstrument(theTestInstrument);

	#include "DepthTest_BottomPortion.h"


/*	Test the cubic amplifier	*/
	theTestInstrument	= LoadInstrument("cubic_amplifier.dsp", 1, 100);

	theAmplitude		= CreateKnob(theTestInstrument, "Amplitude", NULL);

	theTestProbe		= CreateProbe( theTestInstrument, "Output", NULL );
	

	/*	Create an instrument patch which supplies a constant voltage. Set the voltage to .1	*/
	constantVoltagePatch	= LoadPatchTemplate( "ampout.patch" );
	theInputVoltage			= CreateInstrument (constantVoltagePatch, NULL);
		
	if ( theInputVoltage < 0 )
	{
		LogComment("Could not create an instrument for  ampout.patch" );
		PrintfSysErr(theInputVoltage);
	}
		
	theVoltageKnob			= CreateKnob( theInputVoltage, "Amp", NULL );

	/*	Connect the constant voltage instrument to the cubic amplifier		*/		
	ConnectInstrumentParts(	theInputVoltage, "Output", 0, theTestInstrument, "Input", 0 	);

	/*	Test for positive inputs	*/
	for ( i = 0; i < 10; i++ )
	{
		SetKnobPart( theVoltageKnob, 0, .1*(i + 1) );

		StartInstrument( theTestInstrument, NULL );
		StartInstrument( theInputVoltage, NULL );
		
		ReadProbe( theTestProbe, &amplitudeGotten );
		
		StopInstrument( theTestInstrument, NULL );
		StopInstrument( theInputVoltage, NULL );

		amplitudeExpected	= powf( .1*(i + 1) - 1.0, 3.0) + 1.0;
		if ( amplitudeExpected > 1.0 )							/*	Account for clipping	*/
			amplitudeExpected	= 1.0;
			
		BeginTest(FT, "cubic_amplifier.dsp");
		
		if ( (amplitudeGotten - amplitudeExpected < .05) && (amplitudeGotten - amplitudeExpected > -.05) )
			LogResult(PASSED, "");
		else
			LogResult(FAILED, "Expected %f, but got %f from cubic amplifier with Input %f and Amplitude 1.0", 
								amplitudeExpected, amplitudeGotten,ARGUMENT_01, .1*(i + 1) );
	}
	
	/*	Test for negative inputs	*/
	for ( i = 0; i < 10; i++ )
	{
		SetKnobPart( theVoltageKnob, 0, -.1*(i + 1) );

		StartInstrument( theTestInstrument, NULL );
		StartInstrument( theInputVoltage, NULL );
		
		ReadProbe( theTestProbe, &amplitudeGotten );
		
		StopInstrument( theTestInstrument, NULL );
		StopInstrument( theInputVoltage, NULL );

		amplitudeExpected	= powf( -.1*(i + 1) + 1.0, 3.0) - 1.0;
		if ( amplitudeExpected < -1.0 )							/*	Account for clipping	*/
			amplitudeExpected	= -1.0;
			
		BeginTest(FT, "cubic_amplifier.dsp");
		
		if ( (amplitudeGotten - amplitudeExpected < .05) && (amplitudeGotten - amplitudeExpected > -.05) )
			LogResult(PASSED, "");
		else
			LogResult(FAILED, "Expected %f, but got %f from cubic amplifier with Input %f and Amplitude 1.0", 
								amplitudeExpected, amplitudeGotten,ARGUMENT_01, -.1*(i + 1) );
	}

	/*	Test for positive inputs with amplitude set to .5	*/
	for ( i = 0; i < 10; i++ )
	{
		SetKnobPart( theVoltageKnob, 0, .1*(i + 1) );
		SetKnobPart( theAmplitude, 0, .5 );

		StartInstrument( theTestInstrument, NULL );
		StartInstrument( theInputVoltage, NULL );
		
		ReadProbe( theTestProbe, &amplitudeGotten );
		
		StopInstrument( theTestInstrument, NULL );
		StopInstrument( theInputVoltage, NULL );

		amplitudeExpected	= powf( .1*(i + 1) - 1.0, 3.0) + 1.0;
		if ( amplitudeExpected > 1.0 )							/*	Account for clipping	*/
			amplitudeExpected	= 1.0;
		amplitudeExpected	= amplitudeExpected * .5;
		
		BeginTest(FT, "cubic_amplifier.dsp");
		
		if ( (amplitudeGotten - amplitudeExpected < .05) && (amplitudeGotten - amplitudeExpected > -.05) )
			LogResult(PASSED, "");
		else
			LogResult(FAILED, "Expected %f, but got %f from cubic amplifier with Input %f and Amplitude .5", 
								amplitudeExpected, amplitudeGotten,ARGUMENT_01, .1*(i + 1) );
	}
	
	
	/*	Test for positive inputs with amplitude set to -.5	*/
	for ( i = 0; i < 10; i++ )
	{
		SetKnobPart( theVoltageKnob, 0, .1*(i + 1) );
		SetKnobPart( theAmplitude, 0, -.5 );

		StartInstrument( theTestInstrument, NULL );
		StartInstrument( theInputVoltage, NULL );
		
		ReadProbe( theTestProbe, &amplitudeGotten );
		
		StopInstrument( theTestInstrument, NULL );
		StopInstrument( theInputVoltage, NULL );

		amplitudeExpected	= powf( .1*(i + 1) - 1.0, 3.0) + 1.0;
		if ( amplitudeExpected > 1.0 )							/*	Account for clipping	*/
			amplitudeExpected	= 1.0;
		amplitudeExpected	= amplitudeExpected * -.5;
		
		BeginTest(FT, "cubic_amplifier.dsp");
		
		if ( (amplitudeGotten - amplitudeExpected < .05) && (amplitudeGotten - amplitudeExpected > -.05) )
			LogResult(PASSED, "");
		else
			LogResult(FAILED, "Expected %f, but got %f from cubic amplifier with Input %f and Amplitude .5", 
								amplitudeExpected, amplitudeGotten,ARGUMENT_01, .1*(i + 1) );
	}
	
	
	DeleteProbe(theTestProbe);
	UnloadInstrument(theTestInstrument);
	UnloadPatchTemplate(constantVoltagePatch);




	cleanup:
	theError	= CloseAudioFolio();
	CHECKRESULT(FT, theError,"CloseAudioFolio");


	/******************************************************************************/
	TermScaffold();
	
	return 1;
}

