/* @(#) PlaySoundGenDSPs_FT.c 96/07/29 1.3 */
/*	
 *	File:				PlaySoundGenDSPs_FT.c
 *	Author:				Eric L. Hayes
 *	Creation Date:		
 *
 *	Last Modified by:	Eric L. Hayes
 *	Modification Date:	3/15/96
 *
 *	Version:			M2 2.0
 *
 *	This file contains the functional tests for the sound dsps.  The test program gets the
 *	name of the sound instrument from PlaySoundGenDSPs_FT_data.h. The tests require the
 *	user to listen to the output and use the control pad to enter whether the test passed (Shift-A)
 *	or failed (Shift-X).
 *
 *	Before the tests, each instrument is loaded (calcRateDivider argument to LoadInstrument is 1) 
 *	and connected to the line_out instrument. There are four tests.
 *
 *	Test 1:	Play the sound at the default frequency and amplitude.  The user should hear the sound.
 *
 *	Test 2:	Play the sound at the default frequency and at a random amplitiude. If the amplitude is too low, the
 *			sound may not be heard, but the test will still pass.  Look in the debugger window to see what
 *			the amplitude is.
 *
 *	Test 3:	Play the sound at the default rate and at a negative random amplitiude.
 *
 *	Test 4:	Play the sound at the default amplitude and a random frequency. 
 *
 *	Test 5:	Play the sound at the default amplitude and a negative random frequency. 
 *
 *	Test 6:	(For pulse.dsp and pulse_lfo.dsp only) Play the sound at the default amplitude and with a random pulse width. 
 *
 *	Test 7:	(For pulse.dsp and pulse_lfo.dsp only) Play the sound at the default amplitude and with a negative random pulse width. 
 *
 *	Uses the following calls:
 *
 *	OpenAudioFolio	 		CreateCue			LoadInstrument	
 *	GetCueSignal			CreateKnob			StopInstrument		
 *	ConnectInstrumentParts	StartInstrument		SleepUntilTime
 *	SetKnobPart				UnloadInstrument
 *
 *	Dsp Instruments used:
 *
 *	noise.dsp		sawtooth.dsp		rednoise.dsp		square.dsp
 *	triangle.dsp	square_lfo.dsp		pulse.dsp			pulse_lfo.dsp
 *
 *  Copyright (C) 1996, an unpublished work by The 3DO Company.
 *  All rights reserved. This material contains confidential
 *  information that is the property of The 3DO Company. Any
 *  unauthorized duplication, disclosure or use is prohibited.
 *  
 */



#include <audio/audio.h>
#include <stdio.h>
#include <string.h>

#include "TestMacros.h"
#include "TestUtilities.h"
#include <tests/3do_scaf.h>


#define		SOUND_DURATION		240

/**************************************Test Environment Globals****************/
/******************************************************************************/

int main (int argc, char ** argv)
{
	#include "PlaySoundGenDSPs_FT_data.h"
	
	Err		theError;
	float32	amplitude, frequency, pulseWidth;
	int32	p[10], theNumber;
	Item 	theItem, theLineOut, theCue, theAmplitude, theFrequency, thePulseWidth;



	InitScaffold(AUDIO, argc, argv);					/* initializes this program */
	/******************************************************************************/

	theError = OpenAudioFolio();
	CHECKRESULT_SETUP(FT,theError,"OpenAudioFolio",cleanup);
	
	theCue	= CreateCue( NULL );
	CHECKRESULT_SETUP(FT,theCue,"CreateCue",cleanup);

	#define expected 0
	#include "DepthTest_TopPortion.h"

		/****************************************************************************************************/
		theItem		= LoadInstrument(ARGUMENT_01, 1, 100);
		theLineOut	= LoadInstrument("line_out.dsp", 1, 100);
	
		theAmplitude	= CreateKnob( theItem, "Amplitude", NULL );
	
		/****************************************************************************************************/
		ConnectInstrumentParts(	theItem, "Output", 0, 
								theLineOut, "Input", 0 );	/*	Connect the left channel	*/

		ConnectInstrumentParts(	theItem, "Output", 0, 
								theLineOut, "Input", 1 );	/*	Connect the right channel	*/

		/****************************************************************************************************/
		/****************************************************************************************************/
		/*	Play Sound with Default Values										*/
		/****************************************************************************************************/
		BeginTest(FT, "%s", ARGUMENT_01);
		LogComment("%d. Playing %s instrument at default values", p[1], ARGUMENT_01 );

		StartInstrument( theItem, NULL );					/*	Play sound	*/
		StartInstrument( theLineOut, NULL );
		
		/****************************************************************************************************/
		SleepUntilTime( theCue, GetAudioTime() + SOUND_DURATION );		/*	Let the sound play for about 2 seconds	*/

		/****************************************************************************************************/
		StopInstrument( theItem, NULL );					/*	Stop Playing the sound	*/
		StopInstrument( theLineOut, NULL );

		LogResult(InteractiveVerify("You should have heard the sound at its default values. Did Test Pass?"), "");

		/****************************************************************************************************/
		/****************************************************************************************************/
		/*	Play Sound with a Random Amplitude 																*/
		/****************************************************************************************************/
		theNumber	= GetRandomNumber( 0, 100);
		amplitude	= theNumber / 100.0;
		
		BeginTest(FT, "%s", ARGUMENT_01);
		LogComment("%d. Playing %s instrument with %f amplitude", p[1], ARGUMENT_01, amplitude );
		
		SetKnobPart( theAmplitude, 0, amplitude );

		StartInstrument( theItem, NULL );					/*	Play sound	*/
		StartInstrument( theLineOut, NULL );
		
		/****************************************************************************************************/
		
		SleepUntilTime( theCue, GetAudioTime() + SOUND_DURATION );

		/****************************************************************************************************/
		StopInstrument( theItem, NULL );					/*	Stop Playing the sound	*/
		StopInstrument( theLineOut, NULL );

		LogResult(InteractiveVerify("You should have heard the sound at the given amplitude. Did Test Pass?"), "");

		/****************************************************************************************************/
		/****************************************************************************************************/
		/*	Play Sound with a Random Negative Amplitude														*/
		/****************************************************************************************************/
		BeginTest(FT, "%s", ARGUMENT_01);
		LogComment("%d. Playing %s instrument with %f amplitude", p[1], ARGUMENT_01, -amplitude );
		
		SetKnobPart( theAmplitude, 0, -amplitude );

		StartInstrument( theItem, NULL );					/*	Play sound	*/
		StartInstrument( theLineOut, NULL );
		
		/****************************************************************************************************/
		
		SleepUntilTime( theCue, GetAudioTime() + SOUND_DURATION );

		/****************************************************************************************************/
		StopInstrument( theItem, NULL );					/*	Stop Playing the sound	*/
		StopInstrument( theLineOut, NULL );

		LogResult(InteractiveVerify("You should have heard the sound at the given amplitude. Did Test Pass?"), "");

		/****************************************************************************************************/
		/****************************************************************************************************/
		/*	Play Sound with Default Amplitude and a Random Frequency										*/
		/****************************************************************************************************/
		if ( p[1] > 0 )										/*	All the instruments after the first one has a frequency knob	*/
		{
			theFrequency	= CreateKnob( theItem, "Frequency", NULL );
			SetKnobPart( theAmplitude, 0, 1 );				/*	Set the amplitude to the maximum	*/

			if ( (p[1] < 5) || (p[1] == 6)	)				/*	The instruments between 0 and 5, and instrument 8 (pulse.dsp) have a range between -22050 and 22050 Hz	*/	
			{
				theNumber	= GetRandomNumber( 0, 220500);
				frequency	= theNumber / 10.0;
			}
			else											/*	All other instruments are low fequency instruments with a range between -86.1 and 86.1 Hertz			*/
			{
				theNumber	= GetRandomNumber( 0, 861);
				frequency	= theNumber / 10.0;
			}
				
			BeginTest(FT, "%s", ARGUMENT_01);
			LogComment("%d. Playing %s instrument with %f frequency", p[1], ARGUMENT_01, frequency );
		
			SetKnobPart( theFrequency, 0, frequency );

			StartInstrument( theItem, NULL );					/*	Play sound	*/
			StartInstrument( theLineOut, NULL );
		
			/****************************************************************************************************/
		
			SleepUntilTime( theCue, GetAudioTime() + SOUND_DURATION );

			/****************************************************************************************************/
			StopInstrument( theItem, NULL );					/*	Stop Playing the sound	*/
			StopInstrument( theLineOut, NULL );

			LogResult(InteractiveVerify("You should have heard the sound at the given frequency. Did Test Pass?"), "");

		/****************************************************************************************************/
		/****************************************************************************************************/
		/*	Play Sound with Default Amplitude and a Random Negative Frequency								*/
		/****************************************************************************************************/

			BeginTest(FT, "%s", ARGUMENT_01);
			LogComment("%d. Playing %s instrument with %f frequency", p[1], ARGUMENT_01, frequency );
		
			SetKnobPart( theFrequency, 0, -frequency );

			StartInstrument( theItem, NULL );					/*	Play sound	*/
			StartInstrument( theLineOut, NULL );
		
			/****************************************************************************************************/
		
			SleepUntilTime( theCue, GetAudioTime() + SOUND_DURATION );

			/****************************************************************************************************/
			StopInstrument( theItem, NULL );					/*	Stop Playing the sound	*/
			StopInstrument( theLineOut, NULL );

			LogResult(InteractiveVerify("You should have heard the sound at the given frequency. Did Test Pass?"), "");

	}

		/****************************************************************************************************/
		/****************************************************************************************************/
		/*	Play Sound with a Random Pulse Width															*/
		/****************************************************************************************************/
		if ( p[1] > 5 )										/*	All the instruments after the seventh one has a PulseWidth knob	*/
		{
			thePulseWidth	= CreateKnob( theItem, "PulseWidth", NULL );
			SetKnobPart( theAmplitude, 0, 1 );				/*	Set the amplitude to the maximum	*/

			theNumber	= GetRandomNumber( 0, 100);
			pulseWidth	= theNumber / 100.0;
		
			BeginTest(FT, "%s", ARGUMENT_01);
			LogComment("%d. Playing %s instrument with %f Pulse Width", p[1], ARGUMENT_01, pulseWidth );
		
			SetKnobPart( thePulseWidth, 0, pulseWidth );

			StartInstrument( theItem, NULL );					/*	Play sound	*/
			StartInstrument( theLineOut, NULL );
		
			/****************************************************************************************************/
		
			SleepUntilTime( theCue, GetAudioTime() + SOUND_DURATION );

			/****************************************************************************************************/
			StopInstrument( theItem, NULL );					/*	Stop Playing the sound	*/
			StopInstrument( theLineOut, NULL );

			LogResult(InteractiveVerify("You should have heard the sound at the given pulse width. Did Test Pass?"), "");

			/****************************************************************************************************/
			/****************************************************************************************************/
			/*	Play Sound with a Random Negative Pulse Width													*/
			/****************************************************************************************************/

			BeginTest(FT, "%s", ARGUMENT_01);
			LogComment("%d. Playing %s instrument with -%f pulse Width", p[1], ARGUMENT_01, pulseWidth );
		
			SetKnobPart( thePulseWidth, 0, -pulseWidth );

			StartInstrument( theItem, NULL );					/*	Play sound	*/
			StartInstrument( theLineOut, NULL );
		
			/****************************************************************************************************/
		
			SleepUntilTime( theCue, GetAudioTime() + SOUND_DURATION );

			/****************************************************************************************************/
			StopInstrument( theItem, NULL );					/*	Stop Playing the sound	*/
			StopInstrument( theLineOut, NULL );

			LogResult(InteractiveVerify("You should have heard the sound at the given pulse width. Did Test Pass?"), "");

		}

		/****************************************************************************************************/
		UnloadInstrument(theItem);
		UnloadInstrument(theLineOut);
		/****************************************************************************************************/

	#include "DepthTest_BottomPortion.h"
	
	
	DeleteCue( theCue );

	cleanup:
	theError	= CloseAudioFolio();
	CHECKRESULT(FT,theError,"CloseAudioFolio");


	/******************************************************************************/
	TermScaffold();
	
	return 1;
}

