/* @(#) PlaySoundGenDSPs_FT2.c 96/07/29 1.3 */
/*	
 *	File:				PlaySoundGenDSPs_FT2.c
 *	Author:				Eric L. Hayes
 *	Creation Date:		
 *
 *	Last Modified by:	Eric L. Hayes
 *	Modification Date:	3/15/96
 *
 *	Version:			M2 2.0
 *
 *	This file contains the functional tests for the sound dsps.  This test is similar to PlaySoundGenDSPs_FT.c,
 *	except that this test checks to see if the sound instrument plays after being loaded with different
 *	calcRateDivide arguments. The test program gets the name of the sound instrument from PlaySoundGenDSPs_FT2_data.h.
 *	The tests require the user to listen to the output and use the control pad to enter whether the test passed (Shift-A)
 *	or failed (Shift-X).
 *
 *	Before the tests, each instrument is loaded  and connected to the line_out instrument. After every test, read the
 *	frequency knob to see if the frequency corresponds to the calcRateDivide argument supplied to LoadInstrument.
 *	There are three tests.
 *
 *	Test 1:	Load the sound dsp instrument with calcRateDivide argument set to 1
 *
 *	Test 2:	Load the sound dsp instrument with calcRateDivide argument set to 2
 *
 *	Test 3:	Load the sound dsp instrument with calcRateDivide argument set to 8
 *
 *	Uses the following calls:
 *
 *	OpenAudioFolio	 		CreateCue			LoadInstrument	
 *	GetCueSignal			CreateKnob			StopInstrument		
 *	ConnectInstrumentParts	StartInstrument		SleepUntilTime
 *	SetKnobPart				UnloadInstrument
 *
 *	Dsp Instruments used:
 *
 *	noise.dsp		sawtooth.dsp		rednoise.dsp		square.dsp
 *	triangle.dsp	square_lfo.dsp		pulse.dsp			pulse_lfo.dsp
 *
 *	to the output instrument. The first test plays the sound with  calcRateDivide argument set to 1.
 *	The second test plays the sound with  calcRateDivide argument set to 2.
 *	The second test plays the sound with  calcRateDivide argument set to 8.
 *
 *  Copyright (C) 1996, an unpublished work by The 3DO Company.
 *  All rights reserved. This material contains confidential
 *  information that is the property of The 3DO Company. Any
 *  unauthorized duplication, disclosure or use is prohibited.
 *  
 */



#include <audio/audio.h>
#include <stdio.h>
#include <string.h>

#include "TestMacros.h"
#include "TestUtilities.h"
#include <tests/3do_scaf.h>


#define		SOUND_DURATION		240

/**************************************Test Environment Globals****************/
/******************************************************************************/

int main (int argc, char ** argv)
{
	#include "PlaySoundGenDSPs_FT2_data.h"
	
	Err		theError;
	float32	frequency, crdValue;
	int32	p[10];
	Item 	theItem, theLineOut, theCue, theFrequency;



	InitScaffold(AUDIO, argc, argv);					/* initializes this program */
	BeginTest(FT, "PlaySoundGenDSPs_FT2");
	/******************************************************************************/

	theError = OpenAudioFolio();
	CHECKRESULT_SETUP(FT,theError,"OpenAudioFolio",cleanup);
	
	theCue	= CreateCue( NULL );
	CHECKRESULT_SETUP(FT,theCue,"CreateCue",cleanup);

	#define expected 0
	#include "DepthTest_TopPortion.h"

		/****************************************************************************************************/
		/*	Test 1:	Set calcRateDivide Argument to 1														*/
		/****************************************************************************************************/
		crdValue		= 1.0;
		theItem		= LoadInstrument(ARGUMENT_01, 1, 100);
		theLineOut	= LoadInstrument("line_out.dsp", 1, 100);
		
		/****************************************************************************************************/
		ConnectInstrumentParts(	theItem, "Output", 0, 
								theLineOut, "Input", 0 );	/*	Connect the left channel	*/

		ConnectInstrumentParts(	theItem, "Output", 0, 
								theLineOut, "Input", 1 );	/*	Connect the right channel	*/

		/****************************************************************************************************/
		/****************************************************************************************************/
		/*	Play Sound with Default Values										*/
		/****************************************************************************************************/
		BeginTest(FT, "%s", ARGUMENT_01);
		LogComment("%d. Playing %s instrument (calcRateDivide = 1 ) at default values", p[1], ARGUMENT_01 );

		StartInstrument( theItem, NULL );					/*	Play sound	*/
		StartInstrument( theLineOut, NULL );
		
		/****************************************************************************************************/
		SleepUntilTime( theCue, GetAudioTime() + SOUND_DURATION );		/*	Let the sound play for about 2 seconds	*/

		/****************************************************************************************************/
		StopInstrument( theItem, NULL );					/*	Stop Playing the sound	*/
		StopInstrument( theLineOut, NULL );

		/****************************************************************************************************/
		/*	Determine the Frequency Knob's value.  Set and read new frequency knob value					*/
		/****************************************************************************************************/
		switch ( p[1] )										/*	All the instruments after the first one has a frequency knob	*/
		{
			case	0	:
							break;
			case	1	:
			case	2	:
			case	3	:
			case	4	:
			case	5	:
							theFrequency	= CreateKnob( theItem, "Frequency", NULL );
							ReadKnob( theFrequency, &frequency );

							BeginTest(FT, "%s", ARGUMENT_01);
							if ( (frequency - 440.0/crdValue < 1) && (frequency - 440.0/crdValue > -1) )
								LogResult(PASSED, "" );
							else
								LogResult(FAILED, "Expected %f, but got %f from %s", 440.0/crdValue, frequency, ARGUMENT_01 );
		
							SetKnobPart( theFrequency, 0, 1000.0 );
							ReadKnob( theFrequency, &frequency );

							BeginTest(FT, "%s", ARGUMENT_01);
							if ( (frequency - 1000.0 < 1) && (frequency - 1000.0 > -1) )
								LogResult(PASSED, "");
							else
								LogResult(FAILED, "Expected 1000.0 from ReadKnob after Set Knob to 1000.0, but got %f from %s", frequency, ARGUMENT_01 );
							
							break;
			case	6	:
			case	7	:
							theFrequency	= CreateKnob( theItem, "Frequency", NULL );
							ReadKnob( theFrequency, &frequency );

							BeginTest(FT, "%s", ARGUMENT_01);
							if ( (frequency - 17.2/crdValue < 1) && (frequency - 17.2/crdValue > -1) )
								LogResult(PASSED, "");
							else
								LogResult(FAILED, "Expected %f, but got %f from %s", 17.2/crdValue, frequency, ARGUMENT_01 );
		
							SetKnobPart( theFrequency, 0, 30.0 );
							ReadKnob( theFrequency, &frequency );

							BeginTest(FT, "%s", ARGUMENT_01);
							if ( (frequency - 30.0 < 1) && (frequency - 30.0 > -1) )
								LogResult(PASSED, "");
							else
								LogResult(FAILED, "Expected 30.0 from ReadKnob after Set Knob to 30.0, but got %f from %s", frequency, ARGUMENT_01 );
							
							break;
		
		}

		BeginTest(FT, "%s", ARGUMENT_01);
		LogResult(InteractiveVerify("You should have heard the sound. Did Test Pass?"), "");

		UnloadInstrument(theItem);

		/****************************************************************************************************/
		/****************************************************************************************************/
		/*	Test 2:	Set calcRateDivide Argument to 2														*/
		/****************************************************************************************************/
		crdValue		= 2.0;
		theItem		= LoadInstrument(ARGUMENT_01, 2, 100);
	
		/****************************************************************************************************/
		ConnectInstrumentParts(	theItem, "Output", 0, 
								theLineOut, "Input", 0 );	/*	Connect the left channel	*/

		ConnectInstrumentParts(	theItem, "Output", 0, 
								theLineOut, "Input", 1 );	/*	Connect the right channel	*/

		/****************************************************************************************************/
		/****************************************************************************************************/
		/*	Play Sound with Default Values										*/
		/****************************************************************************************************/
		BeginTest(FT, "%s", ARGUMENT_01);
		LogComment("%d. Playing %s instrument (calcRateDivide = 2) at default values", p[1], ARGUMENT_01 );

		StartInstrument( theItem, NULL );					/*	Play sound	*/
		StartInstrument( theLineOut, NULL );
		
		/****************************************************************************************************/
		SleepUntilTime( theCue, GetAudioTime() + SOUND_DURATION );		/*	Let the sound play for about 2 seconds	*/

		/****************************************************************************************************/
		StopInstrument( theItem, NULL );					/*	Stop Playing the sound	*/
		StopInstrument( theLineOut, NULL );

		/****************************************************************************************************/
		/*	Determine the Frequency Knob's value.  Set and read new frequency knob value					*/
		/****************************************************************************************************/
		switch ( p[1] )										/*	All the instruments after the first one has a frequency knob	*/
		{
			case	0	:
							break;
			case	1	:
			case	2	:
			case	3	:
			case	4	:
			case	5	:
							theFrequency	= CreateKnob( theItem, "Frequency", NULL );
							ReadKnob( theFrequency, &frequency );

							BeginTest(FT, "%s", ARGUMENT_01);
							if ( (frequency - 440.0/crdValue < 1) && (frequency - 440.0/crdValue > -1) )
								LogResult(PASSED, "");
							else
								LogResult(FAILED, "Expected %f, but got %f from %s", 440.0/crdValue, frequency, ARGUMENT_01 );
		
							SetKnobPart( theFrequency, 0, 1000.0 );
							ReadKnob( theFrequency, &frequency );

							BeginTest(FT, "%s", ARGUMENT_01);
							if ( (frequency - 1000.0 < 1) && (frequency - 1000.0 > -1) )
								LogResult(PASSED, "");
							else
								LogResult(FAILED, "Expected 1000.0 from ReadKnob after Set Knob to 1000.0, but got %f from %s", frequency, ARGUMENT_01 );
							
							break;
			case	6	:
			case	7	:
							theFrequency	= CreateKnob( theItem, "Frequency", NULL );
							ReadKnob( theFrequency, &frequency );

							BeginTest(FT, "%s", ARGUMENT_01);
							if ( (frequency - 17.2/crdValue < 1) && (frequency - 17.2/crdValue > -1) )
								LogResult(PASSED, "");
							else
								LogResult(FAILED, "Expected %f, but got %f from %s", 17.2/crdValue, frequency, ARGUMENT_01 );
		
							SetKnobPart( theFrequency, 0, 30.0 );
							ReadKnob( theFrequency, &frequency );

							BeginTest(FT, "%s", ARGUMENT_01);
							if ( (frequency - 30.0 < 1) && (frequency - 30.0 > -1) )
								LogResult(PASSED, "");
							else
								LogResult(FAILED, "Expected 30.0 from ReadKnob after Set Knob to 30.0, but got %f from %s", frequency, ARGUMENT_01 );
							
							break;
		
		}

		BeginTest(FT, "%s", ARGUMENT_01);
		LogResult(InteractiveVerify("You should have heard the sound at a lower pitch. Did Test Pass?"), "");

		UnloadInstrument(theItem);

		/****************************************************************************************************/
		/****************************************************************************************************/
		/*	Test 3:	Set calcRateDivide Argument to 8														*/
		/****************************************************************************************************/
		crdValue		= 8.0;
		theItem		= LoadInstrument(ARGUMENT_01, 8, 100);
	
		/****************************************************************************************************/
		ConnectInstrumentParts(	theItem, "Output", 0, 
								theLineOut, "Input", 0 );	/*	Connect the left channel	*/

		ConnectInstrumentParts(	theItem, "Output", 0, 
								theLineOut, "Input", 1 );	/*	Connect the right channel	*/

		/****************************************************************************************************/
		/****************************************************************************************************/
		/*	Play Sound with Default Values										*/
		/****************************************************************************************************/
		BeginTest(FT, "%s", ARGUMENT_01);
		LogComment("%d. Playing %s instrument (calcRateDivide = 8) at default values", p[1], ARGUMENT_01 );

		StartInstrument( theItem, NULL );					/*	Play sound	*/
		StartInstrument( theLineOut, NULL );
		
		/****************************************************************************************************/
		SleepUntilTime( theCue, GetAudioTime() + SOUND_DURATION );		/*	Let the sound play for about 2 seconds	*/

		/****************************************************************************************************/
		StopInstrument( theItem, NULL );					/*	Stop Playing the sound	*/
		StopInstrument( theLineOut, NULL );

		/****************************************************************************************************/
		/*	Determine the Frequency Knob's value.  Set and read new frequency knob value					*/
		/****************************************************************************************************/
		switch ( p[1] )										/*	All the instruments after the first one has a frequency knob	*/
		{
			case	0	:
							break;
			case	1	:
			case	2	:
			case	3	:
			case	4	:
			case	5	:
							theFrequency	= CreateKnob( theItem, "Frequency", NULL );
							ReadKnob( theFrequency, &frequency );

							BeginTest(FT, "%s", ARGUMENT_01);
							if ( (frequency - 440.0/crdValue < 1) && (frequency - 440.0/crdValue > -1) )
								LogResult(PASSED, "");
							else
								LogResult(FAILED, "Expected %f, but got %f from %s", 440.0/crdValue, frequency, ARGUMENT_01 );
		
							SetKnobPart( theFrequency, 0, 1000.0 );
							ReadKnob( theFrequency, &frequency );

							BeginTest(FT, "%s", ARGUMENT_01);
							if ( (frequency - 1000.0 < 1) && (frequency - 1000.0 > -1) )
								LogResult(PASSED, "");
							else
								LogResult(FAILED, "Expected 1000.0 from ReadKnob after Set Knob to 1000.0, but got %f from %s", frequency, ARGUMENT_01 );
							
							break;
			case	6	:
			case	7	:
							theFrequency	= CreateKnob( theItem, "Frequency", NULL );
							ReadKnob( theFrequency, &frequency );

							BeginTest(FT, "%s", ARGUMENT_01);
							if ( (frequency - 17.2/crdValue < 1) && (frequency - 17.2/crdValue > -1) )
								LogResult(PASSED, "");
							else
								LogResult(FAILED, "Expected %f, but got %f from %s", 17.2/crdValue, frequency, ARGUMENT_01 );
		
							SetKnobPart( theFrequency, 0, 10.0 );
							ReadKnob( theFrequency, &frequency );

							BeginTest(FT, "%s", ARGUMENT_01);
							if ( (frequency - 10.0 < 1) && (frequency - 10.0 > -1) )
								LogResult(PASSED, "");
							else
								LogResult(FAILED, "Expected 10.0 from ReadKnob after Set Knob to 10.0, but got %f from %s", frequency, ARGUMENT_01 );
							
							break;
		
		}

		BeginTest(FT, "%s", ARGUMENT_01);
		LogResult(InteractiveVerify("You should have heard the sound at a lower pitch. Did Test Pass?"), "");


		UnloadInstrument(theItem);


		/****************************************************************************************************/
		UnloadInstrument(theLineOut);
		/****************************************************************************************************/

	#include "DepthTest_BottomPortion.h"
	
	
	DeleteCue( theCue );

	cleanup:
	theError	= CloseAudioFolio();
	CHECKRESULT(FT,theError,"CloseAudioFolio");

	/******************************************************************************/
	TermScaffold();
	
	return 1;
}

