/* @(#) Sampler_Drift_FT.c 96/07/29 1.3 */
/*	
 *	File:				Sampler_Drift_FT.c
 *	Author:				Eric L. Hayes
 *	Creation Date:		
 *
 *	Last Modified by:	Eric L. Hayes
 *	Modification Date:	3/15/96
 *
 *	Version:			M2 2.0
 *
 *	This file contains the functional tests for the sampler_drift_v1.dsp.  This test program does two
 *	thimgs: 1) It allows the user to hear that it cam play audio files. 2) Checks to see if the value
 *	of the drift output is in the appropriate range.  This check is done at two different points while
 *	the sample is playing. There are three tests.
 *
 *	Test 1:	Play the sample file at its normal sample rate. The sample should sound normal and the
 *			drift should be 0.
 *
 *	Test 2:	Play the sample at twice its sample rate.  The sample should sound fast and the drift 
 *			should be greater than 0.
 *
 *	Test 3:	Play the sample at half of its sample rate.  The sample should sound slow and the drift
 *			should be less than 0.
 *
 *	Uses the following calls:
 *
 *	OpenAudioFolio	 		CreateCue			LoadInstrument			GetCueSignal
 *	CreateProbeVA			CreateKnob			StopInstrument			CreateAttachment	
 *	ReadProbe				StartInstrument		SleepUntilTime			ConnectInstrumentParts
 *	SetKnob					UnloadInstrument	UnloadSample			WaitSignal
 *	DeleteAttachment		DeleteCue			DeleteProbe
 *
 *  Copyright (C) 1996, an unpublished work by The 3DO Company.
 *  All rights reserved. This material contains confidential
 *  information that is the property of The 3DO Company. Any
 *  unauthorized duplication, disclosure or use is prohibited.
 *  
 */



#include <audio/audio.h>
#include <audio/parse_aiff.h>
#include <stdio.h>

#include "TestMacros.h"
#include "TestUtilities.h"
#include <tests/3do_scaf.h>

/**************************************Test Environment Globals****************/
/******************************************************************************/

int main (int argc, char ** argv)
{
	
	Err		theError;
	float32	rate, driftGotten, driftGotten2;
	int32	theSignal, i, counter = 1;
	Item 	theItem, theLineOut, theSample, theAttachment,
			theCue, theAmplitude, theRate, theTestProbe;



	InitScaffold(AUDIO, argc, argv);					/* initializes this program */
	/******************************************************************************/

	theError = OpenAudioFolio();
	CHECKRESULT_SETUP(FT,theError,"OpenAudioFolio",cleanup);
	
	theCue	= CreateCue( NULL );
	CHECKRESULT_SETUP(FT,theCue,"CreateCue",cleanup);

	#define expected 0

	for ( i = 0; i < 1; i++ )
	{
		/****************************************************************************************************/
		/*	Load the sample instrument and the line_out instrument (which provides the sound).				*/
		/*	Load the sample and attach it to the sample instrument.  Call MonitorAttachment which will		*/
		/*	let you know through theCue when the end of the sample has been reached when the instrument is	*/
		/*	played.  theSignal is theSignal for that particular attachment.									*/
		
		theItem			= LoadInstrument("sampler_drift_v1.dsp", 1, 100);
		CHECKRESULT_SETUP(FT,theItem,"LoadInstrument(sampler_drift_v1.dsp)",cleanup);
		
		theLineOut		= LoadInstrument("line_out.dsp", 1, 100);
		CHECKRESULT_SETUP(FT,theLineOut,"LoadInstrument(line_out.dsp)",cleanup);

		theSample		= LoadSample("44K_16B_M.AIFF");
		CHECKRESULT_SETUP(FT,theSample,"LoadSample(44K_16B_M.AIFF)",cleanup);

		theAttachment	= CreateAttachment( theItem, theSample, NULL );
		CHECKRESULT_SETUP(FT,theAttachment,"CreateAttachment",cleanup);

		MonitorAttachment( theAttachment, theCue, CUE_AT_END );
		theSignal		= GetCueSignal( theCue );
		
		theAmplitude	= CreateKnob( theItem, "Amplitude", NULL );
		CHECKRESULT_SETUP(FT,theAmplitude,"CreateKnob(Amplitude)",cleanup);

		theTestProbe	= CreateProbeVA( theItem, "Drift", AF_TAG_TYPE, AF_SIGNAL_TYPE_WHOLE_NUMBER, TAG_END );
		CHECKRESULT_SETUP(FT,theTestProbe,"CreateProbeVA(Drift)",cleanup);
	
		/****************************************************************************************************/
		/*	Monophonic instrument are listed before the 10th element in the structure						*/
		/*	Stereo instruments are found after the 10th element												*/
		
		theError	= ConnectInstrumentParts(	theItem, "Output", 0, 
												theLineOut, "Input", 0 );		/*	Connect the left channel	*/
		CHECKRESULT_SETUP(FT,theError,"ConnectInstrumentParts",cleanup);

		theError	= ConnectInstrumentParts(	theItem, "Output", 0, 
												theLineOut, "Input", 1 );		/*	Connect the right channel - Mono Sample	*/
		CHECKRESULT_SETUP(FT,theError,"ConnectInstrumentParts",cleanup);
		/****************************************************************************************************/
		/****************************************************************************************************/
		/*	Play Sample with Default Amplitude and Default Sample Rate										*/
		/****************************************************************************************************/
		LogComment("%d. Playing sampler_drift_v1 instrument at default amplitude and sample rate", counter++ );

		StartInstrument( theItem, NULL );					/*	Play sound	*/
		StartInstrument( theLineOut, NULL );
		
		/****************************************************************************************************/
		
		for ( i = 0; i < 44100; i++ )
		{
			if ( i == 22050 )
				ReadProbe( theTestProbe, &driftGotten );

			if ( i == 44099 )
				ReadProbe( theTestProbe, &driftGotten2 );
		}
		

		WaitSignal( theSignal );							/*	Enter a Wait state until sample is finished	*/

		/****************************************************************************************************/

		StopInstrument( theItem, NULL );					/*	Stop Playing the sound	*/
		StopInstrument( theLineOut, NULL );


		BeginTest(FT, "sampler_drift_v1.dsp");
		if ( (int32)driftGotten  == 0 )
			LogResult(PASSED, "");
		else
			LogResult(FAILED, "Expected 0, but got %d from sampler_drift_v1 after playing a 44KHz sample at 44KHz", (int32)driftGotten );


		BeginTest(FT, "sampler_drift_v1.dsp");
		if ( (int32)driftGotten2  == 0 )
			LogResult(PASSED, "" );
		else
			LogResult(FAILED, "Expected 0, but got %d after second reading of sampler_drift_v1 drift output", (int32)driftGotten2 );


		/****************************************************************************************************/
		/****************************************************************************************************/
		/*	Play Sample with Default Amplitude and Double the Sample Rate									*/
		/****************************************************************************************************/
		theRate		= CreateKnob( theItem, "SampleRate", NULL );
		SetKnobPart( theAmplitude, 0, 1 );					/*	Set the amplitude to the maximum	*/
			
		
		rate	= 88200.0;						/*	Normally, the rate is random.  Take this line out later	*/		
		LogComment("%d. Playing 44Khz sample on a sampler_drift_v1 instrument with %f sample rate", counter++, rate );

		SetKnobPart( theRate, 0, rate );			

		StartInstrument( theItem, NULL );		/*	Play sound	*/
		StartInstrument( theLineOut, NULL );
		
		/****************************************************************************************************/
		
		for ( i = 0; i < 88200; i++ )
		{
			if ( i == 44100 )
				ReadProbe( theTestProbe, &driftGotten );

			if ( i == 88199 )
				ReadProbe( theTestProbe, &driftGotten2 );
		}
		

		WaitSignal( theSignal );							/*	Enter a Wait state until sample is finished	*/

		/****************************************************************************************************/

		StopInstrument( theItem, NULL );					/*	Stop Playing the sound	*/
		StopInstrument( theLineOut, NULL );
		
		BeginTest(FT, "sampler_drift_v1.dsp");
		if ( (int32)driftGotten  > 0 )
			LogResult(PASSED, "" );
		else
			LogResult(FAILED, "Got %d from sampler_drift_v1 after playing a 44KHz sample at %f KHz", (int32)driftGotten, rate );


		BeginTest(FT, "sampler_drift_v1.dsp");
		if ( (int32)driftGotten2  > (int32)driftGotten )
			LogResult(PASSED, "" );
		else
			LogResult(FAILED, "Got %d from second reading of sampler_drift_v1 drift output after playing a  44 KHz sample at %f KHz", (int32)driftGotten2, rate);


		/****************************************************************************************************/
		/****************************************************************************************************/
		/*	Play Sample with Default Amplitude and Half the Sample Rate									*/
		/****************************************************************************************************/

		rate	= 22050.0;						/*	Normally, the rate is random.  Take this line out later	*/		
		LogComment("%d. Playing 44Khz sample on a sampler_drift_v1 instrument with %f sample rate", counter++, rate );

		SetKnobPart( theRate, 0, rate );			

		StartInstrument( theItem, NULL );		/*	Play sound	*/
		StartInstrument( theLineOut, NULL );
		
		/****************************************************************************************************/
		
		for ( i = 0; i < 22050; i++ )
		{
			if ( i == 11025 )
				ReadProbe( theTestProbe, &driftGotten );

			if ( i == 22049 )
				ReadProbe( theTestProbe, &driftGotten2 );
		}
		

		WaitSignal( theSignal );							/*	Enter a Wait state until sample is finished	*/

		/****************************************************************************************************/

		StopInstrument( theItem, NULL );					/*	Stop Playing the sound	*/
		StopInstrument( theLineOut, NULL );
		
		BeginTest(FT, "sampler_drift_v1.dsp");
		if ( (int32)driftGotten  < 0 )
			LogResult(PASSED, "");
		else
			LogResult(FAILED, "Got %d from sampler_drift_v1 after playing a 44KHz sample at %f KHz", (int32)driftGotten, rate );


		BeginTest(FT, "sampler_drift_v1.dsp");
		if ( (int32)driftGotten2  < (int32)driftGotten )
			LogResult(PASSED, "");
		else
			LogResult(FAILED, "Got %d from second reading of sampler_drift_v1 drift output after playing a  44 KHz sample at %f KHz", (int32)driftGotten2, rate );


		/****************************************************************************************************/

		DeleteProbe(theTestProbe);
		DeleteAttachment( theAttachment );
		UnloadSample( theSample );
		UnloadInstrument(theItem);
		UnloadInstrument(theLineOut);
		/****************************************************************************************************/
	}	
	
	DeleteCue( theCue );

	cleanup:
	theError	= CloseAudioFolio();
	CHECKRESULT(FT, theError,"CloseAudioFolio");


	/******************************************************************************/
	TermScaffold();
	
	return 1;
}

