/* @(#) Schmidt_Trigger_FT.c 96/07/22 1.2 */
/*	
 *	File:				Schmidt_Trigger_FT.c
 *	Author:				Eric L. Hayes
 *	Creation Date:		
 *
 *	Last Modified by:	Eric L. Hayes
 *	Modification Date:	3/15/96
 *
 *	Version:			M2 2.0
 *
 *	This file contains the functional tests for the schmidt_trigger.dsp.  A structure, TestValues
 *	was created to hold the test values for the instrument's knobs and input.  The strucue also
 *	contains the expected output for those given inputs.  The last element in the structure ia
 *	boolean element named startStop.  When startStop is TRUE, the test program will Stop and then
 *	restart the instrument before reading the output.  THis option allows the test program to test
 *	whether the instrument is properly resetting itself.
 *
 *	The test program goes through a loop in which the test values from the structure are given to
 *	the test instrument.  The test program outputs the results to the screen.
 *
 *	The folowing patches need to be in the current working directory in order for this test to run.
 *
 *		ampout.patch
 *
 *	Uses the following calls:
 *
 *	OpenAudioFolio	 		CreateCue			LoadInstrument			LoadPatchTemplate
 *	CreateProbeVA			CreateKnob			StopInstrument			CreateInstrument	
 *	ReadProbe				StartInstrument		SleepUntilTime			GetAudioTime
 *	SetKnob					UnloadInstrument	ConnectInstrumentParts
 *
 *  Copyright (C) 1996, an unpublished work by The 3DO Company.
 *  All rights reserved. This material contains confidential
 *  information that is the property of The 3DO Company. Any
 *  unauthorized duplication, disclosure or use is prohibited.
 *  
 */



#include <audio/audio.h>
#include <audio/patch.h>
#include <audio/patchfile.h>
#include <stdio.h>

#include "TestMacros.h"
#include "TestUtilities.h"
#include <tests/3do_scaf.h>

/**************************************Test Environment Globals****************/
/******************************************************************************/

typedef	struct
		{
			float32	input;
			float32	setLevel;
			float32	resetLevel;
			Boolean	expectedOutput;
			Boolean	startStop;
		}  TestValues;


#define	NUM_OF_TESTS		15


int main (int argc, char ** argv)
{
	Boolean		oscillate	= FALSE;
	Err			theError;
	float32		theOutputValue, theOutputValue2;
	int			i, j;
	Item 		theItem, theTestProbe, theInputVoltage, constantVoltagePatch, theVoltageKnob,
				sLKnob, rLKnob, theCue;
	TestValues	testList[NUM_OF_TESTS]	= 	{
												0,		0,		0,		FALSE,	TRUE,
												.5,		0,		0,		TRUE,	TRUE,
												-.5,	0,		0,		FALSE,	TRUE,
												.45,	.6,		.3,		FALSE,	TRUE,
												.8,		.6,		.3,		TRUE,	TRUE,
												.45,	.6,		.3,		FALSE,	TRUE,
												
												0,		0,		0,		FALSE,	FALSE,
												.5,		0,		0,		TRUE,	FALSE,
												-.5,	0,		0,		FALSE,	FALSE,
												.45,	.6,		.3,		FALSE,	FALSE,
												.8,		.6,		.3,		TRUE,	FALSE,
												0,		.6,		.3,		FALSE,	FALSE,
												0,		-.4,	-.7,	TRUE,	FALSE,
												-.9,	-.4,	-.7,	FALSE,	FALSE,
												.7,		.4,		.7,		FALSE,	TRUE
												
											};



	InitScaffold(AUDIO, argc, argv);					/* initializes this program */
	/******************************************************************************/

	theError = OpenAudioFolio();
	CHECKRESULT_SETUP(FT,theError,"OpenAudioFolio",cleanup);
	
	theCue	= CreateCue( NULL );
	CHECKRESULT_SETUP(FT,theCue,"CreateCue",cleanup);


	/****************************************************************************************************
	 *	Load the schmidt trigger instrument and the constant voltage patch which will be used as the
	 *	schmidt trigger's input.  Create the knobs for the schmidt trigger and create the knob for the
	 *	constant voltage patch.
	 ****************************************************************************************************/
		
	theItem			= LoadInstrument("schmidt_trigger.dsp", 1, 100);
	CHECKRESULT_SETUP(FT,theItem,"LoadInstrument(schmidt_trigger.dsp)",cleanup);
		
	sLKnob	= CreateKnob( theItem, "SetLevel", NULL );
	CHECKRESULT_SETUP(FT,sLKnob,"CreateKnob(SetLevel)",cleanup);

	rLKnob	= CreateKnob( theItem, "ResetLevel", NULL );
	CHECKRESULT_SETUP(FT,rLKnob,"CreateKnob(ResetLevel)",cleanup);

	theTestProbe	= CreateProbeVA( theItem, "Output", AF_TAG_TYPE, AF_SIGNAL_TYPE_WHOLE_NUMBER, TAG_END );
	CHECKRESULT_SETUP(FT,theTestProbe,"CreateProbeVA",cleanup);
	
	/*	Create an instrument patch which supplies a constant voltage. Set the voltage to .1	*/
	constantVoltagePatch	= LoadPatchTemplate( "ampout.patch" );
	theInputVoltage			= CreateInstrument (constantVoltagePatch, NULL);
		
	if ( theInputVoltage < 0 )
	{
		LogComment("Could not create an instrument for  ampout.patch" );
		PrintfSysErr(theInputVoltage);
	}
		
	theVoltageKnob			= CreateKnob( theInputVoltage, "Amp", NULL );
	CHECKRESULT_SETUP(FT,theVoltageKnob,"CreateKnob(Amp)",cleanup);

	/*	Connect the constant voltage instrument to the schmidt trigger		*/		
	theError	= ConnectInstrumentParts(	theInputVoltage, "Output", 0, theItem, "Input", 0 	);
	CHECKRESULT_SETUP(FT,theError,"ConnectInstrumentParts",cleanup);



	StartInstrument( theInputVoltage, NULL );
	StartInstrument( theItem, NULL );

	for ( i = 0; i < NUM_OF_TESTS; i++ )
	{
		/*	Set the values for the input signal and the reset and set knobs	*/
		SetKnob( theVoltageKnob, testList[i].input );
		SetKnob( sLKnob, testList[i].setLevel );
		SetKnob( rLKnob, testList[i].resetLevel );

		SleepUntilTime( theCue, GetAudioTime() + 2 );
		ReadProbe( theTestProbe, &theOutputValue );
	
		if ( testList[i].startStop == TRUE )
		{
			StopInstrument( theItem, NULL );
			StartInstrument( theItem, NULL );
			SleepUntilTime( theCue, GetAudioTime() + 2 );
			ReadProbe( theTestProbe, &theOutputValue );
		}

		
		if ( testList[i].resetLevel > testList[i].setLevel )
		{
			for ( j = 0; j < 2; j++ )
			{
				ReadProbe( theTestProbe, &theOutputValue );
				SleepUntilTime( theCue, GetAudioTime() + 1 );
				ReadProbe( theTestProbe, &theOutputValue2 );
				
				if ( theOutputValue != theOutputValue2 )
					oscillate	= TRUE;

				ReadProbe( theTestProbe, &theOutputValue );
				SleepUntilTime( theCue, GetAudioTime() + 2 );
				ReadProbe( theTestProbe, &theOutputValue2 );
				
				if ( theOutputValue != theOutputValue2 )
					oscillate	= TRUE;

				ReadProbe( theTestProbe, &theOutputValue );
				SleepUntilTime( theCue, GetAudioTime() + 3 );
				ReadProbe( theTestProbe, &theOutputValue2 );
				
				if ( theOutputValue != theOutputValue2 )
					oscillate	= TRUE;
				
			}
			
			BeginTest(FT, "schmidt_trigger.dsp");
			if ( oscillate  == TRUE )
				LogResult(PASSED, "");
			else
				LogResult(FAILED, "%2d. Output is not oscillating when Input = %f;  SetLevel = %f; ResetLevel = %f.",
								i, testList[i].input, testList[i].setLevel, testList[i].resetLevel );
			
			oscillate	= FALSE;
		}
		else if ( testList[i].expectedOutput  == TRUE )
		{
			BeginTest(FT, "schmidt_trigger.dsp");
			if ( (Boolean)theOutputValue  > 0 )
				LogResult(PASSED, "");
			else
				LogResult(FAILED, "%2d. Expected %d, but got %d when Input = %f;  SetLevel = %f; ResetLevel = %f.",
								i, testList[i].expectedOutput, (Boolean)theOutputValue, testList[i].input, testList[i].setLevel, testList[i].resetLevel );
		}
		else
		{
			BeginTest(FT, "schmidt_trigger.dsp");
			if ( (Boolean)theOutputValue  == 0 )
				LogResult(PASSED, "");
			else
				LogResult(FAILED, "%2d. Expected %d, but got %d when Input = %f;  SetLevel = %f; ResetLevel = %f.",
								i, testList[i].expectedOutput, (Boolean)theOutputValue, testList[i].input, testList[i].setLevel, testList[i].resetLevel );
		}
	}


	StopInstrument( theInputVoltage, NULL );
	StopInstrument( theItem, NULL );

		
	
	DeleteCue( theCue );
	UnloadInstrument( theItem );
	UnloadPatchTemplate( constantVoltagePatch );
	
	cleanup:
	theError	= CloseAudioFolio();
	CHECKRESULT(FT, theError,"CloseAudioFolio");


	/******************************************************************************/
	TermScaffold();
	
	return 1;
}

