/* @(#) Sound3D_FT.c 96/07/29 1.2 */
/*	
 *	File:				Sound3D_FT.c
 *	Author:				Eric L. Hayes
 *	Creation Date:		
 *
 *	Last Modified by:	Eric L. Hayes
 *	Modification Date:	3/22/96
 *
 *	Version:			M2 2.0
 *
 *	This file contains the functional tests 3D sound.  There are three tests.
 *
 *	Test 1:	Move the sound to different angles of theta around the observer's head.  The observer
 *			should hear the sound moving around his head.
 *
 *	Test 2:	Move the sound with increading radius from the observer's head.  The sound should
 *			sound farther and farther away.  Get3DSoundPos should return a larger number as the
 *			sounds move farther away.
 *			
 *	Test 3:	Move a sound with the doppler effect.  The sound should be heard whizzing by.
 *
 *	Audio Functions called:
 *
 *		CloseAudioFolio		CreateCue			CreateAttachment		LoadInstrument			LoadSample
 *		OpenAudioFolio		StartInstrument		StopInstrument			UnloadSample
 *		SleepUntilTime		DeleteAttachment	DeleteCue				UnloadInstrument
 *		Create3DSound		Delete3DSound		Get3DSoundInstrument	Get3dSoundParams
 *		Get3DSoundPos		Move3dSound			Move3DSoundTo			Start3DSound
 *		Stop3DSound
 *		
 *  Copyright (C) 1996, an unpublished work by The 3DO Company.
 *  All rights reserved. This material contains confidential
 *  information that is the property of The 3DO Company. Any
 *  unauthorized duplication, disclosure or use is prohibited.
 *  
 */



#include <audio/audio.h>
#include <audio/parse_aiff.h>
#include <audio/sound3d.h>
#include <stdio.h>
#include <stdlib.h>

#include <TestMacros.h>
#include <tests/3do_scaf.h>




/**************************************Test Environment Globals****************/
/******************************************************************************/

int main (int argc, char ** argv)
{
	Boolean			changeDirection	= FALSE;
	Err				theError;
	int32			theSignal;
	Item 			theSample, sampleInstrument, theCue, sleepCue, sampleAttachment, lineOut,
					ampKnob, freqKnob, sound3dInstrument;
	PolarPosition4D	startPosition, endPosition, newPosition;
	Sound3D			*theSound;
	Sound3DParms 	theSoundParms;


	InitScaffold(AUDIO, argc, argv);					/* initializes this program */

	theError = OpenAudioFolio();
	CHECKRESULT_SETUP(FT, theError,"OpenAudioFolio", cleanup);
	
	theCue	= CreateCue( NULL );
	CHECKRESULT_SETUP(FT, theCue,"CreateCue", cleanup);
	
	sleepCue	= CreateCue( NULL );
	CHECKRESULT_SETUP(FT, sleepCue,"CreateCue", cleanup);
	

 	theError	= Create3DSound( &theSound, NULL );
	CHECKRESULT_SETUP(FT, theError,"Create3DSound( theSound, NULL )", cleanup);


	sound3dInstrument = Get3DSoundInstrument( theSound );


	/*	Load the instrument which will play the sample to be written to the delay line,  load the
		sample, and attach them
	*/
	sampleInstrument = LoadInstrument("sampler_16_v1.dsp", 0, 100);
	CHECKRESULT_SETUP(FT, sampleInstrument,"LoadInstrument(sampler_16_v1.dsp)", cleanup);

	ampKnob			= CreateKnob( sampleInstrument, "Amplitude", NULL );
	CHECKRESULT_SETUP(FT, ampKnob,"CreateKnob(sampler_16_v1.dsp)", cleanup);
	
	freqKnob			= CreateKnob( sampleInstrument, "SampleRate", NULL );
	CHECKRESULT_SETUP(FT, freqKnob,"CreateKnob(sampler_16_v1.dsp)", cleanup);
	
	
	theSample = LoadSample("LongFile_44K_16B_M.AIFF");
	CHECKRESULT_SETUP(FT, theSample,"LoadSample(LongFile_44K_16B_M.AIFF)", cleanup);

	sampleAttachment = CreateAttachment( sampleInstrument, theSample, NULL );
	CHECKRESULT_SETUP(FT, sampleAttachment,"CreateAttachment(sampleInstrument, theSample", cleanup);

	MonitorAttachment( sampleAttachment, theCue, CUE_AT_END );

	theSignal	= GetCueSignal( theCue );
	CHECKRESULT_SETUP(FT, theSignal,"GetCueSignal", cleanup);

	/*	Load the line_out Instrument
	*/
	lineOut = LoadInstrument("line_out.dsp", 0, 100);
	CHECKRESULT_SETUP(FT, lineOut,"LoadInstrument(line_out)", cleanup);



	/*	Connect all of the instruments to the mixer
	*/
	theError	= ConnectInstrumentParts(	sampleInstrument, "Output", 0, sound3dInstrument, "Input", 0 		);
	CHECKRESULT_SETUP(FT, theError,"ConnectInstrumentParts(sampleInstrument Output 0, sound3dInstrument Input 0)", cleanup);

	theError	= ConnectInstrumentParts(	sound3dInstrument, "Output", 0, lineOut, "Input", 0 		);
	CHECKRESULT_SETUP(FT, theError,"ConnectInstrumentParts(sound3dInstrument Output 0, lineOut Input 0)", cleanup);

	theError	= ConnectInstrumentParts(	sound3dInstrument, "Output", 1, lineOut, "Input", 1 		);
	CHECKRESULT_SETUP(FT, theError,"ConnectInstrumentParts(sound3dInstrument Output 1, lineOut Input 1)", cleanup);


	/*	Initialize the position of the sound		*/
   startPosition.pp4d_Radius		= 0;	/* in frame ticks */
   startPosition.pp4d_Theta			= 0;	/* in radians */
   startPosition.pp4d_Phi			= 0;	/* in radians */
   startPosition.pp4d_Time			= 0;	/* in frame ticks */
   startPosition.pp4d_or_Theta		= 0;	/* orientation, in radians */
   startPosition.pp4d_or_Phi		= 0;	/* orientation, in radians */


	/*	Initialize the end position of the sound		*/
   endPosition.pp4d_Radius			= 130;							/* in frame ticks */
   endPosition.pp4d_Theta			= 0;							/* in radians */
   endPosition.pp4d_Phi				= 0;							/* in radians */
   endPosition.pp4d_Time			= 100;							/* in frame ticks */
   endPosition.pp4d_or_Theta		= 0;							/* orientation, in radians */
   endPosition.pp4d_or_Phi			= 0;							/* orientation, in radians */


	theError	= StartInstrument( lineOut, NULL );
	CHECKRESULT_SETUP(FT, theError,"StartInstrument(lineOut, NULL)", cleanup);



/**************************************************************************************************************
 *	Test 1:		Play the sound which centered at the user's head (radius = 0) at various angles of Theta.
 *				The user should hear the sound moving around the ears.
 *
 *************************************************************************************************************/

	printf("\nStart playing 3D Sound at various angles of Theta (Radius = 0)\n\n" );

	theError	= StartInstrument( sampleInstrument, NULL ); 
	CHECKRESULT_SETUP(FT, theError,"StartInstrument(sampleInstrument, NULL)", cleanup);

	theError	= Start3DSound( theSound, &startPosition ); 
	CHECKRESULT_SETUP(FT, theError,"Start3DSound(sampleInstrument, NULL)", cleanup);

	while ( (GetCurrentSignals() & theSignal) != theSignal )
	{
		SleepUntilTime( sleepCue, GetAudioTime() + 240 );

		endPosition.pp4d_Theta						+= .79;							
		endPosition.pp4d_Theta						= s3dNormalizeAngle(endPosition.pp4d_Theta);
		
		theError	=  Move3DSound( theSound, &startPosition, &endPosition );
		CHECKRESULT_SETUP(FT, theError,"Move3DSound()", cleanup);

		SleepUntilTime( sleepCue, GetAudioTime() + 240 );
 		Get3DSoundPos( theSound, &newPosition );
		
		LogComment( "Theta is now %f radians", endPosition.pp4d_Theta );
		LogComment( "Theta returned by Get3DSoundPos is %f",  newPosition.pp4d_Theta );
		
		startPosition.pp4d_Theta			= endPosition.pp4d_Theta;
	}	
	startPosition.pp4d_Theta	= endPosition.pp4d_Theta	= 0;
	ClearCurrentSignals( theSignal );
	
	BeginTest(FT, "Sound3D");
	LogResult(InteractiveVerify("You should have heard the sound move around your head. Did Test Pass?"), "");


/**************************************************************************************************************
 *	Test 2:		Play the sound while increasing the radius.  The sample should sound like it is moving
 *				farther away.
 *
 *************************************************************************************************************/

	printf("\nStart playing 3D Sound at an increasing radius\n\n" );

	theError	= StartInstrument( sampleInstrument, NULL ); 
	CHECKRESULT_SETUP(FT, theError,"StartInstrument(sampleInstrument, NULL)", cleanup);

	theError	= Start3DSound( theSound, &startPosition ); 
	CHECKRESULT_SETUP(FT, theError,"Start3DSound(sampleInstrument, NULL)", cleanup);


	endPosition.pp4d_Time			= 0;							/* in frame ticks */
	while ( (GetCurrentSignals() & theSignal) != theSignal )
	{
		SleepUntilTime( sleepCue, GetAudioTime() + 720 );

		endPosition.pp4d_Radius			+= 100.0;							

		printf( "Radius is now %f frames\n", startPosition.pp4d_Radius );
		
		theError	=  Move3DSound( theSound, &startPosition, &endPosition );
		CHECKRESULT_SETUP(FT, theError,"Move3DSound()", cleanup);

 		theError	=  Get3DSoundParms( theSound, &theSoundParms, sizeof(theSoundParms) );
		CHECKRESULT_SETUP(FT, theError,"Get3DSoundParms()", cleanup);

		SetKnob( ampKnob, theSoundParms.s3dp_DistanceFactor );
		 
		SleepUntilTime( sleepCue, GetAudioTime() + 240 );
 		Get3DSoundPos( theSound, &newPosition );
		LogComment( "Radius returned by Get3DSoundPos is %f",  newPosition.pp4d_Radius );
		
		startPosition.pp4d_Radius			= endPosition.pp4d_Radius;
	}	
	startPosition.pp4d_Radius	= endPosition.pp4d_Radius	= 0;
	Move3DSound( theSound, &startPosition, &endPosition );
	SleepUntilTime( sleepCue, GetAudioTime() + 240 );
	ClearCurrentSignals( theSignal );
	
	BeginTest(FT, "Sound3D");
	LogResult(InteractiveVerify("You should have heard the sound move further away from you. Did Test Pass?"), "");


/**************************************************************************************************************
 *	Test 3:		Use the doppler effect to model a sound which whizzes by the user's head.
 *
 *************************************************************************************************************/

	printf("\nStart playing 3D Sound using the Doppler effect\n\n" );

	theError	= StartInstrument( sampleInstrument, NULL ); 
	CHECKRESULT_SETUP(FT, theError,"StartInstrument(sampleInstrument, NULL)", cleanup);

	theError	= Start3DSound( theSound, &startPosition ); 
	CHECKRESULT_SETUP(FT, theError,"Start3DSound(sampleInstrument, NULL)", cleanup);

	endPosition.pp4d_Radius			= 1000.0;							
	endPosition.pp4d_Theta			= 0;
	endPosition.pp4d_Time			= 0;							
	while ( (GetCurrentSignals() & theSignal) != theSignal )
	{
	
		if ( changeDirection == FALSE )
			endPosition.pp4d_Radius			-= 340.0;
		else
		{
			endPosition.pp4d_Radius			+= 340.0;
			endPosition.pp4d_Theta			= 1.57;
		}
		
		endPosition.pp4d_Time			+= 44100;
		
		if ( endPosition.pp4d_Radius < 0 )
			changeDirection	= TRUE;
		
		theError	=  Move3DSound( theSound, &startPosition, &endPosition );
		CHECKRESULT_SETUP(FT, theError,"Move3DSound()", cleanup);
		
		SleepUntilTime( sleepCue, GetAudioTime() + 240 );

 		theError	=  Get3DSoundParms( theSound, &theSoundParms, sizeof(theSoundParms) );
		CHECKRESULT_SETUP(FT, theError,"Get3DSoundParms()", cleanup);

		LogComment( "Radius is now %f frames.\t\tTheta is now %f", endPosition.pp4d_Radius, endPosition.pp4d_Theta );
		LogComment( "Doppler Shift is now %f", theSoundParms.s3dp_Doppler );
		SetKnob( freqKnob, theSoundParms.s3dp_Doppler*44100.0 );
		 
		startPosition.pp4d_Radius		= endPosition.pp4d_Radius;
		startPosition.pp4d_Time			= endPosition.pp4d_Time;
	}	
	startPosition.pp4d_Theta	= endPosition.pp4d_Theta	= 0;
	startPosition.pp4d_Radius	= endPosition.pp4d_Radius	= 0;
	startPosition.pp4d_Time		= endPosition.pp4d_Time		= 0;
	ClearCurrentSignals( theSignal );
	
	BeginTest(FT, "Sound3D");
	LogResult(InteractiveVerify("You should have notice a change in Doppler Shift variable. Did Test Pass?"), "");




	theError	= Stop3DSound( theSound );
	CHECKRESULT_SETUP(FT, theError,"Stop3DSound()", cleanup);

	StopInstrument( lineOut, NULL );
	StopInstrument( sampleInstrument, NULL );
	
	
	DeleteCue( theCue );
	DeleteCue( sleepCue );
	DeleteAttachment( sampleAttachment );
	Delete3DSound( &theSound );

	UnloadSample( theSample );
	UnloadInstrument( sampleInstrument );
	UnloadInstrument( lineOut );

cleanup:
	theError	= CloseAudioFolio();
	CHECKRESULT(FT, theError,"CloseAudioFolio");

	/******************************************************************************/
	TermScaffold();

	return 0;
}
