/* @(#) SoundSpoolerFT.c 96/07/29 1.2 */
/*	
 *	File:				SoundSpoolerFT.c
 *	Author:				Phil Burk
 *	Creation Date:		
 *
 *	Last Modified by:	Eric L. Hayes
 *	Modification Date:	3/15/96
 *
 *	Version:			M2 2.0
 *
 *	This file contains the functional test for the SoundSpooler.  This test is an old Opera
 *	test which was converted for use with M2.  This test program creates a sawtooth sample.
 *	The SoundSpooler starts spooling the data from the disk and plays the sound.  Using the
 *	control pad the user can simulate different events.  By pressing the A control button,
 *	the user can simulate data starvation, and the spooler will hang (stop).  Releasing button
 *	A will cause the spooler to start playing the sound again.  Pressing Button B causes the
 *	sound spooler to pause, and releasing button B causes the sound spooler to resume again.
 *	Pressing button C will cause the sound spooler to read only one buffer at a time.  The
 *	sound spooler will stay in this mode until the user presses the start button.  Pressing the
 *	X button causes the soundspooler to stop.
 *
 *
 *	Audio Functions called:
 *
 *		CloseAudioFolio			ssplGetUserData			ssplCreateSoundSpooler		ssplSpoolData			
 *		ssplStartSpoolerTags	ssplProcessSignals		ssplPause					ssplResume	
 *		ssplPlayData			ssplStopSpooler			ssplDeleteSoundSpooler		LoadInstrument
 *		StartInstrument			StopInstrument			UnloadInstrument			ConnectInstrumentParts
 *		WaitSignal				OpenAudioFolio
 *
 *  Copyright (C) 1996, an unpublished work by The 3DO Company.
 *  All rights reserved. This material contains confidential
 *  information that is the property of The 3DO Company. Any
 *  unauthorized duplication, disclosure or use is prohibited.
 *  
 */
 
 
/* Include this when using the Audio Folio */
#include <audio/audio.h>
#include <audio/parse_aiff.h>
#include <audio/soundspooler.h>

#include <file/filefunctions.h>
#include <file/filesystem.h>
#include <kernel/debug.h>
#include <kernel/operror.h>
#include <kernel/mem.h>
#include <kernel/types.h>
#include <misc/event.h>
#include <stdio.h>

#include <tests/3do_scaf.h>
#include <TestMacros.h>


#define	PRT(x)	{ printf x; }
#define	ERR(x)	PRT(x)
#define	DBUG(x)	/* PRT(x) */


#define NUM_BUFFERS   (4)
#define NUMCHANNELS (2)   /* Stereo */
#define NUMFRAMES (8*1024)
#define SAMPSIZE (sizeof(int16)*NUMFRAMES*NUMCHANNELS)

/************************************TEST ENV GLOBALS**************************/
Boolean			gContinue;
char *			gProgName = "PlayAIFF.c";		/* contains the name of the program */
char *			gTestName = "playaiff";			/* contains the name of the test    */
int32			gFlag;
int32			gTestLevel = BREADTH;			/* contains what level of testing occurs in this test */
/******************************************************************************/


/************************************************************************
** Fill in a sample with sawtooth waves.
************************************************************************/
void FillBufferWithSaw ( int16 *Data, int32 NumSamples, int32 Octave)
{
	int32  i;
	
/* Fill sample with a sawtooth  in different octaves. */
	for (i=0; i<NumSamples; i++)
	{
		*Data++ = (int16) ((i << (7+Octave)) & 0xFFFF);
	}
	
}

/************************************************************************
** Callback function for buffer completion processor.
************************************************************************/
void MyBufferProcessor( SoundSpooler *sspl, SoundBufferNode *sbn )
{
	LogComment("MyBufferProcessor: 0x%08x done, signal = 0x%08x, BufIndex = %d\n",
					sbn, sbn->sbn_Signal, ssplGetUserData( sspl, sbn ) );
}

/************************************************************************
** Main test.
************************************************************************/
int main(int argc, char *argv[])
{
	char				*Data[NUM_BUFFERS];
	ControlPadEventData	cped;
	Err					theErr;
	Item				OutputIns;
	Item				SamplerIns;
	int32				Result;
	int32				SignalMask;
	int32				MySignal[NUM_BUFFERS], CurSignals;
	int32				i;
	int32				BufIndex, Joy;
	SoundSpooler 		*sspl;



	InitScaffold(AUDIO, argc, argv);					/* initializes this program */
	
	PRT(( "%s started.\n", argv[0]));
	
/* Initalize local variables */
	OutputIns = 0;
	SamplerIns = 0;
	sspl = NULL;
	
/* Initialize the EventBroker. */
	Result = InitEventUtility(1, 0, LC_FocusListener);
	CHECKRESULT_SETUP(FT, Result, "InitEventUtility", cleanup);
	
/* Initialize audio, return if error. */ 
	theErr = OpenAudioFolio();
	CHECKRESULT_SETUP(FT, theErr, "OpenAudioFolio", cleanup);
	
/* Use line_out instead of mixer. */
	OutputIns = LoadInstrument("line_out.dsp",  0, 0);
	CHECKRESULT(FT, OutputIns,"LoadInstrument");

/* Load fixed rate stereo sample player instrument */
	SamplerIns = LoadInstrument("sampler_16_f2.dsp",  0, 100);
	CHECKRESULT(FT, SamplerIns,"LoadInstrument");
	
/* Connect Sampler Instrument to line_out */
	Result = ConnectInstrumentParts (SamplerIns, "Output", 0, OutputIns, "Input", 0);
	CHECKRESULT(FT, Result,"ConnectInstruments");

	Result = ConnectInstrumentParts (SamplerIns, "Output", 1, OutputIns, "Input", 1);
	CHECKRESULT(FT, Result,"ConnectInstruments");

	Result = StartInstrument( OutputIns, NULL );
	CHECKRESULT(FT, Result,"StartInstrument: OutputIns");

/* Allocate and fill buffers with arbitrary test data. */
	for( i=0; i<NUM_BUFFERS; i++ )
	{
		Data[i] = (char *) AllocMem( SAMPSIZE, MEMTYPE_NORMAL );
		if( Data[i] == NULL )
		{
			ERR(( "Not enough memory for sample .\n"));
			goto cleanup;
		}
		FillBufferWithSaw( (int16 *) Data[i], SAMPSIZE/sizeof(int16), i );
	}
	
/* Create SoundSpooler data structure. */
	sspl = ssplCreateSoundSpooler( NUM_BUFFERS, SamplerIns );
	if( sspl == NULL )
	{
		ERR(( "ssplCreateSoundSpooler failed!\n"));
		goto cleanup;
	}
	
/* Fill the sound queue by queuing up all the buffers. "Preroll" */
	BufIndex = 0;
	SignalMask = 0;
	for( i=0; i<NUM_BUFFERS; i++)
	{
/* Dispatch buffers full of sound to spooler. Set User Data to BufIndex.
** ssplSpoolData returns a signal which can be checked to see when the data
** has completed it playback.
*/
		MySignal[BufIndex] = ssplSpoolData( sspl, Data[BufIndex], SAMPSIZE, (void *) BufIndex );
		if( MySignal[BufIndex] <= 0)
		{
			ERR(( "ssplSpoolBuffer failed 0x%x\n", MySignal[BufIndex] ));
			goto cleanup;
		}
		SignalMask |= MySignal[BufIndex];
		BufIndex++;
		if(BufIndex >= NUM_BUFFERS) BufIndex = 0;
	}
	
/* Start Spooler instrument. Will begin playing any queued buffers. */
	Result = ssplStartSpoolerTags( sspl, NULL );
	CHECKRESULT(FT, Result,"ssplStartSpoolerTags");
	
/* Play buffers loop. */
	do
	{
/* Wait for some buffer(s) to complete. */
DBUG(("WaitSignal(0x%x)\n", SignalMask ));
		CurSignals = WaitSignal( SignalMask );
DBUG(("WaitSignal() got 0x%x\n", CurSignals ));

/* Tell sound spooler that the buffer(s) have completed. */
		Result = ssplProcessSignals( sspl, CurSignals, MyBufferProcessor );
		CHECKRESULT(FT, Result,"ssplProcessSignals");
		
/* Read current state of Control Pad. */
		Result = GetControlPad (1, FALSE, &cped);
		if (Result < 0) {
			ERR(( "Error in GetControlPad\n"));
			PrintfSysErr(Result);
		}
		Joy = cped.cped_ButtonBits;

/* Simulate data starvation. Hang if we hit A button. */		
		if(Joy & ControlA)
		{
			PRT(( "Hang spooler.\n"));
			do
			{
				GetControlPad (1, TRUE, &cped);
				Joy = cped.cped_ButtonBits;
			} while(Joy & ControlA);
			PRT(( "Unhang spooler.\n"));
		}
		
/* Pause if we hit button B. */		
		if(Joy & ControlB)
		{
			PRT(( "Pause.\n"));
			ssplPause( sspl );
			do
			{
				GetControlPad (1, TRUE, &cped);
				Joy = cped.cped_ButtonBits;
			} while(Joy & ControlB);
			ssplResume( sspl );
			PRT(( "Resume.\n"));
		}
		
		
/* Single shot mode. Send single buffers by hiting C, until Start hit. */		
		if(Joy & ControlC)
		{
			PRT(( "Single shot mode.\n"));
			do
			{
				GetControlPad (1, TRUE, &cped);
				Joy = cped.cped_ButtonBits;
				if(Joy & ControlC)
				{
					if( ssplPlayData( sspl, Data[BufIndex], SAMPSIZE) > 0 )
					{
						BufIndex++;
						if(BufIndex >= NUM_BUFFERS) BufIndex = 0;
					}
				}
			} while((Joy & ControlStart) == 0);
		}

/*
** Spool as many buffers as are available.
** ssplSpoolData will return positive signals as long as it accepted the data.
*/
		while( ssplSpoolData( sspl, Data[BufIndex], SAMPSIZE, (void *) BufIndex ) > 0 )
		{
/* INSERT YOUR CODE HERE TO FILL UP THE NEXT BUFFER */
			BufIndex++;
			if(BufIndex >= NUM_BUFFERS) BufIndex = 0;
		}
		
	} while( (Joy & ControlX) == 0 );
	
/* Stop Spooler. */
	Result = ssplStopSpooler( sspl );
	CHECKRESULT(FT, Result,"StopSoundSpooler");
	

	for( i=0; i<NUM_BUFFERS; i++ )
		FreeMem( Data[i], SAMPSIZE );

	
	BeginTest(FT, "SoundSpooler");
	LogResult(InteractiveVerify("Did Test Pass?"), "");

cleanup:
	ssplDeleteSoundSpooler( sspl );
	UnloadInstrument( SamplerIns );
	UnloadInstrument( OutputIns );
	
	KillEventUtility();
	theErr	= CloseAudioFolio();
	CHECKRESULT(FT,theErr,"CloseAudioFolio");
	
	TermScaffold();
	return((int) Result);
}
