/************************************************************************/
/* Written by Paige Wuestefeld					     	*/
/* Displays 6 models. The spheres act as chasers, the prelit box acts	*/
/* as a target and the tank and pot are distractors. C + shift starts	*/
/* the chase - on collision a sprite (brick wall) is flashed up on the	*/
/* screen.  The first level is not fogged, the second level has fogged 	*/
/* chasers.  Each level lasts for 5 minutes.				*/
/*									*/
/*Cases tested: 							*/
/*									*/
/*          DynLitTex		tank body, chasers (3DO)	*/
/*          DynLitEnv		tank body, chasers (3DO)	*/
/*	    DynLitTransTex	tank canopy 				*/
/*	    DynLitSpec		tank barrel, Pot			*/
/*	    DynLitSpecTex	fruit chaser (level 1) 			*/
/*	    DynLitFogTex	chasers (1 & 2 on level 2)		*/
/*	    DynLitFogEnv	chaser (3 on level 2)	     		*/
/*									*/
/* Requires:								*/
/*  sphere.bsf, sphere.utf, sphere2.bsf, sphere2.utf			*/
/*  sphere3.bsf, sphere3.utf, tank.bsf, tank.utf, teapot.bsf, teapot.utf*/
/*									*/
/*  Models created by Mike Rosefield					*/
/*									*/
/* Some code taken from helloworld.					*/
/* Modified 11/19 to use MP utility					*/
/************************************************************************/

#ifdef MACINTOSH
#include <kernel:types.h>
#include <kernel:mem.h>
#include <graphics:graphics.h>
#include <graphics:view.h>
#include <graphics:clt:gstate.h>
#include <misc:event.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <graphics:frame2d:f2d.h>
#include <:graphics:font.h>
#include <kernel:random.h>
#include <kernel:time.h>
#else
#include <kernel/types.h>
#include <kernel/mem.h>
#include <graphics/graphics.h>
#include <graphics/view.h>
#include <graphics/clt/gstate.h>
#include <misc/event.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <graphics/frame2d/f2d.h>
#include <graphics/font.h>
#include <kernel/random.h>
#include <kernel/time.h>
#endif

#include "graphicsenv.h"
#include "mercury.h"
#include "matrix.h"
#include "bsdf_read.h"
#include "filepod.h"
#include "SData.h"
#include "SimpleChase.h"
#include "scalemodel.h"
#include "math.h"

/* overarching structure */

  typedef struct AppData {
    CloseData *close;
    Matrix *skewMatrix;
    Matrix *camMatrix;
    uint32 curRenderScreen;
    float rx, ry, rz;		/* rotation of object */
    float tx, ty, tz;		/* location of object */
  }
AppData;

#define RUNMP 1
#define gHitherPlane			(10.0f)
#define kMaxVertsPerPod     4000
#define kMaxCmdListWords    400

/* Initial camera position in world coords */
#define kStartCamPosnX      0.0
#define kStartCamPosnY      0.0
#define kStartCamPosnZ      300.0

#define kRadian             	(PI/180.0)
#define kTurnCameraIncrement	(2.5f * kRadian)
#define kMoveCameraIncrement    (7.5f)
#define kMoveIncrement			(5.0f)
#define NUMSECONDS				300
#define MAXPODS			20
/* these are r g b for both the fog and the clear screen color */
#define fog_red					(0.1f)
#define fog_green				(0.1f)
#define fog_blue				(0.1f)

/* Prototypes for routines only used in this file */
void Program_Init (GraphicsEnv * genv, AppData ** app);
void Program_Begin (GraphicsEnv * genv, AppData * app);
Pod *ChangeCase (Pod * aPod);
void Chase (Vector3D * target, Pod * thePod);
Vector3D *GetCenter (Pod * myPod);
float GetRadius (Pod * myPod);
float GetDist3D (Vector3D * center1, Vector3D * center2);
bool CheckColl (Vector3D * target_current, Vector3D * chaser_current, Vector3D * tc, Vector3D * cc,
		Pod * targetPod, Pod * chaserPod, float safe);
void MoveAbout (Matrix * theMatrix, Vector3D * fuzzytarget);
void HyperSpace (Matrix * theMatrix);

/* --------------------------------------------------------------------- */


int 
main (int argc, char *argv[])
{
  Err err;
  GraphicsEnv *genv;
  AppData *appData;

  (void) &argv;
  (void) &argc;

  /* Setup Display, GState */

  genv = GraphicsEnv_Create ();
  if (genv == NULL) {
    printf ("Couldn't allocate graphics environment.  Exiting.\n");
    exit (1);
  }
  err = GraphicsEnv_Init (genv, 320, 240, 32);
  if (err < 0) {
    PrintfSysErr (err);
    exit (err);
  }

  /* Setup the EventBroker */

  err = InitEventUtility (1, 0, 0);
  if (err < 0) {
    PrintfSysErr (err);
    exit (err);
  }

  Program_Init (genv, &appData);
  Program_Begin (genv, appData);

  printf ("Program exiting.\n");
  GraphicsEnv_Delete (genv);

  return 0;
}


/* --------------------------------------------------------------------- */


void 
Program_Init (GraphicsEnv * genv, AppData ** app)
{
  ViewPyramid vp;
  AppData *appData;

  *app = AllocMem (sizeof (AppData), MEMTYPE_NORMAL);
  if (*app == NULL) {
    printf ("Couldn't create AppData structure.  Exiting\n");
    exit (1);
  }
  appData = *app;

  /* Initialize Mercury */
  appData->close = M_Init (kMaxVertsPerPod, kMaxCmdListWords, genv->gs);
#if RUNMP
  M_InitMP (appData->close, MAXPODS, 100000, 650);
#endif

  if (appData->close == NULL) {
    printf ("Couldn't init Mercury.  Exiting\n");
    exit (1);
  }

  /* required close data arguments */

  appData->close->fwclose = 1.01;
  appData->close->fwfar = 100000.0;
  appData->close->fscreenwidth = genv->d->width;
  appData->close->fscreenheight = genv->d->height;
  appData->close->depth = genv->d->depth;

  /* Setup camera */
  appData->skewMatrix = AllocMem (sizeof (Matrix), MEMTYPE_NORMAL);
  if (appData->skewMatrix == NULL) {
    printf ("Couldn't allocate skew matrix.  Exiting\n");
    exit (1);
  }

  vp.left = -1.0;
  vp.right = 1.0;
  vp.top = 0.75;		/* Basic 4/3 aspect ratio, with positive Y UP */
  vp.bottom = -0.75;
  vp.hither = 1.2;

  Matrix_Perspective (appData->skewMatrix, &vp, 0.0, (float) (genv->d->width),
		      0.0, (float) (genv->d->height), 1.0 / gHitherPlane);

  appData->camMatrix = Matrix_Construct ();
  if (appData->camMatrix == NULL) {
    printf ("Couldn't allocate camera matrix.  Exiting\n");
    exit (1);
  }
  {
    Vector3D center;
    Vector3D position;

    Vector3D_Set (&center, 0.0f, 0.0f, 0.0f);
    Matrix_Identity (appData->camMatrix);
    Matrix_Translate (appData->camMatrix,
		      kStartCamPosnX, kStartCamPosnY, kStartCamPosnZ);
    Matrix_GetTranslation (appData->camMatrix, &position);
    Matrix_LookAt (appData->camMatrix, (Point3D *) & center,
		   (Point3D *) & position, 0.0f);
  }

  M_SetCamera (appData->close, appData->camMatrix, appData->skewMatrix);

}


/* --------------------------------------------------------------------- */

static void 
Controller_Update (ControlPadEventData * cped,
		   Matrix * target, Matrix * pot, Matrix * tank,
		   Matrix * chaser1, Matrix * chaser2, Matrix * chaser3)
{
  Vector3D moveObj1, moveObj2, moveObj3, moveObj4, center;

  Vector3D_Zero (&center);
  Vector3D_Zero (&moveObj1);
  Vector3D_Zero (&moveObj2);
  Vector3D_Zero (&moveObj3);
  Vector3D_Zero (&moveObj4);

  if (cped->cped_ButtonBits & ControlA) {
    if (cped->cped_ButtonBits & ControlUp)
      moveObj1.y = kMoveIncrement;
    if (cped->cped_ButtonBits & ControlDown)
      moveObj1.y = -kMoveIncrement;
    if (cped->cped_ButtonBits & ControlRight)
      moveObj1.x = kMoveIncrement;
    if (cped->cped_ButtonBits & ControlLeft)
      moveObj1.x = -kMoveIncrement;
    if (cped->cped_ButtonBits & ControlLeftShift)
      moveObj1.z = kMoveIncrement;
    if (cped->cped_ButtonBits & ControlRightShift)
      moveObj1.z = -kMoveIncrement;

    Matrix_TranslateByVector (target, &moveObj1);
  }


  if (cped->cped_ButtonBits & ControlB) {
    if (cped->cped_ButtonBits & ControlUp)
      moveObj2.y = kMoveIncrement;
    if (cped->cped_ButtonBits & ControlDown)
      moveObj2.y = -kMoveIncrement;
    if (cped->cped_ButtonBits & ControlRight)
      moveObj2.x = kMoveIncrement;
    if (cped->cped_ButtonBits & ControlLeft)
      moveObj2.x = -kMoveIncrement;
    if (cped->cped_ButtonBits & ControlLeftShift)
      moveObj2.z = kMoveIncrement;
    if (cped->cped_ButtonBits & ControlRightShift)
      moveObj2.z = -kMoveIncrement;
  }
  Matrix_TranslateByVector (pot, &moveObj2);

  if (cped->cped_ButtonBits & ControlC) {
    if (cped->cped_ButtonBits & ControlUp)
      moveObj3.y = kMoveIncrement;
    if (cped->cped_ButtonBits & ControlDown)
      moveObj3.y = -kMoveIncrement;
    if (cped->cped_ButtonBits & ControlRight)
      moveObj3.x = kMoveIncrement;
    if (cped->cped_ButtonBits & ControlLeft)
      moveObj3.x = -kMoveIncrement;
    if (cped->cped_ButtonBits & ControlLeftShift)
      moveObj3.z = kMoveIncrement;
    if (cped->cped_ButtonBits & ControlRightShift)
      moveObj3.z = -kMoveIncrement;
  }
  Matrix_TranslateByVector (tank, &moveObj3);

  if (cped->cped_ButtonBits & ControlD) {
    if (cped->cped_ButtonBits & ControlUp)
      moveObj4.y = kMoveIncrement;
    if (cped->cped_ButtonBits & ControlDown)
      moveObj4.y = -kMoveIncrement;
    if (cped->cped_ButtonBits & ControlRight)
      moveObj4.x = kMoveIncrement;
    if (cped->cped_ButtonBits & ControlLeft)
      moveObj4.x = -kMoveIncrement;
    if (cped->cped_ButtonBits & ControlLeftShift)
      moveObj4.z = kMoveIncrement;
    if (cped->cped_ButtonBits & ControlRightShift)
      moveObj4.z = -kMoveIncrement;

    Matrix_TranslateByVector (chaser1, &moveObj4);
  }

  if (cped->cped_ButtonBits & ControlStart) {
    Matrix_Identity (target);
    Matrix_Identity (pot);
    Matrix_Identity (tank);
    Matrix_Identity (chaser1);
    Matrix_Identity (chaser2);
    Matrix_Identity (chaser3);

    Matrix_Translate (chaser1, 0.0, 0.0, +100.0);
    Matrix_Translate (pot, -100.0, -50.0, -50.0);
    Matrix_RotateX (tank, (kRadian * 180.0));
    Matrix_Translate (tank, -150.0, -75.0, 0.0);
    Matrix_Translate (chaser2, -100.0, -100.0, 0.0);
    Matrix_Translate (chaser3, 100.0, -100.0, 75.0);

  }

  if (cped->cped_ButtonBits & ControlE)
    HyperSpace (target);

}
/* --------------------------------------------------------------------- */
void 
HyperSpace (Matrix * theMatrix)
{

  srand (ReadHardwareRandomNumber ());

  Matrix_SetTranslation (theMatrix, ((rand () % 200) - 100), ((rand () % 200) - 200), ((rand () % 200) - 200));

}

/* --------------------------------------------------------------------- */

void 
Chase (Vector3D * target, Pod * thePod)
{
  int IncX, IncY, IncZ;
  Vector3D *move, *chaser;
  Pod *curPod;

  move = Vector3D_Construct ();
  chaser = Vector3D_Construct ();

  target->x -= 3000.0;
  target->y -= 3000.0;
  target->z -= 3000.0;

  curPod = thePod;
  IncX = 1;
  IncY = 1;
  IncZ = 1;

  while (curPod != NULL) {

    chaser = GetCenter (curPod);
    Vector3D_Add (chaser, chaser, Matrix_GetpTranslation (curPod->pmatrix));

    if (Vector3D_GetX (chaser) < Vector3D_GetX (target))
      move->x = (float) IncX;
    else {
      if (Vector3D_GetX (chaser) > Vector3D_GetX (target))
	move->x = -(float) IncX;
      else
	move->x = 0.0;
    }

    if (Vector3D_GetY (chaser) < Vector3D_GetY (target))
      move->y = (float) IncY;
    else {
      if (Vector3D_GetY (chaser) > Vector3D_GetY (target))
	move->y = -(float) IncY;
      else
	move->y = 0.0;
    }

    if (Vector3D_GetZ (chaser) < Vector3D_GetZ (target))
      move->z = (float) IncZ;
    else {
      if (Vector3D_GetZ (chaser) > Vector3D_GetZ (target))
	move->z = -(float) IncZ;
      else
	move->z = 0.0;
    }

    Matrix_TranslateByVector (curPod->pmatrix, move);

    curPod = curPod->pnext;
    IncX++;
    IncY++;
    IncZ++;

  }


  target->x += 3000.0;
  target->y += 3000.0;
  target->z += 3000.0;

  Vector3D_Destruct (move);
  Vector3D_Destruct (chaser);
}
/* --------------------------------------------------------------------- */
/* --------------------------------------------------------------------- */

void 
MoveAbout (Matrix * theMatrix, Vector3D * fuzzytarget)
{
  Vector3D *move;

  move = Vector3D_Construct ();

  if ((Vector3D_GetX (fuzzytarget) - 3000.0) < 0.0)
    move->x = 1.0;
  else {
    if ((Vector3D_GetX (fuzzytarget) - 3000.0) > 0.0)
      move->x = (-1.0);
    else
      move->x = 0.0;
  }

  if ((Vector3D_GetY (fuzzytarget) - 3000.0) < 0.0)
    move->y = 1.0;
  else {
    if ((Vector3D_GetY (fuzzytarget) - 3000.0) > 0.0)
      move->y = (-1.0);
    else
      move->y = 0.0;
  }

  if ((Vector3D_GetZ (fuzzytarget) - 3000.0) < 0.0)
    move->z = 1.0;
  else {
    if ((Vector3D_GetZ (fuzzytarget) - 3000.0) > 0.0)
      move->z = (-1.0);
    else
      move->z = 0.0;
  }

  Matrix_TranslateByVector (theMatrix, move);
  Vector3D_Destruct (move);
}

/* --------------------------------------------------------------------- */
/* --------------------------------------------------------------------- */
Vector3D *
GetCenter (Pod * myPod)
{

  Vector3D *Center;

  Center = Vector3D_Construct ();
  Vector3D_Set (Center, (myPod->pgeometry->fxmin) + (myPod->pgeometry->fxextent / 2.0),
	     (myPod->pgeometry->fymin) + (myPod->pgeometry->fyextent / 2.0),
	    (myPod->pgeometry->fzmin) + (myPod->pgeometry->fzextent / 2.0));
  return (Center);
}
/* --------------------------------------------------------------------- */
/* --------------------------------------------------------------------- */

float 
GetRadius (Pod * myPod)
{

  float theRadius;

  if (myPod->pgeometry->fxextent > myPod->pgeometry->fyextent) {
    if (myPod->pgeometry->fxextent > myPod->pgeometry->fzextent)
      theRadius = myPod->pgeometry->fxextent / 2.0;
    else
      theRadius = myPod->pgeometry->fzextent / 2.0;
  }
  else {
    if (myPod->pgeometry->fyextent > myPod->pgeometry->fzextent)
      theRadius = myPod->pgeometry->fyextent / 2.0;
    else
      theRadius = myPod->pgeometry->fzextent / 2.0;
  }

  return (theRadius);
}
/* --------------------------------------------------------------------- */
/* --------------------------------------------------------------------- */

float 
GetDist3D (Vector3D * center1, Vector3D * center2)
{

  Vector3D *diff;
  float dist;

  diff = Vector3D_Construct ();

  /* calculate distance between the 2 centers */
  Vector3D_Subtract (diff, center1, center2);
  dist = sqrtf (SQR (diff->x) +
		SQR (diff->y) +
		SQR (diff->z));
  return (dist);

}
/* --------------------------------------------------------------------- */
bool 
CheckColl (Vector3D * target_current, Vector3D * chaser_current, Vector3D * tc, Vector3D * cc,
	   Pod * targetPod, Pod * chaserPod, float safe)
{

  float actualDist;
  Pod *curPod;
  bool coll = FALSE;

  curPod = chaserPod;

  /* check for collision */

  /* add translation vectors to center values */
  Vector3D_Add (target_current, tc, Matrix_GetpTranslation (targetPod->pmatrix));

  while (curPod != NULL) {
    Vector3D_Add (chaser_current, cc, Matrix_GetpTranslation (curPod->pmatrix));
    actualDist = GetDist3D (target_current, chaser_current);

    if (actualDist <= safe) {
      coll = TRUE;
      break;
    }
    else
      coll = FALSE;

    curPod = curPod->pnext;
  }

  return (coll);

}

/* --------------------------------------------------------------------- */
void 
Program_Begin (GraphicsEnv * genv, AppData * app)
{
  ControlPadEventData cped;
  Pod *chasePod, *potPod, *tankPod;
  Pod *targetPod;
  Pod *curPod;
  bool done;
  Bitmap *myDestBM;
  uint32 numPods = 0;
  uint32 numPot, numTank, numChasePods, numPods1, numPods5, numPods6;
  uint32 triTarget, triChase, triPot, triTank;
  int frame, crashes;
  int i, level;
  uint32 MaxVertsPerPod = kMaxVertsPerPod;
  Matrix *gMatrix;
  Vector3D *sphereCenter, *boxCenter;
  Vector3D *pad, *targetCenter, *chaseCenter, *target_current, *chase_current;
  float sphereRadius, boxRadius;
  SpriteObj *sp;
  bool stopChase;
  bool collide, state;
  bool timeUP;
  TimeVal *timesample1, *timesample2;
  Item font;
  TextState *ts = NULL;
  FontTextArray fta[3];
  char *myString, *myString2, *RString;

/* TimeVal structure defined in kernel:time.h */
  timesample1 = AllocMem (sizeof (TimeVal), MEMTYPE_NORMAL);
  timesample2 = AllocMem (sizeof (TimeVal), MEMTYPE_NORMAL);

  targetPod = AllocMem (sizeof (Pod), MEMTYPE_NORMAL);

/* font string setup */
  myString = AllocMem (35 * sizeof (char), MEMTYPE_NORMAL);
  myString2 = AllocMem (35 * sizeof (char), MEMTYPE_NORMAL);
  RString = AllocMem (35 * sizeof (char), MEMTYPE_NORMAL);

  strcpy (myString, "Level 1 - No Fog\nCollisions ");
  strcpy (myString2, "Level 2 - Fogged\nCollisions ");

/*font folio initial setup */
  OpenFontFolio ();
  font = OpenFont ("default_14");

  fta[0].fta_StructSize = sizeof (FontTextArray);
  fta[0].fta_Clip.min.x = 0;
  fta[0].fta_Clip.min.y = 0;
  fta[0].fta_Clip.max.x = (genv->d->width - 1);
  fta[0].fta_Clip.max.y = (genv->d->height - 1);
  memset (&fta[0].fta_Pen, 0, sizeof (PenInfo));
  fta[0].fta_Pen.pen_X = 60.0;
  fta[0].fta_Pen.pen_Y = 40.0;
  fta[0].fta_Pen.pen_BgColor = 0x000000;
  fta[0].fta_Pen.pen_FgColor = 0xc0c0c0;
  fta[0].fta_Pen.pen_XScale = 1.0;
  fta[0].fta_Pen.pen_YScale = 1.0;
  fta[0].fta_String = "Level 1 - No Fog\nCollisions ";
  fta[0].fta_NumChars = strlen (fta[0].fta_String);
  CreateTextState (&ts, font, fta, 1);

/* Load in the models */

/*************/
/*  POT      */
/*************/

/* pot is not grouped correctly - workaround sets their 	*/
/* pmatrices to same location gMatrix 				*/

  Model_LoadFromDisk ("teapot", &numPods, &MaxVertsPerPod);
  printf ("Pot -> DynLit with white specular reflections\n");

/* set the different elements of the object to same pMatrix*/
  gMatrix = Matrix_Construct ();
  Matrix_Identity (gMatrix);

  potPod = gBSDF->pods;
  if (potPod == NULL) {
    printf ("NULL !\n");
    exit (1);
  }

  numPot = gBSDF->numPods;

  curPod = potPod;
  for (i = 0; i < numPot; i++) {
    curPod->pmatrix = gMatrix;
    Model_Scale (curPod, 50.0);
	curPod->pmaterial->shine=0.5;
    curPod = curPod->pnext;
  }
  Matrix_Translate (potPod->pmatrix, -100.0, -50.0, -50.0);

  printf("Reset shine values for pot to %f\n", potPod->pmaterial->shine );

/*************/
/*  TANK     */
/*************/

  Model_LoadFromDisk ("tank", &numPods, &MaxVertsPerPod);
  printf ("Tank Model -> \nDynLitTex body\nDynLitTransTex canopy\nDynLit gun barrel (with specular)\n ");

  tankPod = gBSDF->pods;
  if (potPod == NULL) {
    printf ("NULL !\n");
    exit (1);
  }

  numTank = gBSDF->numPods;

  curPod = tankPod;
  for (i = 0; i < numTank; i++) {
    Model_Scale (curPod, 0.1);
    curPod = curPod->pnext;
  }

  Matrix_RotateX (tankPod->pmatrix, (kRadian * 180.0));
  Matrix_Translate (tankPod->pmatrix, -150.0, -75.0, 0.0);

/*************/
/*  TARGET   */
/*************/

  targetPod = &gBoxDynPod;

/* use prelight function to change normals to RGB then change setup function and turn off lights */
  M_PreLight (targetPod, targetPod, app->close);
  targetPod->pcase = &M_SetupPreLitTrans;
  targetPod->plights = gNoLightList;
  printf ("Target PRELIT with directional light from 3 sides\n");

  TOUCH(targetPod);

/*************/
/* CHASERS   */
/*************/

  Model_LoadFromDisk ("sphere2", &numPods, &MaxVertsPerPod);

  chasePod = gBSDF->pods;
  if (chasePod == NULL) {
    printf ("NULL !\n");
    exit (1);
  }

  numPods1 = gBSDF->numPods;

  Model_Scale (chasePod, 0.40);
  chasePod->pmaterial = &gOpaqueMat;
  chasePod->plights = gModelLightListSpecTex;
  printf("Fruit Sphere -> DynLitSpecTex\n");

  Matrix_Translate (chasePod->pmatrix, 0.0, 0.0, +100.0);
/*************/

  Model_LoadFromDisk ("sphere", &numPods, &MaxVertsPerPod);

  chasePod->pnext = gBSDF->pods;
  if (chasePod->pnext->pnext == NULL) {
    printf ("NULL !\n");
    exit (1);
  }

  numPods6 = gBSDF->numPods;

  Model_Scale ((chasePod->pnext), 0.40);
  printf("Changing one sphere to DynLitEnv Pod\n");
  (chasePod->pnext)->pcase = &M_SetupDynLitEnv;
  (chasePod->pnext)->plights = podenvlights;
  (chasePod->pnext)->pmaterial = &gOpaqueMat;
  printf("Environmental Sphere -> DynLitEnv\n");

  Matrix_Translate (chasePod->pnext->pmatrix, 100.0, -100.0, 0.0);

/**********/
  Model_LoadFromDisk ("sphere3", &numPods, &MaxVertsPerPod);

  chasePod->pnext->pnext = gBSDF->pods;
  if (chasePod->pnext->pnext == NULL) {
    printf ("NULL !\n");
    exit (1);
  }

  numPods5 = gBSDF->numPods;

  Model_Scale ((chasePod->pnext->pnext), 0.40);
  (chasePod->pnext->pnext)->pcase = &M_SetupDynLitTex;
  (chasePod->pnext->pnext)->pmaterial = &gOpaqueMat;
  printf("Fruit Sphere -> DynLitSpecTex\n");

  Matrix_Translate (chasePod->pnext->pnext->pmatrix, -100.0, -100.0, 75.0);
/**********/

  numChasePods = numPods1 + numPods5 + numPods6;

/* initial collision setup */

  sphereCenter = Vector3D_Construct ();
  sphereCenter = GetCenter (chasePod);
  sphereRadius = GetRadius (chasePod);

  boxCenter = Vector3D_Construct ();
  boxCenter = GetCenter (targetPod);
  boxRadius = GetRadius (targetPod);

/* pad box and sphere centers so no negative values */
  pad = Vector3D_Construct ();
  Vector3D_Set (pad, 3000.0, 3000.0, 3000.0);
  targetCenter = Vector3D_Construct ();
  chaseCenter = Vector3D_Construct ();

  Vector3D_Add (chaseCenter, sphereCenter, pad);
  Vector3D_Add (targetCenter, boxCenter, pad);

  TOUCH(sphereCenter);
  TOUCH(boxCenter);

/* create vectors we will be needing later */
  chase_current = Vector3D_Construct ();
  target_current = Vector3D_Construct ();

/* create, load, scale and set sprite attributes */

  sp = Spr_Create (0);
  Spr_LoadUTF (sp, "brick.utf");
  if (sp == NULL) {
    printf ("NULL texture\n");
    exit (1);
  }

  Spr_SetTextureAttr (sp, TXA_ColorOut, TX_BlendOutSelectTex);
  Spr_ResetCorners (sp, SPR_TOPLEFT);
  Spr_Scale (sp, 3.0, 2.5);

  app->curRenderScreen = 0;

  printf ("Finished initialization\n");

  printf ("Instructions:\n");
  printf ("Controls:\n");
  printf ("1 directional light (-1 -1 -1). 2 point lights at (-100, 100, 100) and (100, 100, 100).\n");
  printf ("A + Arrow and Shifts moves transparent red box (target) in X, Y and Z\n");
  printf ("B + Arrow and Shifts moves opaque blue pot in X, Y and Z\n");
  printf ("C + Arrow and Shifts moves tank in X, Y and Z\n");
  printf ("L + Arrow and Shifts moves single textured sphere in X, Y and Z\n");

  printf ("R + Left Shift starts CHASE\n");
  printf ("The Chase: box is target, spheres are chasers, pot and tank are distractors\n");
  printf ("Goal: Avoid collisions with the textured spheres. On collision a 2D sprite \n");
  printf ("      is flashed on the screen. The first level lasts for 5 minutes and is not\n");
  printf ("      fogged. The second has fogged chasers to make the game a little harder\n");

  printf ("R + Right Shift ends chase\n");
  printf ("P randomly repositions Target\n");

  printf ("Pause repositions all objects to initial positions\n");
  printf ("X terminates level\n");

  for (level = 0; level < 2; level++) {

/* initial variable setup */
    frame = 0;
    stopChase = TRUE;
    collide = FALSE;
    state = FALSE;
    timeUP = FALSE;
    done = FALSE;
    crashes = 0;
    triTarget = 0;
    triChase = 0;
    triPot = 0;
    triTank = 0;

  TOUCH(collide);

/* font setup */
    if (level == 0)
      sprintf (RString, "%s %i", myString, crashes);
    else
      sprintf (RString, "%s %i", myString2, crashes);

    fta[0].fta_String = RString;
    fta[0].fta_NumChars = strlen (fta[0].fta_String);
    CreateTextState (&ts, font, fta, 1);

/* sort pods */
    chasePod = M_Sort (numChasePods, chasePod, NULL);
    potPod = M_Sort (numPot, potPod, NULL);
    tankPod = M_Sort (numTank, tankPod, NULL);

    printf ("BEGIN TEST...\n");

/************************************************************************/
/* begin game play stuff						*/
/************************************************************************/

    SampleSystemTimeTV (timesample1);

    do {

      GS_BeginFrame (genv->gs);

      SampleSystemTimeTV (timesample2);

      /* run for NUMSECONDS per level */
      if ((timesample2->tv_sec - timesample1->tv_sec) > NUMSECONDS)
	timeUP = TRUE;

      CLT_ClearFrameBuffer (genv->gs, fog_red, fog_green, fog_blue, 1.0, TRUE, TRUE);

      /* init destination buffer to known state (do each frame) */
      M_DBInit (GS_Ptr (genv->gs), 0, 0, genv->d->width, genv->d->height);

      /* Get Control Pad info */
      GetControlPad (1, FALSE, &cped);

      if ((cped.cped_ButtonBits & ControlX))
	done = TRUE;

      /* set stopChase with R button and shifts */
      if ((cped.cped_ButtonBits & ControlF)) {
	if ((cped.cped_ButtonBits & ControlLeftShift))
	  stopChase = FALSE;
	if ((cped.cped_ButtonBits & ControlRightShift))
	  stopChase = TRUE;
      }

      Controller_Update (&cped, targetPod->pmatrix, potPod->pmatrix, tankPod->pmatrix,
			 chasePod->pmatrix, (chasePod->pnext)->pmatrix, (chasePod->pnext->pnext)->pmatrix);

      if (!stopChase)
	Chase (target_current, chasePod);

      /* on COLLISION draw sprite and increment the collision font line */
      if (CheckColl (target_current, chase_current, targetCenter, chaseCenter,
		     targetPod, chasePod, (sphereRadius + boxRadius)))
	collide = TRUE;
      else
	collide = FALSE;

      /* on collision which is a change in state increment crashes */
      if ((collide != state) && collide)
	crashes++;
      state = collide;

      if (collide) {
	F2_Draw (genv->gs, sp);

	if (level == 0)
	  sprintf (RString, "%s %i", myString, crashes);
	else
	  sprintf (RString, "%s %i", myString2, crashes);
	fta[0].fta_String = RString;
	fta[0].fta_NumChars = strlen (fta[0].fta_String);
	CreateTextState (&ts, font, fta, 1);
      }

      else {

	Matrix_RotateXLocal (chasePod->pnext->pmatrix, kRadian * 2.0);

	Matrix_TurnXLocal (potPod->pmatrix, kRadian);
	Matrix_TurnYLocal (potPod->pmatrix, kRadian);
	Matrix_TurnZLocal (potPod->pmatrix, kRadian);


	if (!stopChase) {
	  /* move the distractor pods in pseudo random fashion */
	  MoveAbout (potPod->pmatrix, target_current);
	  MoveAbout (tankPod->pmatrix, target_current);
	}

	/* Updata camera */
	M_SetCamera (app->close, app->camMatrix, app->skewMatrix);

	/* set src address for transparency */
	myDestBM = (Bitmap *) LookupItem (GS_GetDestBuffer (genv->gs));
	app->close->srcaddr = (uint32) myDestBM->bm_Buffer;

	app->close->fogcolor = M_PackColor (fog_red, fog_green, fog_blue);

	/* draw pods */
	triChase = M_Draw (chasePod, app->close);
	triPot = M_Draw (potPod, app->close);
	triTank = M_Draw (tankPod, app->close);

	/* Potential GOTCHA - need to draw transparent pods last */
	triTarget = M_Draw (targetPod, app->close);
#if RUNMP
	M_DrawEnd (app->close);
#endif

	/* use font */
	DrawText (genv->gs, ts);
      }

      genv->gs->gs_SendList (genv->gs);

      GS_EndFrame (genv->gs);
      app->curRenderScreen = (app->curRenderScreen + 1) % 2;
      GS_SetDestBuffer (genv->gs, genv->bitmaps[app->curRenderScreen]);

      frame++;

    } while (!done && !timeUP);

    printf ("Triangles drawn last frame (unclipped): %i %i %i %i \n", (triTarget & 0xffff), (triChase & 0xffff), 
			(triPot & 0xffff), (triTank & 0xffff) );
    printf ("Triangles drawn last frame (clipped): %i %i %i %i \n", ((triTarget & 0xffff0000)>>16), 
			((triChase & 0xffff0000)>>16), ((triPot & 0xffff0000)>>16), ((triTank & 0xffff0000)>>16) );

	curPod=chasePod;

	while (curPod != NULL) {

	if (curPod->pcase == &M_SetupDynLitSpecTex) {
	/* fruit */
    printf ("Fogging fruit sphere -> DynLitFogTex\n");
    curPod->pcase = &M_SetupDynLitFogTex;
    curPod->plights = gDirPointFogLightList;
    curPod->pmaterial = &gTransMaterial2;}
		
	if (curPod->pcase == &M_SetupDynLitTex) {
	/* 3DO */
    printf ("Fogging 3DO sphere -> DynLitFogTex\n");
    curPod->pmaterial = &gTransMaterial2;
    curPod->plights = gDirPointFogLightList;}

	if (curPod->pcase == &M_SetupDynLitTex) {
	/* sky */
    printf ("Fogging Environmental sphere -> DynLitFogEnv\n");
    curPod->pcase = &M_SetupDynLitFogEnv;
    curPod->plights = gEnvFogLightList;
    curPod->pmaterial = &gTransMaterial2;}
	
	curPod=curPod->pnext;

	} /* while */

    fta[0].fta_String = "Level 2 - Fogged\nCollisions ";
    CreateTextState (&ts, font, fta, 1);

	/* sort them again */
    chasePod = M_Sort (numChasePods, chasePod, NULL);
    potPod = M_Sort (numPot, potPod, NULL);
    tankPod = M_Sort (numTank, tankPod, NULL);

  }
  M_End (app->close);
#if RUNMP
  M_EndMP (app->close);
#endif

  FreeMem(timesample1, sizeof (TimeVal) );
  FreeMem(timesample2, sizeof (TimeVal) );

  FreeMem(targetPod, sizeof (Pod) );

  FreeMem(myString, (35* sizeof (char) ) );
  FreeMem(myString2, (35* sizeof (char) ) );
  FreeMem(RString, (35* sizeof (char) ) );

}				/* main */



