
/*
 * 3do_scaf.c
 *
 *	Copyright:	 1996 by The 3DO Company. All rights reserved.
 *				This material constitutes confidential and proprietary
 *				information of the 3DO Company and shall not be used by
 *				any person or for any purpose except as expressly
 *				authorized in writing by the 3DO Company.
 *
 *	Revision History
 *
 *	YY/MM/DD	Name		Description
 *	--------	----------	--------------------------------------
 *	03/13/96	Bill Weir	Scaffold re-implementation, based on
 *					  		earlier components by Eric Hegstrom,
 * 					  		Eric Matsumo, Nick Triantos, and
 *					  		others
 *
 *  03/27/96    Bill Weir	Added interfaces to Eric Hegstrom's 
 *							MemDebug functionality (see 'scaf_*.*)
 *
 *  05/10/96	Bill Weir 	Added remote-logging
 *
 *  05/16/96 	Bill Weir	Limited stored length of test-program
 * 							name to assure that filenames of local
 *  						and remote logfiles are 31 characters 
 * 							or fewer
 *
 @(#) 3do_scaf.c 96/07/11 1.2
 */


#include <tests/3do_scaf.h>
#include "scaf_MemUtils.h"
#include <kernel/debug.h>
#include <file/fileio.h>
#include <file/filefunctions.h>
#include <misc/event.h>
#include <misc/script.h>


/* local function prototypes */

int32 CheckTestgroup(int32 testgroup);
int32 OpenLog(char * tag, char * stem, char * logname, RawFile ** logfileptr);
int32 OpenThisFile(RawFile **fileptr, char *filename, int32 filemode);
void CloseLog(char * tag);
int32 CopyLogFile(void);
int32 SetParameter(char * p);
int32 BooleanEval(char * string);
char * AppendString(char * origin, char * addloc, char * endloc, char * string);
void SetElapsedTime(void);
void WriteLog(char * buf);
void DisposeLog(char * buf, int32 len);
void HelpMessage(void);
void DiagMessage(char * messagestring);
void Error(char * errorstring);
void ScaffoldError(char * funcname);
void DisplayMessage(char * messagestring);
int32 ReadTimer(void);
static void WaitForNoButtonPresses(void);
#ifdef NOTUSED
int32 SimpleCopyLogFile(void);
#endif


/* parameters */

#define MAXLOGNO 999	/* maximum number <n> for log filename */
#define LBUFLEN	1024	/* length of buffer for log line */
#define EBUFLEN 256		/* length of buffer for error line */
#define RESLEN 64		/* length of buffer for resource-string */
#define MEMLEN 64		/* length of buffer for MemDebug or MemRation */
#define TNAMELEN 1024	/* length of buffer for test-name */

#define TSTAMPLEN 12	/* length of timestamp field */
#define RELEASELEN 15	/* length of release field */
#define TGROUPLEN 15	/* length of testgroup-name field */
#define ETIMELEN 15		/* length of elapsed-time field */
#define TTYPELEN 15		/* length of test-type field */

						/* definitions of control-pad for 'InteractiveVerify' */
#define kDPadMask		(ControlUp | ControlDown | ControlRight | ControlLeft)
#define kButtonMask		(ControlLeftShift | ControlRightShift | ControlX | \
							ControlStart | ControlA | ControlB | ControlC)
#define kShiftMask		(ControlLeftShift | ControlRightShift)
#define kPassedMask		(ControlA | ControlB | ControlC)
#define kFailedMask		(ControlX)
#define kInteractiveMask	(ControlStart)



/* local datastructures */

struct
{
	char timestamp[TSTAMPLEN + 1];		/* initialized by InitScaffold */
	char progname[32];					/* initialized by InitScaffold */	
	eTestGroup testgroup;				/* initialized by InitScaffold */
	char testgroupname[32];				/* initialized by InitScaffold */
	char release[RELEASELEN + 1];		/* initialized by InitScaffold */
	char globallogstem[64];				/* initialized by InitScaffold */
	char locallogstem[64];				/* initialized by ChangeLogName or InitScaffold */
	eActionCode logaction;				/* initialized by InitScaffold */
	int32 display;						/* initialized by InitScaffold */
	int32 file;							/* initialized by InitScaffold */
	int32 database;						/* initialized by InitScaffold */
	int32 deferlog;						/* initialized by InitScaffold */
	int32 loglevel;						/* initialized by InitScaffold */
	int32 priority;						/* initialized by InitScaffold */
	int32 semaphores;					/* initialized by InitScaffold */
	int32 memory;						/* initialized by InitScaffold */
	int32 processes;					/* initialized by InitScaffold */
	int32 items;						/* initialized by InitScaffold */
	char resourcestring[RESLEN + 1];	/* initialized by InitScaffold */
	char memdebug[MEMLEN + 1];			/* initialized by InitScaffold */
	char memration[MEMLEN + 1];			/* initialized by InitScaffold */
	int32 memcheck;						/* initialized by InitScaffold */
	int32 starttime;					/* initialized by InitScaffold */
	int32 pausetime;					/* initialized by InitScaffold */
	char testname[TNAMELEN + 1];		/* per test, added by BeginTest */
	int32 testtype;						/* per test, added by BeginTest */
	char testtypestring[TTYPELEN + 1];	/* per test, added by BeginTest 
											(padded to 15 characters) */
	char elapsedstring[ETIMELEN + 1];	/* per log-instance, added by 
											SetElapsedTime for 
											log-functions */

} session;						/* this may be extended by addition of members */


/* testgroup names, indexed by numeric identifer */

char *testgroups[10] =
{
	"", 
	"KERNEL",
	"FILESYSTEM",
	"AUDIO",
	"GRAPHICS_3D",
	"GRAPHICS_2D",
	"DATA_STREAM",
	"PERIPHERALS",
	"INTERNATIONAL",
	"GRAPHICS_CLT"
};

int32 initflag = (int32)0;		/* initialization flag, which is 0 before 
									initialization with 'InitScaffold', 
									remains 1 until 'TermScaffold', and is
									then set to -1 */
								
int32 newlogflag = (int32)0;	/* flag which is set to 1 if 'ChangeLogName' 
									is invoked before 'InitScaffold' */
	
char remotelogname[64];			/* name of unique remote-logfile 
									'<testgroup>_<testprogram>_<n>.log' */
RawFile *remotelogfile;			/* stream identifier of remote-logfile */

char globallogname[64];			/* name of global-logfile 'scaffold<n>.log' */
RawFile *globallogfile;			/* stream identifier of global-logfile */
int32 globalwriteflag;			/* flag to indicate whether global-logfile has
									received one or more log-entries */

char locallogname[64];			/* name of local-logfile '<progname><n>.log' or
									'<whatever><n>.log' where <whatever> is set
									by ChangeLogName */
RawFile *locallogfile;			/* stream identifier of local-logfile */
int32 localwriteflag;			/* flag to indicate whether local-logfile has
									received one or more log-entries */

char logbuf[LBUFLEN];			/* buffer for line to be logged */
char *dbuffer = (char *)NULL;	/* address of allocated buffer for deferred
									logging */
char *dbufferptr = (char *)NULL;/* pointer into 'dbuffer', to the location at
									which the next log-write is to occur for
									deferred logging */
char *dbufferover;				/* location at which overflow will occur in 
									writes to 'dbuffer' */
								
char dbtarget[128]; 			/* database-target directory specification */

char e[EBUFLEN];				/* error string */

int32 debug = (int32)0; 		/* debug flag, set to 1 to enable diagnostics
									or to 0 to disable diagnostics */ 



/* Scaffold API functions (called from testprograms) */

int32 InitScaffold(eTestGroup testgroup, int32 argc, char * argv[])
{
	int32 argerrors;	/* count of errors in processing arguments */
	int32 rflag;	  	/* flag to be set for required '-R' argument */
	char *s;
	char *t;
	char *p;
	int32 i;
	int32 j;
	int32 fill;
	Err retval;
	char commandline[512];

	/* check for previous initialization, and return 0 after 
		message is previously initialized */
	if(initflag > (int32)0) 
	{
		(void)sprintf(e, "%s %s",
			"encountered 'InitScaffold'",
			"after Scaffold already initialized");
		Error(e);
		return((int32)0);
	}
	
	/* buffer the command-line, with a space to separate
		arguments */
	s = commandline;
	for(i = (int32)0; i < argc; i++)
	{
		t = argv[i];
		while(*t != '\0') *s++ = *t++;
		*s++ = ' ';		
	}
	*(s-1) = '\0';
	
	/* set logfile stream-pointers NULL to indicate that the 
		logfiles are not open */
	globallogfile = (RawFile *)NULL;
	locallogfile = (RawFile *)NULL;
	
	/* set deferred-logging buffer's origin and current-location
		pointers NULL to force default to immediate logging */
	dbuffer = (char *)NULL;
	dbufferptr = (char *)NULL;
	
	/* initialize flags and counters */
	argerrors = (int32)0;
	rflag = (int32)0;

	/* set timestamp field to '-', because timestamp is not 
		available */
	s = session.timestamp;
	t = s + TSTAMPLEN;
	*s++ = '-';
	while(s < t) *s++ = ' ';
	*t = '\0';

	/* if the specified testgroup is not recognized, issue an 
		error-message and set flag to abort after argument-
		checking */
	if(CheckTestgroup(testgroup))
	{
		(void)sprintf(e,"ERROR: testgroup %d not recognized",
			testgroup);
		Error(e);
		argerrors++;
	}
	
	/* save the program-name, truncated to 23-character 
		length if necessary after removing prepended path-
		components; this limitation is to assure 
		that full local and remote log-names do not exceed
		31 characters even with addition of '_nnn.log' */
	s = argv[0];
	while((t = strchr(s, '/')) != (char *)NULL)
	{
		s = t + 1;
	}
	(void)strncpy(session.progname, s, (int32)31);
	if(strlen(session.progname) > (int32)23)
	{
		session.progname[23] = '\0';
		if(debug)
		{
			(void)sprintf(e, "%s '%s'",
				"program-name (truncated to 23-character length) is",
				&(session.progname[0]));
		}
	}
	else if(debug) 
	{
		(void)sprintf(e, "%s '%s'",
			"program-name (at full-length) is",
			&(session.progname[0]));
	}
	if(debug) DiagMessage(e);
		
	/* set default values for environment parameters */
	session.logaction = LOG;
	session.deferlog = (int32)0;
	session.display = (int32)1;
	session.file = (int32)1;
	session.database = (int32)1;
	session.loglevel = (int32)100;
	session.priority = (int32)100;
	session.resourcestring[0] = '\0';
	session.memdebug[0] = '\0';
	session.memration[0] = '\0';
	session.memcheck = (int32)(-1);
		
	/* specify database target directory, which is a folder in the 
		'remote' directory aliased to the remote-server database,
		such as 'Tools/The QA Server/QA Tools/RemoteDB/'; note that 
		this target-directory specification must end in '/', since
		the filename will be concatenated directly onto it */
	(void)strcpy(dbtarget, "/remote/ResultsDB/");
		
	/* begin scan of command-line arguments for required -R and
		optional -S arguments */
	if(debug)
	{
		for(i = (int32)1; i < argc; i++)
		{
			(void)sprintf(e, 
				"pre-scan argument%d is %s",
				i, &argv[i][0]);
			DiagMessage(e);
		}
	}
	for(i=(int32)1; i < argc; )
	{
		/* if an optional '-S' argument, identify the
			parameter and set its value; do not advance 
			'i', which will index the next argument after
			down-shuffle */
		if((argv[i][0] == '-') && (argv[i][1] == 'S'))
		{
			/* locate actual argument, seeking it first 
				directly following '-S'; set number of 
				slots to fill by subsequent down-shuffle */
			if(argv[i][2] != '\0') 
			{
				(void)strcpy(logbuf, &argv[i][2]);
				fill = (int32)1;
			}	
			else 
			{
				/* if next argument is missing or is a flag, 
					the parameter-assignement is certainly  
					missing, so error */
				if((i > (argc - 1)) || (argv[i + 1][0] == '-'))
				{
					argerrors++;
					logbuf[0] = '\0';
					fill = (int32)1;
				}
					
				/* otherwise, the next argument may be the
					parameter-assignment, and since this cannot
					be checked more conclusively, point to the 
					argument and set to shuffle higher-indexed 
					arguments down 2 slots */
				else
				{
					(void)strcpy(logbuf, argv[i + 1]);
					fill = (int32)2;
				}
			}
				
			/* identify parameter name and value, set 
				the parameter's value; for "MemDebug" 
				parameters, 'SetParameter' will also 
				invoke the initialization function */
			if(logbuf[0] != '\0')
			{
				if(SetParameter(logbuf)) argerrors++;
			}
			
			/* shuffle arguments */
			for(j=i+fill; j < argc; j++)
			{
				argv[j-fill] = argv[j];
			}
			argc -= fill;
				
		}  /* end of processing for -S flag */			

		/* if the required '-R' argument, identify the
			release; do not advance 'i', which will 
			index the next argument after down-shuffle */
		else if((argv[i][0] == '-') && (argv[i][1] == 'R'))
		{
			/* warn if -R flag is repeated; this will not
				cause return-value 0, and the last-appearing
				release-specification will be used */
			if(rflag)
			{
				Error("found repeated '-R' argument");
			}
				
			/* locate actual argument, seeking it first 
				directly following '-R'; set number of 
				slots to fill by subsequent down-shuffle */
			if(argv[i][2] != '\0') 
			{
				/* point to argument and set to shuffle 
					higher-indexed arguments down 1 slot */
				s = &argv[i][2];
				fill = (int32)1;
			}	
			else 
			{
				/* if next argument is missing or is a flag, 
					the release-identifier is certainly missing, 
					so error */
				if((i >= (argc - 1)) || (argv[i+1][0] == '-'))
				{
					argerrors++;
					s = (char *)NULL;
					fill = (int32)1;
				}
					
				/* otherwise, the next argument may be a 
					release-identifier, and since this cannot
					be checked more conclusively, point to 
					argument and set to shuffle higher-indexed 
					arguments down 2 slots */
				else
				{
					s = argv[i + 1];
					fill = (int32)2;
				}
			}
				
			/* if release-identifier is present, check its
				length and store it if not too long */
			if(s != (char *)NULL)
			{
				/* determine the identifier length */
				j = strlen(s);
					
				/* if too long, error */
				if(j > RELEASELEN)
				{
					(void)sprintf(e, "%s '%s' %s",
						"the -R argument", s, 
						"is too long -- argument ignored");
					Error(e);
					argerrors++;
				}

				/* if short enough, count and store it;
					pad release-identifier with spaces to 
					full width of field */
				else
				{
					rflag++;
					p = session.release;
					t = p + RELEASELEN;
					while(*s != '\0') *p++ = *s++;
					while(p < t) *p++ = ' ';
					*p = '\0';
				}
			}
				
			/* shuffle remaining arguments */
			if(debug)
			{
				(void)sprintf(e, 
					"shuffling by %d starting with arg%d",
					fill, i + fill);
				DiagMessage(e);
			}
			for(j=i+fill; j < argc; j++)
			{
				if(debug)
				{
					(void)sprintf(e,
						"moving '%s' from index %d to index %d",
						argv[j], j, j - fill);
					DiagMessage(e);
				}
				argv[j-fill] = argv[j];
			}
			argc -= fill;
				
		} /* end of processing for -R flag */
			
		/* if any other flag argument, or if non-flag argument, 
			skip by advancing index 'i'; note that for -R and -S 
			flags, processed above, index does not advance because
			arguments have been shuffled so next argument has the 
			same index as the -R or -S argument */
		else i++;
	}
	if(debug)
	{
		for(i = (int32)1; i < argc; i++)
		{
			(void)sprintf(e, 
				"post-scan argument%d is %s", 
				i, &argv[i][0]);
			DiagMessage(e);
		}
	}

	/* when all arguments have been scanned, check for the
		required '-R' flag; if this flag was absent, issue 
		an error message and exit from parent test-program */
	if(!rflag)
	{
		(void)sprintf(e, "%s -- %s", 
			"required -R flag is absent", 
			"cannot proceed with test-program");
		Error(e);
		exit(1);
	}
			
	/* if the '-R' flag was present but argument-errors
		occurred, issue an error message and exit from parent 
		test-program */
	if(argerrors)
	{
		(void)sprintf(e, "%s -- %s", 
			"argument-errors were present", 
			"cannot proceed with test-program");
		Error(e);
		exit(1);
	}

	/* if the '-R' flag was present and no argument-errors
		occurred, proceed to open a unique local logfile, 
		and ultimately return the updated value of 'argc';
		set default name-stem for global-log */
	(void)strcpy(session.globallogstem, "scaffold");
			
	/* attempt to open global-log, with the uniform stem 
		specified by 'session.globallogstem'; the locally-
		unique name is returned in 'globallogname', the
		file-descriptor in 'globallogfile' if successful;
		for failure, the error-count is increased and 
		will be trapped below */
	if(OpenLog("global", session.globallogstem, 
		globallogname, &globallogfile))
			argerrors++;
			
	/* if global-log was successfully opened, attempt to 
		open local-log; if the file 'session.locallogstem'
		exists, it will be opened for appending; if that 
		file does not exist, a unique file will be created 
		for this log, by appending <n> to the stem */
	else
	{
		/* if local-log's name-stem has been specified by 
			'ChangeLogName', prior to initialization, or by 
			including a '-Slogstem' flag on the test-program's 
			command-line, 'session.locallogstem' will have 
			been set already; if not, use the default, which  
			is the program-name, as stem; the stem is limited
			to 23-character length to assure that the full
			local or remote logname will not exceed 31 
			characters */
		if(!newlogflag) (void)strcpy(session.locallogstem, 
										session.progname);
										
		/* attempt to open the local-log, with the stem 
			'session.locallogstem'; the locally-unique name is 
			returned in 'locallogname', the file-descriptor 
			in 'locallogfile', if successful; for failure, 
			the error-count is increased and will be trapped 
			below */
		if(OpenLog("local", session.locallogstem, 
							locallogname, &locallogfile))
			argerrors++;
	}
	
	/* if a log-opening error occurred, issue an 
		error-message and exit from parent test-program */
	if(argerrors)
	{
		(void)sprintf(e, "%s -- %s", 
			"log-opening errors were present", 
			"cannot proceed with test-program");
		Error(e);
		exit(1);
	}
	
	/* if no log-opening errors occurred, set 'initflag' to 1  
		for other functions and prepare to return the proper 
		argument-count */
	initflag = (int32)1;
	
	/* create logstring including command-line, and write 
		to log */
	(void)sprintf(logbuf, "TEST-PROGRAM:  %s", commandline); 
	WriteLog(logbuf);
		
	/* initialize the timer; if the return-value is 0, the
		timer did not initialize, and this is indicated by 
		both 'starttime' and '.pausetime' having value 0; 
		if the return-value is non-zero, initialization was 
		successful, and the returned value is used as an 
		offset by 'SetElapsedTime' */
	session.pausetime = (int32)0;
	session.starttime = ReadTimer();

	/* return the updated value of 'argc' after squeezing out 
		'-R' and '-S' flags; note that the test-program exited  
		above if required '-R' flag is absent, if unrecoverable  
		argument errors occurred, or if a log could not be
		opened */
	return(argc);
}


void TermScaffold(void)
{
	int32 nodelete;
	char tgroupname[32];
	char dbstemname[64];
	
	if(initflag <= (int32)0) 
	{
		ScaffoldError("TermScaffold");
		return;
	}
	
	/* if logging was deferred, write the buffer's accumulated 
		contents to the global log, and to the local log if enabled, 
		then free the buffer */ 
	if(dbufferptr != (char *)NULL) 
	{			
		DisposeLog(dbuffer, (int32)(dbufferptr - dbuffer));
		(void)free(dbuffer);
	}

	/* close the global-logfile and the local-logfile; the local-
		logfile will be deleted if it was not written */
	CloseLog("global");
	CloseLog("local");
	/* note that 'globallogname' may be used, and that 'locallogname'
		may be reused, for subsequent copy to remote */
		
	/* write to the persistent database if enabled at this point, if 
		global-logfile exists, and if any log-entries were written */
	if(session.database && 
		(globallogfile != (RawFile *)NULL) && 
			globalwriteflag)
	{
		/* create compressed testgroup-name, by trimming padding
			spaces which were added for logging */
		(void)strcpy(tgroupname, session.testgroupname);
		*(strchr(tgroupname, ' ')) = '\0';

		/* 'dbtarget', which was set by 'InitScaffold', names the  
			standard alias-folder for the remote logging database; 
			the destination stem-name of the copied file is 
			'<dbtarget><testprogram-name>', where <dbtarget> is
			a full path such as '/remote/ResultsDB/'; the string
			'_<n>.log' will be appended to this by 'OpenLog' */
		(void)sprintf(dbstemname, "%s%s", dbtarget, session.progname); 

		/* assign a unique destination-name: 'OpenLog' with "remote"
			argument attempts to open the log with successive trials 
			with ascending <n>, but does not then open the new file
			for writing, because the file will be created by copy;
			'remotelogname' receives the unique name and
			'remotelogfile' is ignored until the file is opened 
			for writing */
		if(OpenLog("remote", dbstemname, remotelogname, &remotelogfile))
		{
			/* if this process fails, set flag to assure that the 
				global logfile will not be deleted */
			(void)sprintf(e, "%s '%s' %s -- %s",
				"log-entries in file", globallogname, 
				"were not accessible for copy to the results database",
				"please inform your logging administrator");
			Error(e);
			nodelete = (int32)1;
		}
	
		/* if unique destination-name was set, attempt to copy the file */
		else
		{
			/* if copy fails, flag is set to inhibit delete of source-file;
				if copy succeeds, flag is clear to allow delete */ 
			nodelete = CopyLogFile();
			
			/* message if copy failed */
			if(nodelete)
			{
				(void)sprintf(e, "%s '%s' %s -- %s",
					"log-entries in file", globallogname, 
					"were not being copied to the results database",
					"please inform your logging administrator");
				Error(e);
			}
		}
	}
	
	/* if file exists, but if no log entries were written to it or if
		logging to results database is not enabled, allow deletion of  
		the file */
	else if(globallogfile != (RawFile *)NULL) nodelete = (int32)0;
	
	/* if file does not exist, there is no need to delete it, so inhibit */
	else nodelete = (int32)1;
	
	/* delete global-logfile if it exists, unless error occurred in copying */
	if(!nodelete)
	{
		DeleteFile(globallogname);
		if(debug)
		{
			(void)sprintf(e, "%s '%s'",
				"deleted global-logfile", globallogname);
			DiagMessage(e);
		}
	}

	/* if MemDebug functionality was initialized, terminate it now */
	if(session.memdebug[0] != '\0') TermMemDebug();
	if(session.memration[0] != '\0') TermMemRation();
	if(session.memcheck != (int32)(-1)) TermMemCheck();
	
	/* set the initialization-flag to -1 to indicate that 
		Scaffold was open but has now been closed */
	initflag = (int32)(-1);
	
	return;
}

void BeginTest(int32 test_type, char * test_name, ...)
{
	char *s;
	char *t;
	va_list args;
	
	if(initflag <= (int32)0) 
	{
		ScaffoldError("BeginTest");
		return;
	}
	
	/* create the timestamp; initially this is simply '-', 
		but later an absolute timestamp will be provided;
		note that this timestamp must be of length 
		TSTAMPLEN characters */	
	(void)strcpy(session.timestamp, "-           ");
	
	/* re-create string to describe test-type(s) in its buffer 
		(TTYPELEN-character field-length) */
	s = session.testtypestring;
	*s = '\0';
	t = s + TTYPELEN;
	if(test_type & BAT) 
		s = AppendString(session.testtypestring, s, t, "BAT");
	if(test_type & FT) 
		s = AppendString(session.testtypestring, s, t, "FT");
	if(test_type & EST) 
		s = AppendString(session.testtypestring, s, t, "EST");
	if(test_type & DT) 
		s = AppendString(session.testtypestring, s, t, "DT");
	if(test_type & LST) 
		s = AppendString(session.testtypestring, s, t, "LST");
	if(test_type & BT) 
		s = AppendString(session.testtypestring, s, t, "BT");
	if(test_type & PT) 
		s = AppendString(session.testtypestring, s, t, "PT");
	if(test_type & BREADTH) 
		s = AppendString(session.testtypestring, s, t, "BREADTH");
	if(test_type & DEPTH) 
		s = AppendString(session.testtypestring, s, t, "DEPTH");
	if(test_type & HIGH_RISK) 
		s = AppendString(session.testtypestring, s, t, "RISK");
	if(test_type & IT) 
		s = AppendString(session.testtypestring, s, t, "IT");
	while(s < t) *s++ = ' ';
	*t = '\0';
	if(debug)
	{
		(void)sprintf(e, "%s '%s'",
			"generated test-type string", 
			session.testtypestring);
		DiagMessage(e);
	}

	/* copy test-name ('printf' format) to its buffer (indefinite 
		field-length) */
	va_start(args, test_name);
	if(args == NULL) (void)strcpy(session.testname, test_name);
	else (void)vsprintf(session.testname, test_name, args);
	va_end(args);
	if(debug)
	{
		(void)sprintf(e, "%s '%s'",
			"generated test-name string", 
			session.testname);
		DiagMessage(e);
	}
	
	/* create the logstring; '.testgroupname' has a 
		built-in trailing pad of two spaces */	
	if(debug)
	{
		(void)sprintf(e, "%s '%s'",
			"BeginTest test-name string is", 
			session.testname);
		DiagMessage(e);
		(void)sprintf(e, "%s '%s'",
			"BeginTest release string is", 
			session.release);
		DiagMessage(e);
/* PREVIOUS
		(void)sprintf(e, "%s '%s'",
			"BeginTest timestamp string is", 
			session.timestamp);
		DiagMessage(e);
		(void)sprintf(e, "%s '%s'",
			"BeginTest test-type string is", 
			session.testtypestring);
		DiagMessage(e);
		(void)sprintf(e, "%s '%s'",
			"BeginTest resource string is", 
			session.resourcestring);
		DiagMessage(e);
*/
	}
	(void)sprintf(logbuf, "%s<%s>  %s",
		session.testgroupname, session.testname, session.release);
/* PREVIOUS
	(void)sprintf(logbuf, "<%s>\t%s%s%s%s",
		session.testname, session.release, session.timestamp, 
		session.testtypestring, session.resourcestring);
*/

	/* write logstring to log */
	WriteLog(logbuf);
	
	/* reinitialize the timer; if both 'starttime' and 
		'.pausetime' have value 0, the timer did not 
		initialize so this need not be re-attempted; 
		if the return-value is non-zero, initialization was 
		successful, and the returned value is used as an 
		offset by 'SetElapsedTime' */
	if(session.pausetime || session.starttime)
	{
		session.pausetime = (int32)0;
		session.starttime = ReadTimer();
	}

	return;
}


void PauseTimer(void)
{
	if(initflag <= (int32)0) 
	{
		ScaffoldError("PauseTimer");
		return;
	}

	/* act only if timer is functioning and not already paused */
	if(session.starttime && !session.pausetime)
	{
		/* set <current-time> - '.starttime' into '.pausetime', 
			to record the elapsed time from 'BeginTest' to 
			'PauseTimer'; this value will be used by 'ResumeTimer'
			to reset '.starttime' */ 
		session.pausetime = ReadTimer();
		session.pausetime -= session.starttime;
	
		/* reset '.starttime' to 0 to assure that '-' will be
			assigned instead of an elapsed-time, while paused */
		session.starttime = (int32)0;
	}
			
	return;
}


void ResumeTimer(void)
{
	if(initflag <= (int32)0) 
	{
		ScaffoldError("ResumeTimer");
		return;
	}
	
	/* act only if timer is functioning and paused */
	if(!session.starttime && session.pausetime)
	{
		/* set <current> - '.pausetime' into '.starttime', to
			reset the effective start-time, allowing for the
			interval paused */
		session.starttime = ReadTimer();
		session.starttime -= session.pausetime;
		
		/* reset '.pausetime' to 0 to assure that elapsed-time
			values, and not '-', will be printed while no longer
			paused */
		session.pausetime = (int32)0;
	}
	
	return;
}


void ChangeResource(int32 resource_identifier, int32 quantum)
{
	if(initflag <= (int32)0) 
	{
		ScaffoldError("ChangeResource");
		return;
	}
	
	DisplayMessage("'ChangeResource' is not yet implemented");
	return;
}


/*	Control-pad actions for 'InteractiveVerify' are as follows: 
 *
 *		Result			Press
 *		------------- 	--------------------------------------------
 *		PASSED			Shift-A, Shift-B, or Shift-C (either shift)
 *		FAILED			Shift-X (either shift)
 *		UNRESOLVED		Shift-Play (either shift)
 *
 *
 *  this function returns negative values for error-conditions, 
 *  otherwise positive values for results
 */

eTestResult InteractiveVerify(char *prompt_string, ...)
{
	Err theErr;
	eTestResult result; 
	ControlPadEventData theCPED;
	Boolean evtUtilWasInited;
	Boolean gotValidButtonCombo;
	int32 bits;
	va_list args;

	/* set indefinite-result as default return-value */
	result = (eTestResult)UNRESOLVED;
	
	/* check for proper initialization */
	if(initflag <= (int32)0) 
	{
		ScaffoldError("InteractiveVerify");
		return(result);
	}
	
	/* use 'logbuf' to contain prompt ('printf' format) */
	va_start(args, prompt_string);
	if(args == NULL) (void)strcpy(logbuf, prompt_string);
	else (void)vsprintf(logbuf, prompt_string, args);
	va_end(args);
	
	/* issue prompt to User */
	(void)printf("%s\n", logbuf);

	/* determine if the event-broker is already attached, 
	    so it can be left attached at the conclusion of 
		this function rather than being detached */	
	if(KillEventUtility() < 0) evtUtilWasInited = FALSE;
	else evtUtilWasInited = TRUE;

	/* attach or reattach now, with error return if unable to
		do so */		
	theErr = InitEventUtility(1, 0, FALSE);
	if (theErr < 0) return (eTestResult)theErr;

	/* wait for quiescence */	
	WaitForNoButtonPresses();
	
	/* get result from control-pad */
	gotValidButtonCombo = FALSE;
	do {
		theErr = GetControlPad(1, TRUE, &theCPED);
		bits = theCPED.cped_ButtonBits;
		if (theErr < 0)
			return (eTestResult)theErr;
		bits &= kButtonMask;
		if (((bits & kShiftMask) != 0) &&
			((bits & kPassedMask) != 0) &&
			((bits & (kShiftMask | kPassedMask)) == bits)) 
		{
			gotValidButtonCombo = TRUE;
			result = PASSED;
		}
		if (((bits & kShiftMask) != 0) &&
			((bits & kFailedMask) != 0) &&
			((bits & (kShiftMask | kFailedMask)) == bits)) 
		{
			gotValidButtonCombo = TRUE;
			result = FAILED;
		}
		if (((bits & kShiftMask) != 0) &&
			((bits & kInteractiveMask) != 0) &&
			((bits & (kShiftMask | kInteractiveMask)) == bits))
		{
			gotValidButtonCombo = TRUE;
			result = (eTestResult)UNRESOLVED;
		}
		
	} while (!gotValidButtonCombo);

	/* again wait for quiescence */	
	WaitForNoButtonPresses();
	
	/* if event-broker was not attached at entry to this 
		function, detach it now, otherwise leave it attached */	
	if (evtUtilWasInited == FALSE) (void)KillEventUtility();

	/* return the retrieved result */	
	return(result);
}


/* note that log-path stem can be changed only by invoking this function 
	BEFORE invoking 'InitScaffold'; following 'InitScaffold', this function
	provides an error message */

void ChangeLogName(char * log_path)
{
	/* if not yet initialized, set the new log-name stem and set a 
		flag to assure that the default will not be used */
	if(!initflag)
	{
		(void)strncpy(session.locallogstem, log_path, (int32)31);
		session.locallogstem[23] = '\0';
		newlogflag = (int32)1;
	}
	
	/* if initialized but not yet terminated, close the old log,
		and open the new one; remove the old log if but only if 
		nothing was written to it */
		
	else if (initflag == (int32)1)
	{
		/* close the old local-log; if it was not written, delete it */
		CloseLog("local");
		
		/* save the new stem */ 	
		(void)strncpy(session.locallogstem, log_path, (int32)31);
		session.locallogstem[23] = '\0';
		
		/* open the new log */
		(void)OpenLog("local", session.locallogstem, locallogname, &locallogfile);
	}
	
	/* if already terminated, it is too late to change log-name */
	else
	{
		(void)sprintf(e, "%s '%s' %s",
			"encountered Scaffold-function", "ChangeLogName",
			"after Scaffold termination");
		Error(e);
	}
	
	return;
}


void ChangeLogTargets(Boolean display, Boolean file, Boolean database)
{
	if(initflag <= (int32)0) 
	{
		ScaffoldError("ChangeLogTargets");
		return;
	}
	
	/* set 1 or 0 to correspond to Boolean for logging to display; note 
		that this value is effective at the point of invocation of each 
		logging function, causing that log-entry to be written to the 
		display, or not */
	if(display == TRUE) session.display = (int32)1;
	else session.display = (int32)0;
	
	/* set 1 or 0 to correspond to Boolean for logging to file; note that
		this value is effective at the point of invocation of each logging 
		function, causing that log-entry to be written to file, or not; at
		change of log with 'ChangeLogName' and at 'TermScaffold', the log
		is deleted if nothing has been written to it */
	if(file == TRUE) session.file = (int32)1;
	else session.file = (int32)0;
	
	/* set 1 or 0 to correspond to Boolean for logging to database;
		note that this value is effective at the point of invocation of 
		each logging function, to determine whether or not its log-entry is
		included in the collection of entries which are candidates for 
		database logging, and then when 'TermScaffold' is invoked, to 
		determine whether the collection of log-entries is copied to the 
		logging server or not */
	if(database == TRUE) session.database = (int32)1;
	else session.database = (int32)0;
	
	return;
}


void ChangeLogAction(eActionCode action_code)
{
	if(initflag <= (int32)0) 
	{
		ScaffoldError("ChangeLogAction");
		return;
	}
	
	if((action_code < 1) || (action_code > 3))
	{
		Error("specified 'action_code' is out-of-range");
	}
	else session.logaction = (int32)action_code;
	return;
}


void DeferLog(int32 log_bytes)
{
	if(initflag <= (int32)0) 
	{
		ScaffoldError("DeferLog");
		return;
	}

	/* allocate a buffer containing the specified number of bytes */
	dbuffer = malloc(log_bytes);
	
	/* if the allocation failed, give an error message and set
		the current-location pointer to the buffer origin; a
		NULL value will force immediate logging */
	if(dbuffer == (char *)NULL)
	{
		(void)sprintf(e, "%s -- %s",
			"failed to allocate buffer for deferred-logging",
			"immediate-logging will occur");
		Error(e);
		dbufferptr = (char *)NULL;
	}
	
	/* if the allocation succeeded, set the current-location 
		pointer to the buffer origin; a non-NULL value will 
		force deferred logging; the buffer-size is saved for
		reference in 'session.deferlog' */
 	else
	{
		dbufferptr = dbuffer;
		dbufferover = dbuffer + log_bytes - 1;
		session.deferlog = log_bytes;
	}
	return;
}


void LogError(int32 errno)
{
	/* dimension error-string one more than result-string to allow for
		'ERROR=' versus 'RESULT=' */
	char errnostring[14];
	char *s;
	char *t;
	int32 len;

	if(initflag <= (int32)0) 
	{
		ScaffoldError("LogError");
		return;
	}
	
	/* set the elapsed-time for this logging instance */
	SetElapsedTime();
	
	/* create padded string representing hexadecimal error-number */
	s = errnostring;
	t = errnostring + 14;
	(void)sprintf(s, "0x%08x", errno);
	s = strchr(errnostring, '\0');
	while(s < t) *s++ = ' ';
	*t = '\0';
	
	/* retrieve the message associated with 'errno' */
	len = GetSysErr(e, (EBUFLEN - 1), errno);
	e[len] = '\0';
	
	/* create the logstring and write it to the log; 
		'.elapsedstring' is right justified in the first 
		field, providing a substantial indent */
	(void)sprintf(logbuf, "%sERROR=%s%s%s%s\r",
		session.elapsedstring, errnostring, session.release, 
		session.testtypestring, session.testname);
	WriteLog(logbuf);
	
	/* copy a blank elapsed-time field to logstring; note 
		that the  number of spaces must be ETIMELEN + 4 */
	(void)sprintf(logbuf, "               %s", e);
	WriteLog(logbuf);
	
	return;
}


void LogResult(eTestResult result, char * string, ...)
{
	/* dimension result-string one less than error-string to allow for
		'RESULT=' versus 'ERROR=' */
	char resultstring[13];
	char *s;
	char *t;
	va_list args;

	if(initflag <= (int32)0) 
	{
		ScaffoldError("LogResult");
		return;
	}
	
	/* set the elapsed-time for this logging instance */
	SetElapsedTime();
	
	/* create padded string corresponding to integer result */
	s = resultstring;
	t = resultstring + 13;
	if(result == PASSED) (void)strcpy(s, "PASSED");
	else if(result == FAILED) (void)strcpy(s, "FAILED"); 
	else if(result == UNRESOLVED) (void)strcpy(s, "UNRESOLVED"); 
	else if(result == SETUP_FAILED) (void)strcpy(s, "SETUP_FAILED"); 
	else if(result == SKIPPED) (void)strcpy(s, "SKIPPED"); 
	else if(result == UNAVAILABLE) (void)strcpy(s, "UNAVAILABLE");
	s = strchr(resultstring, '\0');
	while(s < t) *s++ = ' ';
	*t = '\0';

	/* create a logstring containing the result, and log it; 
		'.elapsedstring' is right justified in the first 
		field, providing a substantial indent */
	(void)sprintf(logbuf, "%sRESULT=%s%s%s%s", 
		session.elapsedstring, resultstring, session.release, 
		session.testtypestring, session.testname);
	WriteLog(logbuf);

	/* if a comment is included, create a logstring containing 
		the comment ('printf' format) and log it */
	if((string != (char *)NULL) && (*string != '\0'))
	{
		/* copy a blank elapsed-time field to logstring; note 
			that the  number of spaces must be ETIMELEN + 4 */
		(void)strcpy(logbuf, "               ");
		s = strchr(logbuf, '\0');

		/* append the comment string ('printf' format) 
			and log it */
		va_start(args, string);
		if(args == NULL) strcpy(s, string);
		else vsprintf(s, string, args);
		va_end(args);
		WriteLog(logbuf);
	}
	
	return;
}


void LogComment(char * string, ...)
{
	char *s;
	va_list args;
	
	if(initflag <= (int32)0) 
	{
		ScaffoldError("LogComment");
		return;
	}
	
	/* set the elapsed-time for this logging instance */
	SetElapsedTime();
	
	/* copy elapsed-time field to logstring; '.elapsedstring'
		 is right justified, providing a substantial indent */
	(void)sprintf(logbuf, "%s", 
		session.elapsedstring);
	s = strchr(logbuf, '\0');

	/* append comment field to logstring ('printf' format) */
	va_start(args, string);
	if(args == NULL) strcpy(s, string);
	else vsprintf(s, string, args);
	va_end(args);
	
	/* write logstring to log */
	WriteLog(logbuf);
	
	return;
}


void LogDiagnostic(int32 level, char * string, ...)
{
	char *s;
	va_list args;

	if(initflag <= (int32)0) 
	{
		ScaffoldError("LogDiagnostic");
		return;
	}
	
	/* check level in allowed range 0-100 */
	if((level < (int32)0) || (level > (int32)100))
	{
		(void)sprintf(e, "%s -- %s",
			"specified 'level' must be in range 0-100",
			"diagnostic not logged");
		Error(e);
	}
	
	/* log only if in-range 'level' exceeds environment's 
		'loglevel' */
	else if(level > session.loglevel)
	{
		/* set the elapsed-time for this logging instance */
		SetElapsedTime();
	
		/* copy elapsed-time field to logstring; '.elapsedstring'
			 is right justified, providing a substantial indent */
		(void)sprintf(logbuf, "%s", session.elapsedstring);
		s = strchr(logbuf, '\0');

		/* append comment field to logstring ('printf' format) */
		va_start(args, string);
		if(args == NULL) strcpy(s, string);
		else vsprintf(s, string, args);
		va_end(args);
	
		/* write logstring to log */
		WriteLog(logbuf);
	}	
	return;
}


void ChangeLogLevel(int32 loglevel)
{
	if(initflag <= (int32)0) 
	{
		ScaffoldError("ChangeLogLevel");
		return;
	}
	
	if((loglevel < 0) || (loglevel > 100))
	{
		(void)sprintf(e, "%s -- %s",
			"requested 'loglevel' must be in range 0-100",
			"not altered");
		Error(e);
	}
	else session.loglevel = loglevel;
	return;
}


void DisplayUnconditional(char * string, ...)
{
	char *s;
	va_list args;

	if(initflag <= (int32)0) 
	{
		ScaffoldError("DisplayUnconditional");
		return;
	}
	
	/* set the elapsed-time for this logging instance */
	SetElapsedTime();
	
	/* copy elapsed-time field to logstring */
	(void)sprintf(logbuf, "%s", session.elapsedstring);
	s = strchr(logbuf, '\0');

	/* write string to logstring ('printf' format) */
	va_start(args, string);
	if(args == NULL) strcpy(s, string);
	else vsprintf(s, string, args);
	va_end(args);
	
	/* write logstring to display */
	DisplayMessage(logbuf);		
		
	return;
}


void DisplayConditional(char *string, ...)
{
	char *s;
	va_list args;

	if(initflag <= (int32)0) 
	{
		ScaffoldError("DisplayConditional");
		return;
	}
	
	/* make display conditional on logging enabled to display */
	if(session.display)
	{
		/* set the elapsed-time for this logging instance */
		SetElapsedTime();
	
		/* copy elapsed-time field to logstring */
		(void)sprintf(logbuf, "%s", session.elapsedstring);

		/* write string to logstring ('printf' format) */
		va_start(args, string);
		if(args == NULL) strcpy(logbuf, string);
		else vsprintf(logbuf, string, args);
		va_end(args);
	
		/* write logstring to display */
		DisplayMessage(logbuf);		
	}
	
	return;
}


/* function 'SetResultExpected' is NOT used in tests, as are other
	Scaffold functions, but rather in utilities which set the 
	result currently expected for a test */
	
void SetResultExpected(int32 testgroup, char * testname, eTestResult result)
{
	DisplayMessage("'SetResultExpected' is not yet implemented");
	return;
}



/* MEMORY-DIAGNOSTIC FUNCTIONS */


/* wrapper for 'StartMemDebug', provided for consistency with other 
	'InitXXX functions */
	
void InitMemDebug(char * flag_string)
{
	StartMemDebug(flag_string);
	return;
}


/* wrapper for 'StartMemRation', provided for consistency with other 
	'InitXXX functions */
	
void InitMemRation(char * flag_string)
{
	StartMemRation(flag_string);
	return;
}

/* the other functions of this group are defined in modules 'MemDebug.c',
	'MemRation.c' and 'MemUtils.c' */
	


/* UTILITY FUNCTIONS */


/* check testgroup-identifier specified by 'testgroup'; return 0 if 
	recognized, with the identifier in 'session.testgroup' and the
	testgroup-name in 'session.testgroupname'; return 1 if not
	recognized */

int32 CheckTestgroup(int32 testgroup)
{
	int32 retval;
	char *s;
	char *t;
	
	/* if testgroup-identifier is not within range, return 1 */
	if((testgroup < KERNEL) || (testgroup > GRAPHICS_CLT)) 
		retval = (int32)1;
		
	/* if testgroup-identifier is within range, store it and
		the testgroup-name */
	else
	{
		if(debug)
		{
			(void)sprintf(e, "%s '%s'",
				"testgroup is identified as",
				testgroups[testgroup]);
			DiagMessage(e);
		}
		session.testgroup = testgroup;
		(void)strcpy(session.testgroupname, testgroups[testgroup]);
		s = strchr(session.testgroupname, '\0');
		t = session.testgroupname + TGROUPLEN;
		while(s < t) *s++ = ' ';
		*t = '\0';
		retval = (int32)0;
	}
	
	return(retval);
}


/* SetParameter: for the parameter-assignment expression pointed to 
	by 'p', in the form '<parameter>=<value>', assign the value 
	to the parameter */

 int32 SetParameter(char *p)
{
	int32 eflag;	/* local error-flag */
	char *v;		/* pointer to value in string */
	char *q;
	int32 n;
	int32 *nptr; 

	if(debug) 
	{
		(void)sprintf(e, "entering SetParameter with '%s'", p);
		DiagMessage(e);
	}
	
	/* clear local error-flag */
 	eflag = (int32)0;
	
	/* find '=' separating parameter-name and value */
	v = strchr(p, '=');
	if(v == (char *)NULL) 
	{
		Error("parameter-expression lacks '='");
		eflag++;
	}
	else
	{
		/* terminate parameter-name, pointed to by 'p', over '=',
			and point to value pointed with 'v' */
		*v++ = '\0';
		if(!strcmp(p, "logaction"))
		{
			if(!strcmp(v, "LOG")) session.logaction = LOG;
			else if(!strcmp(v, "NOLOG")) 
				session.logaction = NOLOG;
			else if(!strcmp(v, "QUERY")) 
				session.logaction = QUERY;
			else
			{
				(void)sprintf(e, "'%s' %s %s",
					p, "parameter value must be",
					"'LOG', 'NOLOG', or 'QUERY'");
				Error(e);
				eflag++;
			} 
		}
		else if(!strcmp(p, "display"))
		{
			n = BooleanEval(v);
			if(n >= (int32)0) session.display = n;
			else
			{
				(void)sprintf(e, "'%s' %s %s",
					p, "parameter value must be",
					"'TRUE' or 'FALSE'");
				Error(e);
				eflag++;
			} 
		}
		else if(!strcmp(p, "file"))
		{
			n = BooleanEval(v);
			if(n >= (int32)0) session.file = n;
			else
			{
				(void)sprintf(e, "'%s' %s %s",
					p, "parameter value must be",
					"'TRUE' or 'FALSE'");
				Error(e);
				eflag++;
			} 
		}
		else if(!strcmp(p, "database"))
		{
			n = BooleanEval(v);
			if(n >= (int32)0) session.database = n;
			else
			{
				(void)sprintf(e, "'%s' %s %s -- %s",
					p, "parameter value must be",
					"'TRUE' or 'FALSE'",
					"value not changed");
				Error(e);
				eflag++;
			} 
		}
		else if(!strcmp(p, "deferlog"))
		{
			q = v; 
			while((*q != '\0') && (isdigit(*q))) q++;
			if((q != v) && (*q == '\0'))   
			{
				n = atoi(v);
				if(n <= (int32)0)
				{
					(void)sprintf(e, "'%s' %s %s -- %s",
						p, "parameter value must be",
						"greater than 0",
						"value not changed");
					Error(e);
					eflag++;
				}
				else session.deferlog = n;
			}
			else
			{
				(void)sprintf(e, "'%s' %s -- %s",
					p, "parameter value must be numeric",
					"value not changed");
				Error(e);
				eflag++;
			} 
		}
		else if(!strcmp(p, "logfile") || !strcmp(p, "logstem"))
		{
			/* both 'logfile' and 'logstem' are recognized */
			(void)strncpy(session.locallogstem, v, (int32)31);
			session.locallogstem[23] = '\0';
			newlogflag = (int32)1;

		}
		else if(!strcmp(p, "loglevel"))
		{
			q = v; 
			while((*q != '\0') && (isdigit(*q))) q++;
			if((q != v) && (*q == '\0'))   
			{
				n = atoi(v);
				if((n < (int32)0) || (n > (int32)100))
				{
					(void)sprintf(e, "'%s' %s %s -- %s",
						p,"parameter must be",
						"in range 0-100",
						"value not changed");
					Error(e);
					eflag++;
				}
				else session.loglevel = n;
			}
			else
			{
				(void)sprintf(e, "'%s' %s %s",
					p, "parameter value must be",
					"numeric -- value not changed");
				Error(e);
				eflag++;
			} 
		}
		else if(!strcmp(p, "priority"))
		{
			q = v; 
			while((*q != '\0') && (isdigit(*q))) q++;
			if((q != v) && (*q == '\0'))   
			{
				n = atoi(v);
				if((n < (int32)10) || (n > (int32)199))
				{
					(void)sprintf(e, "'%s' %s %s -- %s",
						p, "parameter must be",
						"in range 10-199",
						"value not changed");
					Error(e);
					eflag++;
				}
				else session.priority = n;
			}
			else
			{
				(void)sprintf(e, "'%s' %s %s",
					p,"parameter value must be",
					"numeric -- value not changed");
				Error(e);
				eflag++;
			} 
		}
		else if(!strcmp(p, "help"))
		{
			n = BooleanEval(v);
			if(n == (int32)1) HelpMessage();
			else
			{
				(void)sprintf(e, "'%s' %s %s",
					p,"parameter value must be",
					"'TRUE'");
				Error(e);
				eflag++;
			}
		}
		else if(!strcmp(p, "MemDebug"))
		{
			/* copy value, and execute the function with the value as 
				its string-argument */
			(void)strcpy(session.memdebug, v);
			InitMemDebug(v);
		}
		else if(!strcmp(p, "MemRation"))
		{
			(void)strcpy(session.memration, v);
			InitMemRation(v);
		}
		else if(!strcmp(p, "MemCheck"))
		{
			n = BooleanEval(v);
			if(n >= (int32)0) 
			{
				session.memcheck = n;
				if(n) InitMemCheck(TRUE);
				else InitMemCheck(FALSE);
			}
			else if(!strcmp(v, "VERBOSE") || !strcmp(v, "verbose") || !strcmp(v, "Verbose"))
				InitMemCheck(TRUE);
			else if(!strcmp(v, "NORMAL") || !strcmp(v, "normal") || !strcmp(v, "Normal"))
				InitMemCheck(FALSE);
			else
			{
				(void)sprintf(e, "'%s' %s %s -- %s",
					p, "parameter value must be",
					"'TRUE' or 'VERBOSE', else 'FALSE' or 'NORMAL'",
					"value not changed");
				Error(e);
				eflag++;
			} 
		}
		else
		{
			(void)sprintf(e, "parameter '%s' %s", p, 
				"is unrecognized and will be ignored");
			Error(e);
			eflag++;
		}
	}
	return(eflag);
}


/* evaluates Boolean expression: 'TRUE', 'True', 'true', 'T', 't', and '1' return 1;
	'FALSE', 'False', 'false', 'F', 'f', and '0' return 0; other strings return -1
	to represent UNRECOGNIZED */

int32 BooleanEval(char *string)
{
	int32 retval;
	
	if(!strcmp(string, "TRUE") ||
	   !strcmp(string, "True") ||
	   !strcmp(string, "true") ||
	   !strcmp(string, "T")    ||
	   !strcmp(string, "t")    ||
	   !strcmp(string, "1")) 
	   			retval = (int32)1;
			
	else if(!strcmp(string, "FALSE") ||
	   		!strcmp(string, "False") ||
	   		!strcmp(string, "false") ||
	   		!strcmp(string, "F")    ||
	   		!strcmp(string, "f")    ||
	   		!strcmp(string, "0")) 
	   			retval = (int32)0;
	else retval = (int32)(-1);
	return(retval);
}


/* appends the string 'string' to the buffer whose origin is pointed to by
	'origin', at the location pointed to by 'addloc', if space remains to
	add the string (and a preceding ',' if not at the origin) prior to 
	the location pointed to by 'endloc' */
	
char *AppendString(char *origin, char *addloc, char *endloc, char *string)
{
	int32 len;
	int32 span;
	
	/* set length of string to be added to buffer, 
		including ',' if this is not the first addition */
	len = strlen(string);
	span = len;
	if(addloc != origin) span++;
	if((addloc + span) >= endloc)
	{
		(void)sprintf("%s -- %s '%s'",
			"test-type string overflow", "failed to add", string);
		Error(e);
	}
	else
	{
		if(addloc != origin) *addloc++ = ',';
		(void)strcpy(addloc, string);
		addloc = addloc + len;
	}
	
	/* return pointer to terminal null */
	return(addloc);
}


/* set a string expressing the elapsed-time from Scaffold 
	initialization into 'session.elapsedstring', which makes
	it available to logging functions */
	
void SetElapsedTime(void)
{
	int32 etime;
	int32 sec;
	int32 msec;
	char *tag;
	char *s;
	char *t;
	char etimestring[ETIMELEN];

	/* if timer is not functioning, or if it is paused, display
		'-' instead of an elapsed-time */
	if(session.pausetime || !session.starttime)
	{
		(void)strcpy(etimestring, "-");
	}
	
	/* if timer is functioning and not paused, attempt to
		read the timer and set the elapsed-time value */
	else
	{
		/* read system's timer (which records the interval since
			system-boot in microseconds), and compute the 
			elapsed-time from Scaffold initialization in msec */
		etime = ReadTimer();

		/* in event of error or timer not enabled, write '-' in 
			place of elapsed-time */
		if(!etime)
		{
			(void)strcpy(etimestring, "-");
		}
		else
		{
			etime -= session.starttime;
			sec = etime / 1000;
			msec = etime - (sec * 1000);
			if(msec >= 100) tag = "";
			else if(msec >= 10) tag = "0";
			else tag = "00";
			(void)sprintf(etimestring, "%d.%s%d", 
				sec, tag, msec);
		}
	}
	
	/* pad '-' or elapsed-time with spaces to full width of field, 
		with the entry right-justified to a two-space pad */
	s = session.elapsedstring;
	t = session.elapsedstring + ETIMELEN - strlen(etimestring) - 2;
	while(s < t) *s++ = ' ';
	t = etimestring;
	while(*t != '\0') *s++ = *t++;
	*s++ = ' ';
	*s++ = ' ';
	*s = '\0';
	
	return;
}



/* FILE-ACCESS FUNCTIONS */


/* open the global-log or a local-log, or determine a unique destination
	filename for a remote log: 'tag' is "global", "local", or "remote"
	to designate the type; 'stem' is the stem upon which the logname 
	will be constructed; 'logname' points to a buffer 
	which will receive the full pathname of the log, and 'logfileptr' 
	points to a location which will receive the file-descriptor pointer;
	for local-log, if the file named by 'stem' exists, it will be
	opened for appending of new entries, but if it does not exist a new
	file with a unique name formed by appending <n> to the stem will
	be opened to receive just the log-entries of this testprogram; for
	global-log, <n> is always appended and the file is opened to receive
	just the log-entries of this testprogram; for remote-log, <n> is 
	always appended and the unique name is returned, but the file is
	NOT opened because copy rather than direct logging will be done;
	this function returns 0 for success, 1 for failure following a 
	message */

int32 OpenLog(char *tag, char *stem, char *logname, RawFile **logfileptr)
{
	int32 i;
	Err retval;
	int32 filemode;
	
	/* if tag is "local", first seek the file specified exactly by  
		the stem; if found (i = -1) this file will be opened below for 
		append; if not found, a unique filename with <n> appended to 
		the stem will be constructed and opened for writing below */
	filemode = (int32)0;
	if(!strcmp(tag, "local")) 
	{
		/* if stem contains '.log', use stem only for check; if not,
			append '.log' to stem for check */
		(void)strcpy(logname, stem);
		if(strstr(stem, ".log") == (char *)NULL) 
			(void)strcat(logname, ".log");
			
		/* attempt to open file for reading to check its existence */
		retval = OpenRawFile(logfileptr, logname, FILEOPEN_READ);
	
		/* if file can be opened, close it to prepare for re-opening to 
			append log-entries, below, and set flag 'i' to -1 to indicate
			the type of file-opening required */
		if(retval >= (Err)0)
		{
			CloseRawFile(*logfileptr);
			i = (int32)(-1);
			filemode = FILEOPEN_WRITE_EXISTING;
		}
		
		/* if file cannot be opened, it does not exist, so a unique 
			file-name with appended <n> will be sought below */
		else i = (int32)0;
	}
	else i = (int32)0;
	
	/* if tag is "local" and file corresponding to stem was not found
		above, or if tag is "global" or "remote", select a unique name
		for the logfile, <stem><n>.log, where <n> is the smallest 
		integer which will make the file's name unique */
	if(!i)
	{
		/* attempt to open files for reading, beginning with <n> 
			given value 0, and incrementing until the limit MAXLOGNO 
			is reached and an error is generated, indicating the file 
			could not be opened */
		for(i = (int32)0; i <= (int32)MAXLOGNO; i++)
		{
			/* create next sequential name and attempt to open the file;
				note that 'OpenRawFile' expects a pointer to the location
				in which it will deposit the file-descriptor pointer, so
				the direct-referenced 'logfileptr' is appropriate */
			(void)sprintf(logname, "%s_%d%s", stem, i, ".log");
			retval = OpenRawFile(logfileptr, logname, FILEOPEN_READ);
					
			/* if the file can be opened, close it and cycle for the 
				next; if it cannot be opened, break to open it for  
				writing with the currently-selected name */
			if(retval >= (int32)0) CloseRawFile(*logfileptr);
			else 
			{
				filemode = FILEOPEN_WRITE;
				break;
			}
		}
	}
			
	/* if name could not be assigned, issue error message and skip 
		to return 1 */
	if(i > (int32)MAXLOGNO)
	{
		(void)sprintf(e, "%s %s %s '%s' -- %s %d %s",
			"unable to assign", tag, "logfile's name with stem",
			stem, "no names with <n> less than", (int32)MAXLOGNO, 
			"are available");
		Error(e);
		i = (int32)1;				
	}

	/* if name was assigned, for "local" and "global" types only, 
		attempt to open the file; if unable to open, issue error 
		message and skip to return 0; "remote" type is not opened
		because transfer will be made by copy */
	else if(strcmp(tag, "remote"))
	{
		/* attempt to open the file */
		if(debug)
		{
			(void)sprintf(e, "%s %s %s '%s' %s %d",
				"about to open", tag, "file", logname, 
				"with filemode", filemode);
			DiagMessage(e);
		}
		i = OpenThisFile(logfileptr, logname, filemode);
		
		/* if successful */
		if(!i)
		{
			/* clear a flag which will be set when log is written */
			if(!strcmp(tag, "local")) localwriteflag = (int32)0;
			else globalwriteflag = (int32)0;
			
			/* if "local" type and appending, seek end-of-file */
			if((!strcmp(tag, "local")) && 
				(filemode == FILEOPEN_WRITE_EXISTING))
			{
				if(debug)
				{
					(void)sprintf(e, "%s %s %s '%s'",
						"about to seek end-of-file for", tag, "file",  
						logname);
					DiagMessage(e);
				}
				if(SeekRawFile(*logfileptr, 0, FILESEEK_END) < (Err)0)
				{
					(void)sprintf(e, "%s '%s'",
						"unable to seek end of file", logname);
					Error(e);
				}
				else if(debug)
				{
					(void)sprintf(e, "%s %s %s '%s'",
						"successfully sought end-of-file for", tag, "file",  
						logname);
					DiagMessage(e);
				}
			}
		}
	}
	
	/* for "remote" type, assure return-value of 0 unless a unique 
		name could not be assigned, above */
	else
	{
		if(debug)
		{
			(void)sprintf(e, "%s '%s'", 
				"selected unique remote logfile name", logname);
			DiagMessage(e);
		}
		i = 0;
	}
	
	/* return 0 for success, 1 for error */
	return(i);
}


/* open the file named by 'filename' with mode specified by 'filemode',
	depositing the file-descriptor in the location pointed to by 'fileptr';
	this function is invoked from 'OpenLogFile' and from 'CopyLogFile' */
			
int32 OpenThisFile(RawFile **fileptr, char *filename, int32 filemode)
{
	char *tag1;
	char *tag2;
	Err code;
	int32 retval;

	/* announce */
	if(debug)
	{
		if((filemode == FILEOPEN_WRITE) || (filemode == FILEOPEN_WRITE_NEW)) 
			tag1 = "new";
		else tag1 = "existing";
		if(filemode == FILEOPEN_READ) tag2 = "reading";
		else tag2 = "writing";
		(void)sprintf(e, "opening %s file '%s' for %s", 
			tag1, filename, tag2);
		DiagMessage(e);
	}
		
	/* attempt to open file */	
	code = OpenRawFile(fileptr, filename, filemode);
		
	/* if unable to open */
	if((code < (Err)0) || (*fileptr == (RawFile *)NULL)) 
	{
		if(code < (Err)0) 
		{
			PrintfSysErr(code);
			DiagMessage("file was not opened");
		}
		else
		{
			(void)sprintf(e, "%s",
				"file-pointer was returned NULL, without error");
			DiagMessage(e);
		}
		(void)sprintf(e, "%s '%s'",
			"unable to open file", filename);
		Error(e);

		/* set return-value 1 */
		retval = (int32)1;
	}
		
	/* if opened successfully */
	else 
	{
		if(debug)
		{
			(void)sprintf(e, "%s '%s'",
				"successfully opened logfile", 
				filename);
			DiagMessage(e);
		}
			
		/* set return-value 0 */
		retval = (int32)0;
	}
	
	return(retval);
}


/* close either the global log or the local log: 'tag' is either "global" 
	or "local" to designate the type; this function is invoked from 
	'TermScaffold' for local-log and global-log, and from 'CopyLogFile'
	for global-log and remote-log */
	
void CloseLog(char *tag)
{
	char *name;
	int32 deleteflag;

	/* clear flag which permits file-deletion after close */
	deleteflag = (int32)0; 

	/* for global-log, close it and delete if empty (when first
		closed) or if copy was successful (for second close, 
		after copy is attempted) */	
	if(!strcmp(tag, "global")) 
	{
		CloseRawFile(globallogfile);
		name = globallogname;
		if(!globalwriteflag) deleteflag = (int32)1;
	}
	
	/* for local-log, close it and delete if empty */
	else if(!strcmp(tag, "local")) 
	{
		CloseRawFile(locallogfile);
		name = locallogname;
		if(!localwriteflag) deleteflag = (int32)1;
	}
	
	/* for remote-log, just close it */
	else if(!strcmp(tag, "remote"))
	{
		CloseRawFile(remotelogfile);
		name = remotelogname;
	}
	else
	{
		(void)sprintf(e, "file to be closed (type '%s') %s",
			tag, "must be 'global', 'local', or 'remote'");
		Error(e);
		return;
	}
	
	/* announce */
	if(debug)
	{
		(void)sprintf(e, "%s %s %s '%s'",
			"closed", tag, "logfile", name);
		DiagMessage(e);
	}
	
	/* delete log-file if conditions for this were met above */
	if(deleteflag)
	{
		DeleteFile(name);
		if(debug)
		{
			(void)sprintf(e, "%s %s %s '%s'",
				"deleted", tag, "logfile", 
				name);
			DiagMessage(e);
		}
	}
	return;
}


/* copy contents of global-logfile to remote-logfile */

int32 CopyLogFile(void)
{
	int32 retval;
	char buf[1024];
	int32 len;
	int32 totallen;

	/* initialize accumulated length */
	totallen = (int32)0;
	
	/* attempt to open source for reading */
	if(debug)
	{
		(void)sprintf(e, "%s '%s' to '%s'", 
			"opening files for copy of", globallogname, remotelogname);
		DiagMessage(e);
	}
	retval = OpenThisFile(&globallogfile, globallogname, FILEOPEN_READ);
	
	/* if successful, attempt to open destination, copy, and close
		destination, then close source */
	if(!retval)
	{
		/* attempt to open destination-file for writing */
		retval = OpenThisFile(&remotelogfile, remotelogname, FILEOPEN_WRITE);

		/* if destination-file was opened successfully, read from source-file 
			to buffer and write from buffer to destination-file until all 
			bytes have been read, then close destination-file */
		if(!retval)
		{
			if(debug)
			{
				DiagMessage("beginning file-copy");
			}
			len = (int32)1024;
			while(!retval)
			{
				/* attempt to read full buffer from source-file */
				len = ReadRawFile(globallogfile, buf, len);
				
				/* if read was successful and characters were 
					returned, accumulate length and write the
					buffer to destination file */
				if(len > (int32)0) 
				{
					/* accumulate length */
					totallen += len;
					
					/* write to remote */
					if(WriteRawFile(remotelogfile, buf, len) < 0)
					{
						(void)sprintf(e, "%s '%s' -- %s",
							"encountered error in writing file",
							remotelogname,
							"suspending copy-operation");
						Error(e);
						retval = (int32)1;
					}
					
					/* if successful, set span of 1024 bytes 
						for next read */
					else len = (int32)1024;
				}
				
				/* if length is 0, this terminates the copy 
					successfully */
				else if(len == 0) break;
				
				/* if error on read-attempt, suspend copy */
				else 
				{
					(void)sprintf(e, "%s '%s' -- %s",
						"encountered error in reading file",
						globallogname,
						"suspending copy-operation");
					Error(e);
					retval = (int32)1;
				}
			}
			CloseLog("remote");
		}

		/* close source-file; 'globalwriteflag' remains set so this file 
			will not be deleted by 'CloseLog' */
		CloseLog("global");
	}
	
	/* announce success if message is enabled; failure will be announced by
		'TermScaffold' */
	if(!retval & debug)
	{
		(void)sprintf(e, "copy succeeded -- %d bytes", totallen);
		DiagMessage(e);
	}

	/* return 0 for success, 1 for failure */
	return(retval);		
}


#ifdef NOTUSED

int32 SimpleCopyLogFile(void)
{
	char cmdline[1024];
	int32 retval;
	
	(void)sprintf(cmdline, "copy %s %s", globallogname, 
		remotelogname);
	if(debug)
	{
		(void)sprintf(e, "%s '%s'", "log-copy command-line is", 
			cmdline);
		DiagMessage(e);
	}
	retval = system(cmdline);
	if(retval)
	{
		/* message */
		(void)sprintf(e,"%s '%s'",
			"log-copy failed due to undetermined error on command-line",
			cmdline);
		Error(e);
	}
	return(retval);
}

#endif


/* write a complete line, pointed to by 'buf', to all 
	enabled logs; note that because the line may first
	have to be terminated, 'buf' must point to a buffer 
	of sufficient size, rather to than a fixed string; 
	if logging is deferred and space is available, the 
	log-line is buffered in the deferred-log buffer 
	'dbuffer' at the location pointed to by 
	'dbufferptr' */

void WriteLog(char *buf)
{
	char *s;
	int32 len;
	
	/* determine length of line */
	len = strlen(buf);
	
	/* if logging is being deferred, append to the 
		deferred-log buffer 'dbuffer', at the location 
		of its pointer 'dbufferptr', rather than writing
		immediately to the display and/or logs */
	if(dbufferptr != (char *)NULL) 
	{
		if((dbufferptr + len + 9) >= dbufferover)
		{
			(void)sprintf(e, "%s '%s' %s -- %s",
				"logging", buf, "will overflow buffer",
				"logging suspended");
			Error(e);
		}
		else
		{
			if(debug)
			{
				(void)strcpy(dbufferptr, "DEFERRED ");
				dbufferptr += 9;
			}
			(void)strcpy(dbufferptr, buf);
			dbufferptr += len;
			if(len && (*(dbufferptr - 1) != '\r'))
			{
				*dbufferptr++ = '\r';
				*dbufferptr = '\0';
			}
		}
	}

	/* if logging is immediate, write to the display 
		if it is enabled, and to enabled logs */
	else
	{
		DisposeLog(buf, len);
	}
	return;
}


/* write log-buffer to the display if display is enabled, and to 
    local log if log is enabled and open; this function is invoked
	line by line from 'WriteLog' if logging is immediate, or only 
	at the conclusion of the test program from 'TermScaffold' if
	logging is deferred */

void DisposeLog(char *buf, int32 len)
{
	/* write to the display if enabled */	
	if(session.display)
	{
		kprintf("%s\r", buf);
	}

	/* append '\r' to line if line does not already end with '\r' */
	if(len && (*(buf + len - 1) != '\r'))
	{
		*(buf + len) = '\r';
		len++;
		*(buf + len) = '\0';
	}
	
	/* write to the local-logfile if enabled and log is open */
	if((session.file) && (locallogfile != (RawFile *)NULL))
	{
		WriteRawFile(locallogfile, buf, len);
		localwriteflag = (int32)1;
	}
	
	/* write to the global-logfile if enabled and log is open;
		this file will later be copied to the results-database
		by 'TermScaffold' */
	if((session.database) && (globallogfile != (RawFile *)NULL))
	{
		WriteRawFile(globallogfile, buf, len);
		globalwriteflag = (int32)1;
	}
		
	return;
}


void HelpMessage()
{
	kprintf("See the document '3DO Test-Automation Frameworks'\r");
	kprintf("  for details on use of the Scaffold API\r");
	return;
}


void DisplayMessage(char *messagestring)
{
	kprintf("%s\r", messagestring);
	return;
}


void DiagMessage(char *messagestring)
{
	kprintf("DIAGNOSTIC: %s\r", messagestring);
	return;
}


void Error(char *errorstring)
{
	kprintf("SCAFFOLD ERROR: %s\r", errorstring);
	return;
}


/* issue error message if Scaffold is not yet initialized 
	(initflag = 0) or if Scaffold has already been terminated
	(initflag < 0) */
	
void ScaffoldError(char *funcname)
{
	if(!initflag)
		(void)sprintf(e, "%s '%s' %s",
			"encountered Scaffold-function", funcname,
			"prior to Scaffold initialization");
	else
		(void)sprintf(e, "%s '%s' %s",
			"encountered Scaffold-function", funcname,
			"after Scaffold termination");
	Error(e);
	return;
}


/* read the system timer and return the elapsed-time in 
	milliseconds since system bootup, or 0 in the event 
	of error */

int32 ReadTimer(void)
{
	TimeVal tv;
	int32 ret;
	
	/* get the time-value */
	SampleSystemTimeTV(&tv);
	
	/* compute the timer value in msec */
	ret = tv.tv_sec * 1000;
	ret += tv.tv_usec / 1000;
	
	return(ret);
}


static void WaitForNoButtonPresses(void)
{
	Err	ereturn;
	ControlPadEventData	theCPED;
	
	do 
	{
		ereturn = GetControlPad(1, FALSE, &theCPED);
		if (ereturn < 0) return;
		theCPED.cped_ButtonBits &= kButtonMask;
		
	} while(theCPED.cped_ButtonBits != 0);
}


