/*
	File:		testvideostream.c

	Contains:	datstreaming tests based on VideoPlayer

	Written by:	amasri

	Copyright:	 1996 by The 3DO Company. All rights reserved.
				This material constitutes confidential and proprietary
				information of the 3DO Company and shall not be used by
				any Person or for any purpose except as expressly
				authorized in writing by the 3DO Company.

	Change History (most recent first):

		 <8>	  5/7/96	PUT YOUR INITIALS HERE		fix disposedataacq
		 <7>	  5/7/96	am		added EZFlix tests
		<4>	 	3/14/96		am		rework with VideoPlayer and 3do_Scaf.h
		<3+>	  3/7/96	am		remove controlsubscriber.h
		 <3>	  3/7/96	am		brought into conformance with M2 2.0 data stream libs

	To Do:

 @(#) testvideostream.c 96/07/10 1.1

*/
#ifndef PRINT_LEVEL		/* may be set by a compiler command-line switch */
#define PRINT_LEVEL 1	/* print error msgs */
#endif

#include <string.h>				/* memset */
#include <kernel/debug.h>
#include <kernel/mem.h>			/* for AllocMemTrack, FreeMemTrack   */
#include <audio/audio.h>
#include <graphics/view.h>		/* VIEWTAG_BITMAP */

#include <streaming/datastream.h>
#include <streaming/datastreamlib.h>
#include <streaming/dsstreamdefs.h>
#include <streaming/saudiosubscriber.h>
#include <streaming/satemplatedefs.h>
#include <streaming/sacontrolmsgs.h>
#include <streaming/mpegvideosubscriber.h>


#include "testvideostream.h"
#include "joypad.h"
#include "tests.h"

/************************************/
/* Constants and structures         */
/************************************/
#undef kDefaultNumBuffers					/* override the setting in preparestream.h */
#define kDefaultNumBuffers		6			/* suggested number of stream buffers to use */


/************************************/
/* Local utility routine prototypes */
/************************************/
static int32	InitVideoPlayerFromStreamHeader(PlayerPtr ctx, char* streamFileName,
					uint32 bitDepth);
static int32	HandleMPEGFrames(PlayerPtr ctx);
Err DisplayEZFX(PlayerPtr ctx);
Err DisplayMPEG(PlayerPtr ctx);

/* globals and test flags */
extern char* sTestName;
static uint32 gVideoType;
bool fDrawEZflixToBuffer = FALSE;
bool fSendFreeEZFlixSignal = FALSE;
bool fClockTest = FALSE;
bool fEZFlixCurrTime = FALSE;
bool fEZFlixDuration = FALSE;

/*********************************************************************************************
 * This function is called each time through the player's main loop and can be used to
 * implement any user interface actions necessary. Returning a zero causes playback to
 * continue. Any other value causes playback to cease and immediately return from
 * the PlayVideoStream() function.
 *********************************************************************************************/
int32 myUserFunction(PlayerPtr ctx)
	{
	int32			status = kDSNoErr;
	JoyPadStatePtr	jp;

	/* Our only additional context is a pointer to the joypad state. */
	jp = (JoyPadStatePtr) ctx->userContext;

	/* Read control pad #1. */
	GetJoyPad(jp, 1);


	if ( jp->xBtn )
		/* "X" aka Stop button ==> Return 1 to stop the stream */
		return 1;

	else if ( jp->upArrow )
		{
		/* Jump forward to the next marker in the stream. */
		status = DSGoMarker(ctx->messageItem, NULL, ctx->streamCBPtr, 1,
							GOMARKER_FORWARD);
		if ( status == kDSBranchNotDefined )
			{
			PERR(("Can't skip forward: The stream needs a marker table.\n"));
			status = kDSNoErr;
			}
		else if ( status == kDSRangeErr )
			{
			/* Skip forward range error. Beyond end? If we're looping, jump to
			 * the start of the next loop. */
			if ( ctx->fLoop )
				status = DSGoMarker(ctx->messageItem, NULL, ctx->streamCBPtr, 0,
					GOMARKER_ABSOLUTE);
			else
				{
				PERR(("Skip forward range error. Beyond end?\n"));
				status = kDSNoErr;
				}
			}
		FAIL_NEG("DSGoMarker fwd", status);
		}

	else if ( jp->downArrow )
		{
		/* Jump backward to the previous marker in the stream. */
		status = DSGoMarker(ctx->messageItem, NULL, ctx->streamCBPtr, 1,
							GOMARKER_BACKWARD);
		if ( status == kDSBranchNotDefined )
			{
			PERR(("Can't skip backward: The stream needs a marker table.\n"));
			status = kDSNoErr;
			}
		else if ( status == kDSRangeErr )
			/* Skip backward range error. Beyond start? Jump back to start. */
			status = DSGoMarker(ctx->messageItem, NULL, ctx->streamCBPtr, 0,
				GOMARKER_ABSOLUTE);
		FAIL_NEG("DSGoMarker bwd", status);
		}

	else if ( jp->leftArrow )
		{
		/* Jump to the beginning */
		/* For streamer testing: unshifted-leftArrow ==> ASAP flush branch (the
		 * normal way to branch in response to a user button push),
		 * shifted arrow => butt-joint branch. */
		status = DSGoMarker(ctx->messageItem, NULL, ctx->streamCBPtr, 0,
			( jp->rightShift | jp->leftShift ) ?
				GOMARKER_ABSOLUTE | GOMARKER_NO_FLUSH_FLAG : GOMARKER_ABSOLUTE);
		FAIL_NEG("DSGoMarker 0", status);
		}

	else if ( jp->startBtn )
		{
		/* Pause / Unpause the stream */
		if ( DSIsRunning(ctx->streamCBPtr) )
			{
			/* Note: Stop without SOPT_FLUSH so we can later resume. */
			status = DSStopStream(ctx->messageItem, NULL, ctx->streamCBPtr, SOPT_NOFLUSH);
			FAIL_NEG("DSStopStream", status);
			}
		else
			{
			status = DSStartStream(ctx->messageItem, NULL, ctx->streamCBPtr, SOPT_NOFLUSH);
			FAIL_NEG("DSStartStream", status);
			}
		}

	else if ( jp->aBtn ) {
		/* Execute branching tests. Execute a different test each time the a button
		 * is pressed. We use the flags to assure that each test is only performed
		 * once. Control drops through the if statements.
		 */
		if (!fClockTest){
			status = DoClockTest(ctx);
			fClockTest = TRUE;
			goto FAILED;
		}
		if (!fEZFlixCurrTime & gVideoType == 'EZFX'){
			status = DoEZFlixCurrTimeTest(ctx->channelPtr);
			fEZFlixCurrTime = TRUE;
			goto FAILED;
		}
		if (!fEZFlixDuration & gVideoType == 'EZFX'){
			status = DoEZFlixDurationTest(ctx->channelPtr);
			fEZFlixDuration = TRUE;
			goto FAILED;
		}
	}
FAILED:
	return status;
	}

Err DisplayMPEG(PlayerPtr ctx) {
	/* NOTE: A better scheme here would be to also wait for a control port event
	 * signal (rather than an audio Cue to poll the control port) and for a
	 * time-to-display-frame signal (rather than a subroutine that sleeps
	 * until time to display and sleeps until the display has begun). */
	int32			status = 0;
	int32			playerResult;
	uint32			signalMask;
	uint32			currentSignal;
	Item			msgItem;
	Boolean			keepRunning = TRUE;
	
	/* Set up the mask for signals we're expecting */
	signalMask =	ctx->frameCompletionPortSignal |		/* Frame Completions */
					ctx->messagePortSignal |				/* End of stream replys */
					ctx->userCueSignal;						/* Periodic user events */

	/* --------------- Main playback loop ---------------
	 * NOTE: A better scheme here would be to also wait for a control port event
	 * signal (rather than an audio Cue to poll the control port) and for a
	 * time-to-display-frame signal (rather than a subroutine that sleeps
	 * until time to display and sleeps until the display has begun). */
	while ( keepRunning )
		{
		currentSignal = WaitSignal(signalMask);
		
		/* If a user function is defined, then call it.
		 * If it returns a non-zero value, then exit. */
		if ( currentSignal & ctx->userCueSignal )
			if ( ctx->userFn != NULL )
				{
				playerResult = (*ctx->userFn)(ctx);
				if ( playerResult )
					keepRunning = FALSE;
				else
					SignalAtTime(ctx->userCue,
						GetAudioTime() + kUserFunctionPeriod);
				}

		/* Check for a new-frame-to-display signal. */
		if ( currentSignal & ctx->frameCompletionPortSignal )
			{
			status = HandleMPEGFrames(ctx);
			FAIL_NEG("HandleMPEGFrames", status);
			}
		
		/* Check for end-of-stream signal. */
		if ( currentSignal & ctx->messagePortSignal ) {
			while ( (msgItem = GetMsg(ctx->messagePort)) != 0 )
				{
				FAIL_NEG("PlayVideoStream GetMsg", msgItem);
				if ( msgItem == ctx->endOfStreamMessageItem )
						keepRunning = FALSE;
				}
			}
		}
FAILED:
	return status;
}

Err	TestVideoStream(ScreenInfo *screenInfoPtr, uint32 bitDepth,
		char *streamFileName, PlayVideoUserFn userFn, void *userContext,
		bool fLoop)
	{
	Err				status;
	int32			playerResult = 0;
	PlayerPtr		ctx;
	Player			playerContext;
	DSRequestMsg	EOSMessage;
	

	memset(ctx = &playerContext, 0, sizeof(playerContext));

	/* Remember the screen context pointer that we will use for drawing. */
	ctx->screenInfoPtr = screenInfoPtr;
	
	/* Initialize the context's fLoop flag from the fLoop arg.
		 * The userFcn can access and modify this flag. */
		ctx->fLoop = fLoop;
	
		/* Open the specified file, locate the stream header, and initialize
		 * all resources necessary for performing stream playback. */
		status = InitVideoPlayerFromStreamHeader(&playerContext, streamFileName, bitDepth);
		FAIL_NEG("InitVideoPlayerFromStreamHeader", status);
	
		/* Remember the user function pointer, if any */
		ctx->userFn = userFn;
		ctx->userContext = userContext;
	
		/* If we were supplied with a user function, we have to initiate a
		 * wake up siginal to make sure it gets called regularly */
		if (userFn)
			{
			status = CreateItem(MKNODEID(AUDIONODE, AUDIO_CUE_NODE), NULL);
			FAIL_NEG("Create Audio userCue Item", status);
			ctx->userCue = status;
			
			ctx->userCueSignal = GetCueSignal(ctx->userCue);
			}
	
		/* Register for end of stream notification. Do it before DSStartStream so
	 * we won't miss it in a short or empty stream. */
	status = DSWaitEndOfStream(ctx->endOfStreamMessageItem, &EOSMessage, ctx->streamCBPtr);
	FAIL_NEG("DSWaitEndOfStream", status);

	PRNT(("Prerolling the stream\n"));
	status = DSPreRollStream(ctx->messageItem, NULL, ctx->streamCBPtr, 2);
	sTestName = "DSPreRollStream";
	BeginTest(FT, sTestName);
	if (status < 0){
		LogResult( FAILED, "%s fails with error %d", sTestName, (int)status);
		goto FAILED;
	}
	else {
		LogResult( PASSED, "%s passes", sTestName);
	}

	/* Start the stream running */
	PRNT(("Starting the stream\n"));
	status = DSStartStream(ctx->messageItem, NULL, ctx->streamCBPtr, 0);
	sTestName = "DSStartStream";
	BeginTest(FT, sTestName);
	if (status < 0){
		LogResult( FAILED, "%s fails with error %d", sTestName, (int)status);
		goto FAILED;
	}
	else {
		LogResult( PASSED, "%s passes", sTestName);
	}
	FAIL_NEG("DSStartStream", status);

	/* Prime the user cue signal if present */
	if ( ctx->userCue )
		SignalAtTime(ctx->userCue, GetAudioTime() + kUserFunctionPeriod);
	
	if (gVideoType == MPEG)
		DisplayMPEG(ctx);
	else if (gVideoType == EZFX)
		DisplayEZFX(ctx);

	/* Set up the mask for signals we're expecting */

	/* Unconditionally stop the stream */
	sTestName = "DSStopStream";
	BeginTest(FT, sTestName);
	status = DSStopStream(ctx->messageItem, NULL, ctx->streamCBPtr, SOPT_FLUSH);
	if (status < 0){
		LogResult( FAILED, "%s fails with error %d", sTestName, (int)status);
		goto FAILED;
	}
	else {
		LogResult( PASSED, "%s passes", sTestName);
	}
	PRNT(("Stopped the stream\n"));

#if 0	/* there's no CinePak subscriber in this program */
	/* Flush anything held by the CinePak subscriber */
	if ( ctx->cpakContextPtr != NULL )
		{
		FlushCPakChannel(ctx->cpakContextPtr, ctx->cpakChannelPtr, 0);
		}
#endif 

FAILED:

	/* Cleanup; deallocate */
	DismantlePlayer(ctx);

	return ( status < 0 ) ? status : playerResult;
	}

/*******************************************************************************************
 * Routine to get the stream header info, and perform all necessary allocations
 * and initializations necessary for stream playback.
 *******************************************************************************************/
static Err	InitVideoPlayerFromStreamHeader(PlayerPtr ctx, char* streamFileName,
		uint32 bitDepth)
	{
	int32				status;
	long				subscriberIndex;
	long				channelNum;
	int32				i;
	
	SAudioCtlBlock		ctlBlock;	/* Control blocks for communicating with subscribers */
	MPEGVideoCtlBlock	mpegCtl;
	DSHeaderSubsPtr		subsPtr;
	
	Boolean				fStreamHasAudio = false;	/* assume no audio subscriber */

	/* Get the stream header loaded */
	sTestName = "FindAndLoadStreamHeader";
	BeginTest(FT, sTestName);
	status = FindAndLoadStreamHeader(&ctx->hdr, streamFileName);
	if (status == kDSHeaderNotFound){
			LogResult( FAILED, "%s fails: no header was found.", sTestName );
			goto FAILED;
		}
	else if (status != 0)  {
		LogResult( FAILED, "%s fails with error %d", sTestName, (int)status );
		goto FAILED;
	}
	else {
		LogResult( PASSED, "%s passes", sTestName );
	}
	
	/* Make sure this playback code is compatible with the version of the
	 * data in the stream file. */
	/*if ( ctx->hdr.headerVersion != DS_STREAM_VERSION )
		return kDSVersionErr;*/

	/* Allocate the stream buffers and build the linked list of
	 * buffers that are input to the streamer. */
	sTestName = "CreateBufferList";
	BeginTest(FT, sTestName);
	ctx->bufferList = CreateBufferList(ctx->hdr.streamBuffers, ctx->hdr.streamBlockSize);
	if (ctx->bufferList == NULL ){
		LogResult( FAILED, "%s fails to allocate buffer", sTestName );
		status = kDSNoMemErr;
		goto FAILED;
	}
	else {
		LogResult( PASSED, "%s passes", sTestName);
	}


	/* We need to create a message port and message items
	 * to communicate with the data streamer. */
	ctx->messagePort = NewMsgPort(&ctx->messagePortSignal);
	FAIL_NEG("NewMsgPort", ctx->messagePort);


	ctx->messageItem = CreateMsgItem(ctx->messagePort);
	FAIL_NEG("CreateMsgItem", ctx->messageItem);


	ctx->endOfStreamMessageItem = CreateMsgItem(ctx->messagePort);
	FAIL_NEG("Create EOS MsgItem", ctx->endOfStreamMessageItem);

	
	status = CreateItem(MKNODEID(AUDIONODE, AUDIO_CUE_NODE), NULL);
	FAIL_NEG("Create Audio synchCue Item", status);
	ctx->synchCue = status;


	/* Also allocate a message port for recieving frame completion signals */
	ctx->frameCompletionPort = NewMsgPort(&ctx->frameCompletionPortSignal);
	FAIL_NEG("NewMsgPort frameCompletionPort", ctx->frameCompletionPort);


	/* Open a data acquisition thread on the specified file */
	status = ctx->acqMsgPort = DNewDataAcq(streamFileName, ctx->hdr.dataAcqDeltaPri);
	if (status < 0){
		sTestName = "NewDataAcq";
		BeginTest(FT, sTestName);
		LogResult( FAILED, "%s fails with error %d", sTestName, (int)status);
		goto FAILED;
	}
	else {
		sTestName = "NewDataAcq";
		BeginTest(FT, sTestName);
		LogResult( PASSED, "%s passes", sTestName);
	}

	sTestName = "NewDataStream";
	BeginTest(FT, sTestName);
	status = NewDataStream(&ctx->streamCBPtr,ctx->bufferList,ctx->hdr.streamBlockSize,ctx->hdr.streamerDeltaPri,ctx->hdr.numSubsMsgs);
	if (status < 0){
		LogResult( FAILED, "%s fails with error %d", sTestName, (int)status);
		goto FAILED;
	}
	else {
		LogResult( PASSED, "%s passes", sTestName);
	}

	/* Connect the stream to its data supplier */
	status = DSConnect(ctx->messageItem, NULL, ctx->streamCBPtr, ctx->acqMsgPort);
	if (status < 0){
		sTestName = "DSConnect";
		BeginTest(FT, sTestName);
		LogResult( FAILED, "%s fails with error %d", sTestName, (int)status);
		goto FAILED;
	}
	else {
		sTestName = "DSConnect";
		BeginTest(FT, sTestName);
		LogResult( PASSED, "%s passes", sTestName);
	}

	/* Loop through the subscriber descriptor table and initialize all
	 * subscribers specified in the table. */
	for ( subscriberIndex = 0;
			subsPtr = ctx->hdr.subscriberList + subscriberIndex,
			subsPtr->subscriberType != 0;
			subscriberIndex++ )
		{
		switch ( subsPtr->subscriberType )
			{
			case MPVD_CHUNK_TYPE:
				/* Remember the target bit depth of the playback we're starting */
				ctx->bitDepth = bitDepth;

				/* This is an MPEG stream, so ignore other video types */
				gVideoType = MPEG;
				
				/* Instantiate an MPEG video subscriber and sign it up for
				 * a DataStreamer subscription (now in one easy step!) */
				sTestName = "NewMPEGVideoSubscriber";
				BeginTest(FT, sTestName);
				status = NewMPEGVideoSubscriber(ctx->streamCBPtr,subsPtr->deltaPriority, ctx->messageItem, bitDepth);
				if (status < 0){
					LogResult( FAILED, "%s fails with error %d", sTestName, (int)status);
					goto FAILED;
				}
				else {
					LogResult( PASSED, "%s passes", sTestName);
				}
				ctx->mpegVideoSubMsgPort = status;
				
				
				/* The MPEG Subscriber requires a message port to send information
				 * about decompressed frames to.  This enables the application to 
				 * take care of all display activities. */
				mpegCtl.SetClientPort.clientPort = ctx->frameCompletionPort;
				sTestName = "DSControl3";
				BeginTest(FT, sTestName);
				status = DSControl(ctx->messageItem,NULL,ctx->streamCBPtr,MPVD_CHUNK_TYPE,kMPEGVideoCtlOpSetClientPort,&mpegCtl);
				if (status < 0){
					LogResult( FAILED, "%s fails with error %d", sTestName, (int)status);
					goto FAILED;
				}
				else {
					LogResult( PASSED, "%s passes", sTestName);
				}
				/* Now send the MPEGVideoSubscriber a pointer to the frame buffer to
				 * decompress into.  Note that we're only sending a chunk of memory, so we
				 * could theoretically provide a pointer to any kind of DMA-able memory. */
				mpegCtl.SetFrameBuffers.frameBuffers.numFrameBuffers = cntFrameBuffers;
				mpegCtl.SetFrameBuffers.frameBuffers.frameBufferSize = (bitDepth == 16) ?
					NTSC_FRAME_WIDTH * NTSC_FRAME_HEIGHT * 2 :
					NTSC_FRAME_WIDTH * NTSC_FRAME_HEIGHT * 4;
					
				mpegCtl.SetFrameBuffers.frameBuffers.bitDepth = bitDepth;

				for ( i = 0; i < cntFrameBuffers; i++ )
					mpegCtl.SetFrameBuffers.frameBuffers.frameBuffer[i] =
						ctx->screenInfoPtr->bitmap[i]->bm_Buffer;
		
				sTestName = "DSControl1";
				BeginTest(FT, sTestName);
				status = DSControl(ctx->messageItem,NULL,ctx->streamCBPtr,MPVD_CHUNK_TYPE,kMPEGVideoCtlOpSetFrameBuffers,&mpegCtl);
				if (status < 0){
					LogResult( FAILED, "%s fails with error %d", sTestName, (int)status);
					goto FAILED;
				}
				else {
					LogResult( PASSED, "%s passes", sTestName);
				}
				break;

			case EZFLIX_CHUNK_TYPE:
				/* This is an EZFlix stream, so ignore other video types */
				gVideoType = EZFX;
				sTestName = "NewEZFlixSubscriber";
				BeginTest(FT, sTestName);
				status = NewEZFlixSubscriber( ctx->streamCBPtr, subsPtr->deltaPriority, ctx->messageItem, 1, &ctx->contextPtr);
				if (status < 0){
					LogResult( FAILED, "%s fails with error %d", sTestName, (int)status);
					goto FAILED;
				}
				else {
					LogResult( PASSED, "%s passes", sTestName);
				}
				sTestName = "InitEZFlixStreamState";
				BeginTest(FT, sTestName);
				status = InitEZFlixStreamState(ctx->streamCBPtr,ctx->contextPtr,&ctx->channelPtr,0,true );
				if (status < 0){
					LogResult( FAILED, "%s fails with error %d", sTestName, (int)status);
					goto FAILED;
				}
				else {
					LogResult( PASSED, "%s passes", sTestName);
				}
				break;
			
			case SNDS_CHUNK_TYPE:
				sTestName = "NewSAudioSubscriber";
				BeginTest(FT, sTestName);
				status = NewSAudioSubscriber(ctx->streamCBPtr, subsPtr->deltaPriority,ctx->messageItem);
				if (status < 0){
					LogResult( FAILED, "%s fails with error %d", sTestName, (int)status);
					goto FAILED;
				}
				else {
					LogResult( PASSED, "%s passes", sTestName);
				}
				fStreamHasAudio = true;
				break;

			default:
				PERR(("InitVideoPlayerFromStreamHeader - unknown subscriber type '%.4s'\n",
					(char*)&subsPtr->subscriberType));
				status = kDSInvalidTypeErr;
				goto FAILED;
			}
		}

	/* If the stream has audio, then do some additional initializations. */
	if ( fStreamHasAudio )
		{
		/* Preload audio instrument templates, if any are specified */
		if ( ctx->hdr.preloadInstList != 0 ){
			ctlBlock.loadTemplates.tagListPtr = ctx->hdr.preloadInstList;
			sTestName = "DSControl2";
			BeginTest(FT, sTestName);
			status = DSControl(ctx->messageItem, NULL, ctx->streamCBPtr, SNDS_CHUNK_TYPE, kSAudioCtlOpLoadTemplates, &ctlBlock);
			if (status < 0){
				LogResult( FAILED, "%s fails with error %d", sTestName, (int)status);
				goto FAILED;
			}
			else {
				LogResult( PASSED, "%s passes", sTestName);
			}
		}
		/* Enable a second audio channel so we can set up a test */
		ctx->hdr.enableAudioChan = 0x0003;
	
		/* Enable any audio channels whose enable bit is set.
		 * NOTE: Channel zero is enabled by default, so we don't check it. */
		for ( channelNum = 1; channelNum < 32; channelNum++ ){
			/* If the bit corresponding to the channel number is set,
			 * then tell the audio subscriber to enable that channel. */
			if ( ctx->hdr.enableAudioChan & (1L << channelNum) ){
				sTestName = "DSSetChannel";
				BeginTest(FT, sTestName);
				status = DSSetChannel(ctx->messageItem, NULL, ctx->streamCBPtr, SNDS_CHUNK_TYPE, channelNum, CHAN_ENABLED, CHAN_ENABLED);
				if (status < 0){
					LogResult( FAILED, "%s fails with error %d", sTestName, (int)status);
					goto FAILED;
				}
				else {
					LogResult( PASSED, "%s passes", sTestName);
				}
			}
		}
		DoChannelMaskTest(ctx);	
		/* Set the audio clock to use the selected channel */	
		ctlBlock.clock.channelNumber = ctx->hdr.audioClockChan;
		status = DSControl(ctx->messageItem, NULL, ctx->streamCBPtr, SNDS_CHUNK_TYPE,
								 kSAudioCtlOpSetClockChan, &ctlBlock);
		FAIL_NEG("DSControl - setting audio clock chan", status);
	}

	return kDSNoErr;

FAILED:
	DismantlePlayer(ctx);
	return status;
	}


/*******************************************************************************************
 * Routine to free all resources associated with a Player structure. Assumes that all
 * relevant fields are set to ZERO when the struct is initialized.
 *
 *	NOTE:	THE ORDER OF THE FOLLOWING DISPOSALS IS IMPORTANT. DO NOT CHANGE UNLESS
 *			YOU KNOW WHAT YOU ARE DOING.
 *
 *******************************************************************************************/
void	DismantlePlayer(PlayerPtr ctx)
	{
	Err status;
	
	/* If we have a frame hanging out, return it to the subscriber before
	 * we turn off the subscribers. */
	if (gVideoType == MPEG){
		if ( ctx->lastFrameMsg )
		ReplyMsg(ctx->lastFrameMsg, 0, ctx->lastMPEGBufferPtr,
			sizeof(MPEGBuffer));
	}
	status = DisposeDataStream(ctx->messageItem, ctx->streamCBPtr);
	sTestName = "DisposeDataStream";
	BeginTest(FT, sTestName);
	if (status < 0){
		LogResult( FAILED, "%s fails with error %d", sTestName, (int)status);
		goto FAILED;
	}
	else {
		LogResult( PASSED, "%s passes", sTestName);
	}
	status = DDisposeDataAcq(ctx->acqMsgPort);
	sTestName = "DisposeDataAcq";
	BeginTest(FT, sTestName);
	if (status < 0){
		LogResult( FAILED, "%s fails with error %d", sTestName, (int)status);
		goto FAILED;
	}
	else {
		LogResult( PASSED, "%s passes", sTestName);
	}	
	if (gVideoType == EZFX){
		sTestName = "DestroyEZFlixStreamState";
		BeginTest(FT, sTestName);
		status = DestroyEZFlixStreamState( ctx->contextPtr, ctx->channelPtr );
		if (status < 0) {
			LogResult( FAILED,"%s fails with error %d", sTestName, (int)status);
		}
		else  {
			LogResult( PASSED, "%s passes", sTestName);		
		}
	}
	FreeMemTrack(ctx->bufferList);
	DeleteMsg(ctx->messageItem);
	DeleteMsg(ctx->endOfStreamMessageItem);
	DeleteMsgPort(ctx->messagePort);
	DeleteMsgPort(ctx->frameCompletionPort);
	DeleteItem(ctx->synchCue);
	
	memset(ctx, 0, sizeof(*ctx));		/* ensures double-dismantle is safe */
FAILED:
	;
}

Err DisplayEZFX(PlayerPtr ctx){
	int32			status;
	int32			playerResult = 0;
	Bitmap*			bitmap;
	uint32			time;
	uint32			frameCount;
	int32			drawingScreen;
	bool 			fStreamIsRunning;

	/* Make sure the first screen is displayed, and we're drawing into
	 * the second screen.
	 */


	status = ModifyGraphicsItemVA(ctx->screenInfoPtr->viewItem,
							VIEWTAG_BITMAP,
								ctx->screenInfoPtr->bitmapItem[0],
							TAG_END);

	if (status < 0) 
		goto FAILED;
	drawingScreen = 1;
	
	/* Get a pointer to the bitmap we will draw into */
	bitmap = ctx->screenInfoPtr->bitmap[ drawingScreen ];
	time = GetAudioTime();
	frameCount = 0;
	while ( TRUE )
	{
		/* If a user function is defined, then call it. If the user function
		 * returns a non-zero value, then exit. */
		if ( ctx->userFn )
		{
			playerResult = (*ctx->userFn)(ctx);
			if ( playerResult )
				break;
		}

		/* Exit if we received the end-of-stream msg reply.
		 * status < 0 is an error. status > 0 is our EOS-reply msg Item. */
		status = GetMsg(ctx->messagePort);
		FAIL_NEG("PlayEZFlixStream GetMsg", status);
		if ( status > 0 )
		{
			time = GetAudioTime() - time;
			TOUCH(time);	/* avert compiler warning of variable set but unused */
			break;
		}

		/* In case there's no EZFlix in this stream, skip
		 * trying to draw something that doesn't exist.
		 */
		if ( ctx->contextPtr != 0 )
		{
			/*
			 * Draw next EZFlix frame directly into the screen buffer.
			 */
			if (DrawEZFlixToBuffer( ctx->contextPtr, ctx->channelPtr, bitmap )) {	
				if (!fDrawEZflixToBuffer){
					sTestName = "DrawEZFlixToBuffer";
					BeginTest(FT, sTestName);
					LogResult( PASSED, "%s passes", sTestName);		
					fDrawEZflixToBuffer = TRUE;
				}
				if (frameCount == 100)
					DoClockTest(ctx);
				if (frameCount == 25) {
					sTestName = "DSIsRunning";
					BeginTest(FT, sTestName);
					fStreamIsRunning = DSIsRunning(ctx->streamCBPtr);
					if (fStreamIsRunning){
						LogResult( PASSED, "%s passes", sTestName);		
					}
					else {
						LogResult( FAILED, "%s fails", sTestName);		
					}
				}
								
				/* Display the screen we just drew, and switch to drawing into
				 * the next screen.
				 */
				status = ModifyGraphicsItemVA(ctx->screenInfoPtr->viewItem,
							VIEWTAG_BITMAP,
								ctx->screenInfoPtr->bitmapItem[drawingScreen],
							TAG_END);
				FAIL_NEG("ModifyGraphicsItemVA", status);

				if (++drawingScreen >= cntFrameBuffers)
					drawingScreen = 0;
				bitmap = ctx->screenInfoPtr->bitmap[ drawingScreen ];

				/* TBD
				 * Flash-write black into the next screen.  The size of frames coming
				 * from the stream can change at any time, so we want to be sure we always
				 * draw over a black background to avoid garbage on the edges when a smaller
				 * frame follows a larger frame.
				 */

				++frameCount;
			}
		}
	}
	FAILED:

	/* Cleanup; deallocate [NB: will be done elsewhere] */
	/* DismantlePlayer(ctx); */
	return ( status < 0 ) ? status : playerResult;
}



/*******************************************************************************************
 * Handle an MPEG frame completion message: schedule its display time and return the previous
 * frame buffer to the streamer for refilling.
 *
 * [TBD] This is currently implemented by waiting here, in this subroutine. This delays the
 *   app's response to other signals like control port signals. The player's main loop should
 *   instead wait on these two additional signals (time to display a buffer, time to return a
 *   buffer for refilling) and do the work then.
 * [TBD] When we're stopped--at least at EOS--don't return buffers for refilling.
 *******************************************************************************************/
int32		HandleMPEGFrames(PlayerPtr ctx)
	{
	int32					status = kDSNoErr;
	Item					theMsg;
	Message					*msgPtr;
	ScreenInfo				*const screenInfo = ctx->screenInfoPtr;
	
	/* Get any pending frames off the frame completion port and display them */
	while ( (theMsg = GetMsg(ctx->frameCompletionPort)) != 0 )
		{
		if ( theMsg > 0 )
			{
			msgPtr = (Message *)LookupItem(theMsg);
				/* [TBD] Should probably error check item lookup */
			
			if ( msgPtr != NULL )
				{
				const MPEGBufferPtr		theBuffer = (MPEGBufferPtr)msgPtr->msg_DataPtr;
				const int32				screenIndex = (int32)theBuffer->userData;

				/* wait until the frame is ready to be displayed, then throw it
				 * up on the screen */
				SleepUntilTime(ctx->synchCue, theBuffer->displayTime);

				status = ModifyGraphicsItemVA(screenInfo->viewItem,
							VIEWTAG_BITMAP, screenInfo->bitmapItem[screenIndex],
							TAG_END);
				FAIL_NEG("HandleMPEGFrames ModifyGraphicsItemVA", status);
				
				/* wait until the previous buffer is no longer being displayed */
				WaitSignal(screenInfo->renderSignal);

				/* Now that we're done with it, return the last frame to the
				 * subscriber to be recycled. */
				if (ctx->lastFrameMsg != 0)
					{
					status = ReplyMsg(ctx->lastFrameMsg, 0, ctx->lastMPEGBufferPtr,
								sizeof(MPEGBuffer));
					FAIL_NEG("HandleMPEGFrames ReplyMsg", status);
					}

				/* Remember this frame so that we can reply to it when the next
				 * frame is ready to display. */
				ctx->lastFrameMsg		= theMsg;
				ctx->lastMPEGBufferPtr	= theBuffer;
				}
			}
		}

FAILED:
	return status;
	}
